/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2008 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Ilia Alshanetsky <iliaa@php.net>                            |
  |          Graham Kelly <graham@php.net>                               |
  +----------------------------------------------------------------------+
*/
/* $Id: optimize_fcr.c,v 1.16 2008/08/07 11:37:19 graham Exp $ */


#include "zend.h"
#include "zend_API.h"
#include "zend_constants.h"
#include "zend_compile.h"
#include "zend_extensions.h"
#include "zend_ini.h"

#include "SAPI.h" 
#include "php.h"

#include "ext/standard/basic_functions.h"
#include "ext/standard/pageinfo.h"
#include "ext/standard/info.h"
#include "ext/standard/md5.h"
#include "ext/standard/sha1.h"
#include "ext/standard/crc32.h"
#include "ext/standard/php_string.h"
#include "ext/standard/php_math.h"
#include "ext/standard/php_rand.h"
#include "ext/standard/html.h"
#include "ext/standard/php_versioning.h"

#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif

#include "php_optimizer.h"
#include "optimize.h"
#include "optimizer_debug.h"
#include "optimize_fcr.h"


PHPAPI extern char *php_ini_scanned_files;
static void optimize_fcall_fcr(zend_op* op, zend_op* prev, CB* cb, zend_op_array* op_array, zend_op** Ts, HashTable* fh, fcr_func_entry* fe, int cofc, zend_bool* optimization_done OPTIMIZER_STATS_PARAM TSRMLS_DC);

const fcr_func_entry fn[FN_BASIC] = {
	FCR_FUNC(defined,               DEFINED,               OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(extension_loaded,      EXTENSION_LOADED,      OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(function_exists,       FUNCTION_EXISTS,       OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_callable,           IS_CALLABLE,           OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(class_exists,          CLASS_EXISTS,          OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(interface_exists,      INTERFACE_EXISTS,      OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(strlen,                STRLEN,                OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(getlastmod,            GETLASTMOD,            OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(getmyuid,              GETMYUID,              OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(getmygid,              GETMYGID,              OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(getmyinode,            GETMYINODE,            OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(get_current_user,      GET_CURRENT_USER,      OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(phpversion,            PHPVERSION,            OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(zend_version,          ZEND_VERSION,          OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(php_sapi_name,         PHP_SAPI_NAME,         OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(php_logo_guide,        PHP_LOGO_GUIDE,        OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(zend_logo_guide,       ZEND_LOGO_GUIDE,       OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(php_real_logo_guide,   PHP_REAL_LOGO_GUIDE,   OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(php_uname,             PHP_UNAME,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(php_ini_scanned_files, PHP_INI_SCANNED_FILES, OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(md5,                   MD5,                   OPTIMIZER_IS_LONG | OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(sha1,                  SHA1,                  OPTIMIZER_IS_LONG | OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(crc32,                 CRC32,                 OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(pi,                    PI,                    OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(getrandmax,            GETRANDMAX,            OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(mt_getrandmax,         MT_GETRANDMAX,         OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(get_magic_quotes_gpc,  GET_MAGIC_QUOTES_GPC,  OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(ord,                   ORD,                   OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(chr,                   CHR,                   OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(strtoupper,            STRTOUPPER,            OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(strtolower,            STRTOLOWER,            OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(strcmp,                STRCMP,                OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(addslashes,            ADDSLASHES,            OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(htmlspecialchars,      HTMLSPECIALCHARS,      OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(htmlentities,          HTMLENTITIES,          OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(dirname,               DIRNAME,               OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY) 
	FCR_FUNC(basename,              BASENAME,              OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(realpath,              REALPATH,              OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(readlink,              READLINK,              OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(file_exists,           FILE_EXISTS,           OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_link,               IS_LINK,               OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_dir,                IS_DIR,                OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_file,               IS_FILE,               OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_executable,         IS_EXECUTABLE,         OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_readable,           IS_READABLE,           OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(is_writeable,          IS_WRITEABLE,          OPTIMIZER_IS_BOOL,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(fileatime,             FILEATIME,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(filectime,             FILECTIME,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(filegroup,             FILEGROUP,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(fileinode,             FILEINODE,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(filemtime,             FILEMTIME,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(fileowner,             FILEOWNER,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(fileperms,             FILEPERMS,             OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(filesize,              FILESIZE,              OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(filetype,              FILETYPE,              OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(strncmp,               STRNCMP,               OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(strcasecmp,            STRCASECMP,            OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(strncasecmp,           STRNCASECMP,           OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(version_compare,       VERSION_COMPARE,       OPTIMIZER_IS_LONG | OPTIMIZER_IS_BOOL, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(dechex,                DECHEX,                OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(hexdec,                HEXDEC,                OPTIMIZER_IS_LONG | OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(decbin,                DECBIN,                OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(decoct,                DECOCT,                OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(octdec,                OCTDEC,                OPTIMIZER_IS_LONG | OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(str_repeat,            STR_REPEAT,            OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(intval,                INTVAL,                OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(substr,                SUBSTR,                OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(acos,                  ACOS,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(acosh,                 ACOSH,                 OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(asin,                  ASIN,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(asinh,                 ASINH,                 OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(atan,                  ATAN,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(atanh,                 ATANH,                 OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(cos,                   COS,                   OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(cosh,                  COSH,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(sin,                   SIN,                   OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(sinh,                  SINH,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(tan,                   TAN,                   OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(tanh,                  TANH,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(exp,                   EXP,                   OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(log10,                 LOG10,                 OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(sqrt,                  SQRT,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(atan2,                 ATAN2,                 OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(ceil,                  CEIL,                  OPTIMIZER_IS_DOUBLE | OPTIMIZER_IS_BOOL_F, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(floor,                 FLOOR,                 OPTIMIZER_IS_DOUBLE | OPTIMIZER_IS_BOOL_F, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(fmod,                  FMOD,                  OPTIMIZER_IS_DOUBLE, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(ini_get,               INI_GET,               OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY)
	FCR_FUNC(ip2long,               IP2LONG,               OPTIMIZER_IS_LONG,   FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(long2ip,               LONG2IP,               OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(trim,                  TRIM,                  OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(rtrim,                 RTRIM,                 OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(ltrim,                 LTRIM,                 OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(chop,                  CHOP,                  OPTIMIZER_IS_STRING, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	
	FCR_FUNC(printf,                PRINTF,                OPTIMIZER_IS_LONG,     FN_OPTIMIZABLE | FN_EXTERNAL_MODIFIER | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(fopen,                 FOPEN,                 OPTIMIZER_IS_RESOURCE | OPTIMIZER_IS_BOOL_F, FN_OPTIMIZABLE | FN_EXTERNAL_MODIFIER)
	FCR_FUNC(readfile,              READFILE,              OPTIMIZER_IS_LONG | OPTIMIZER_IS_BOOL_F,     FN_OPTIMIZABLE | FN_EXTERNAL_MODIFIER)
	FCR_FUNC(file,                  FILE,                  OPTIMIZER_IS_ARRAY | OPTIMIZER_IS_BOOL_F,    FN_OPTIMIZABLE | FN_EXTERNAL_MODIFIER)
	FCR_FUNC(file_get_contents,     FILE_GET_CONTENTS,     OPTIMIZER_IS_STRING | OPTIMIZER_IS_BOOL_F,   FN_OPTIMIZABLE | FN_EXTERNAL_MODIFIER)
	FCR_FUNC(file_put_contents,     FILE_PUT_CONTENTS,     OPTIMIZER_IS_LONG | OPTIMIZER_IS_BOOL_F,     FN_OPTIMIZABLE | FN_EXTERNAL_MODIFIER)
	
	FCR_FUNC(rad2deg,               RAD2DEG,               OPTIMIZER_IS_DOUBLE,   FN_OPTIMIZABLE | FN_OPTIMIZABLE_STATIC | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(deg2rad,               DEG2RAD,               OPTIMIZER_IS_DOUBLE,   FN_OPTIMIZABLE | FN_OPTIMIZABLE_STATIC | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(abs,                   ABS,                   OPTIMIZER_IS_DOUBLE | OPTIMIZER_IS_LONG | OPTIMIZER_IS_BOOL_F, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
	FCR_FUNC(get_cfg_var,           GET_CFG_VAR,           OPTIMIZER_IS_STRING | OPTIMIZER_IS_BOOL_F, FN_OPTIMIZABLE_STATIC_ONLY | FN_OPTIMIZE_SILENCE)
#if 0
    /* "max", */
    /* "min", */
    /* "pow", */
    /* "round", */
    
    /* Array Functions */
	/* "array_count_values", */
    /* "array_fill_keys", */
    /* "array_fill", */
    /* "array_flip", */
    /* "array_key_exists", */
    /* "array_keys", */
    /* "array_merge", */
    /* "array_pad", */
    /* "array_product", */
    /* "array_push", */
    /* "array_reverse", */
    /* "array_sum", */
    /* "array_unique", */
    /* "array_values", */
    /* "arsort", */
    /* "asort", */
    /* "compact", */
    /* "count", */
    /* "in_array", */
    /* "list", */
    /* "rsort", */
    /* "sizeof", */
    /* "sort", */
    
    /* f functions */
	/* "sprintf", */
	/* "vprintf", */
	/* "vsprintf", */
	/* "sscanf", */
	
	/* Variable Functions */
	/* "doubleval", */
	/* "empty", */
	/* "floatval", */
	/* "gettype", */
	/* "is_array", */
	/* "is_binary", */
	/* "is_bool", */
	/* "is_buffer", */
	/* "is_double", */
	/* "is_float", */
	/* "is_int", */
	/* "is_integer", */
	/* "is_long", */
	/* "is_null", */
	/* "is_numeric", */
	/* "is_object", */
	/* "is_real", */
	/* "is_resource", */
	/* "is_scalar", */
	/* "is_string", */
	/* "is_unicode", */
	/* "isset", */
	/* "serialize", */


    /* "get_class_methods", */
	/* "get_class_vars", */
	/* "get_class", */
	/* "get_parent_class", */
	/* "is_a", */
	/* "is_subclass_of", */
	/* "method_exists", */
	/* "property_exists", */
	
	/* apache_get_version */
	
	/* range, */
	/* get_html_translation_table, */
	/* get_loaded_extensions, */
	/* stream_get_transports, */
#endif
};


#ifndef INADDR_NONE
# define INADDR_NONE ((unsigned long int) -1)
#endif

/* Windows safe versions of acosh, asinh, and atanh based on the php_* versions in math.c by kalle */
static double optimizer_acosh(double z)
{
#if HAVE_ACOSH
	return acosh(z);
#else
	return (log(x + sqrt(x * x - 1)));
#endif
}

static double optimizer_asinh(double z)
{
#if HAVE_ASINH
	return asinh(z);
#else
	return(log(z + sqrt(1 + pow(z, 2))) / log(M_E));
#endif
}

static double optimizer_atanh(double z)
{
#if HAVE_ATANH
	return atanh(z);
#else
	return (0.5 * log((1 + z) / (1 - z)));
#endif
}


zend_bool fcall_return_value_used(zend_op* op) 
{
	if (get_next_znode_pos(op, 1)) {
		return 1;
	}
	
	return 0;
}


/*
 * Optimize calls to md5 when it has a constant value being hashed
 * 
 */
static void optimize_md5(zend_op** a, zend_op** b, CB* cb, zend_op** Ts, zend_bool raw TSRMLS_DC) 
{
	zend_op *prev = *a;
	char md5str[33];
	PHP_MD5_CTX context;
	unsigned char digest[16];

	convert_to_string(&__OP1_VAL(prev));
	md5str[0] = '\0';
	PHP_MD5Init(&context);
	PHP_MD5Update(&context, (const unsigned char *)OP1_VAL_S(prev), OP1_VAL_SL(prev));
	PHP_MD5Final(digest, &context);

	if (raw) {
		optimize_fcall_to_string((char *)digest, sizeof(digest), a, b, (*b)->extended_value, cb, Ts);
	} else {
		make_digest(md5str, digest);
		optimize_fcall_to_string(md5str, sizeof(md5str)-1, a, b, (*b)->extended_value, cb, Ts);
	}
}



/*
 * Optimize sha1 when it has a constant value being hashed
 * 
 */
static void optimize_sha1(zend_op** a, zend_op** b, CB* cb, zend_op** Ts, zend_bool raw TSRMLS_DC) 
{
	zend_op *prev = *a;
	char sha1str[41];
	PHP_SHA1_CTX context;
	unsigned char digest[20];

	convert_to_string(&__OP1_VAL(prev));
	sha1str[0] = '\0';
	PHP_SHA1Init(&context);
	PHP_SHA1Update(&context, (const unsigned char *)OP1_VAL_S(prev), OP1_VAL_SL(prev));
	PHP_SHA1Final(digest, &context);

	if (raw) {
		optimize_fcall_to_string((char *)digest, sizeof(digest), a, b, (*b)->extended_value, cb, Ts);
	} else {
		make_sha1_digest(sha1str, digest);
		optimize_fcall_to_string(sha1str, sizeof(sha1str)-1, a, b, (*b)->extended_value, cb, Ts);
	}
}



/*
 * Optimize crc32 when it has a constant value being hashed
 * 
 */
static void optimize_crc32(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;
	char *p = OP1_VAL_S(prev);
	int len, nr = OP1_VAL_SL(prev);
	php_uint32 crcinit = 0;
	register php_uint32 crc;

	crc = crcinit^0xFFFFFFFF;

	for (len =+nr; nr--; ++p) {
		crc = ((crc >> 8) & 0x00FFFFFF) ^ crc32tab[(crc ^ (*p)) & 0xFF ];
	}

	optimize_fcall_to_long(crc^0xFFFFFFFF, a, b, 1, cb, Ts);
}



/*
 * Optimize extension loaded.
 * 
 */
static int optimize_extension_loaded(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;
	Bucket *p = module_registry.pListHead;

	convert_to_string(&__OP1_VAL(prev));
	while (p != NULL) {
		zend_module_entry *m = (zend_module_entry*)p->pData;
		if (m->type == MODULE_PERSISTENT && !strcasecmp(m->name, OP1_VAL_S(prev))) {
			if (b != NULL) {
				optimize_fcall_to_true(a, b, 1, cb, Ts);
			}
			return 1;
		}
		p = p->pListNext;
	}
	return 0;
}



/*
 * Optimize class_exists 
 * TODO: add a check for classes located within the same source file... while 
 *   this doesnt make much sense for someone to check for the existance of a 
 *   class in the same file... its not much work to optimize it out as well
 */
static zend_bool optimize_class_exists(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;
	zend_class_entry **ce;
	char *lc_name;
	zend_bool retval;

	convert_to_string(&__OP1_VAL(prev));
	lc_name = do_alloca(OP1_VAL_SL(prev) + 1);
	zend_str_tolower_copy(lc_name, OP1_VAL_S(prev), OP1_VAL_SL(prev));

	if (zend_hash_find(EG(class_table), lc_name, OP1_VAL_SL(prev)+1, (void **) &ce) == SUCCESS && !((*ce)->ce_flags & ZEND_ACC_INTERFACE)) {
		optimize_fcall_to_true(a, b, 1, cb, Ts);
		retval = 1;
	} else {
		retval = 0;
	}
	free_alloca(lc_name);
	return retval;
}



/*
 * Optimize interface_exists 
 * TODO: add a check for interfaces located within the same soruce file.
 */
static zend_bool optimize_interface_exists(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;
	zend_class_entry **ce;
	char *lc_name;
	zend_bool retval;

	convert_to_string(&__OP1_VAL(prev));
	lc_name = do_alloca(OP1_VAL_SL(prev) + 1);
	zend_str_tolower_copy(lc_name, OP1_VAL_S(prev), OP1_VAL_SL(prev));

	if (zend_hash_find(EG(class_table), lc_name, OP1_VAL_SL(prev)+1, (void **) &ce) == SUCCESS && (*ce)->ce_flags & ZEND_ACC_INTERFACE) {
		optimize_fcall_to_true(a, b, 1, cb, Ts);
		retval = 1;
	} else {
		retval = 0;
	}
	free_alloca(lc_name);
	return retval;
}



/*
 * Optimize function exists
 * TODO: Add a check for functions located within the same source file
 */
static int optimize_function_exists(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;
	zend_function *func;
	int retval = 0;
	char *lcname;

	convert_to_string(&__OP1_VAL(prev));
	lcname = estrndup(OP1_VAL_S(prev), OP1_VAL_SL(prev));
	zend_str_tolower(lcname, OP1_VAL_SL(prev));

	if (zend_hash_find(EG(function_table), lcname, OP1_VAL_SL(prev) + 1, (void **)&func) == SUCCESS) {
		if (func->type == ZEND_INTERNAL_FUNCTION && func->internal_function.handler == zif_display_disabled_function) {
			/* disabled native function */
		} else {
			if (b != NULL) {
				optimize_fcall_to_true(a, b, 1, cb, Ts);
				retval = 1;
			}
		}
	}
	efree(lcname);
	return retval;
}



/*
 * Optimize defined
 * 
 */
static void optimize_defined(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;
	char *p;	

	convert_to_string(&__OP1_VAL(prev));
	if ((p = memchr(OP1_VAL_S(prev), ':', OP1_VAL_SL(prev))) && *(p + 1) == ':') { /* class constant */
		*p = '\0';
		if (optimize_get_constant(p+2, OP1_VAL_S(prev) + OP1_VAL_SL(prev) - p-2, OP1_VAL_S(prev), p - OP1_VAL_S(prev), NULL TSRMLS_CC)) {
			*p = ':';
			goto od;
		}
		*p = ':';
	} else if (optimize_get_constant(OP1_VAL_S(prev), OP1_VAL_SL(prev), NULL, 0, NULL TSRMLS_CC)) { /* basic constant */
od:
		optimize_fcall_to_true(a, b, 1, cb, Ts);
	}
}


/*
 * Optimzie str_repeat() for constant values
 * 
 */
static void optimize_str_repeat(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	zend_op *prev = *a;

	char *r, *p;
	int r_len, p_len;

	long mult;

	zend_op* arg2 = prev; 
	zend_op* arg1 = PREV_OP(arg2);

	convert_to_long(&__OP1_VAL(arg2));
	convert_to_string(&__OP1_VAL(arg1));

	p = OP1_VAL_S(arg1);
	p_len = OP1_VAL_SL(arg1);
	mult = OP1_VAL_L(arg2);

	r_len = p_len * mult;
	r = (char *)safe_emalloc(p_len, mult, 1);
	
	if (p_len == 1) {
		memset(r, *p, mult); 
	} else {
		char *s, *e, *ee;
		int l = 0;

		memcpy(r, p, p_len);
		s = r;
		e = r + p_len;
		ee = r + r_len;
		
		while (e<ee) {
			l = (e-s) < (ee-e) ? (e-s) : (ee-e);
			memmove(e, s, l);
			e += l;
		}
	}
	r[r_len] = '\0';
	optimize_fcall_to_string(r, r_len, a, b, (*b)->extended_value, cb, Ts);

	efree(r);
	zval_dtor(&__OP1_VAL(arg1));
	SET_TO_NOP(arg1);
}



/*
 * Optimize calls to substr() with all static params
 *
 */
static void optimize_substr(zend_op** prev, zend_op** op, CB* cb, zend_op** Ts OPTIMIZER_STATS_PARAM TSRMLS_DC) 
{
	zval str;
 	zend_op* str_op;
 	zend_op* f_op;
 	zend_op* l_op = NULL;
 	long f = 0, l = 0;
 	
 	if ((*op)->extended_value == 3) {
 		l_op = PREV_OP(*op);
 		f_op = PREV_OP(l_op);
 		str_op = PREV_OP(f_op);
 		convert_to_long(&__OP1_VAL(l_op));
 		l = OP1_VAL_L(l_op);
 	} else {
 		f_op = PREV_OP(*op);
 		str_op = PREV_OP(f_op);
 	}
 					
 	convert_to_long(&__OP1_VAL(f_op));
 	f = OP1_VAL_L(f_op);
 					
 	convert_to_string(&__OP1_VAL(str_op));
 	str = __OP1_VAL(str_op);
 					
 	if ((*op)->extended_value == 2) {
 		l = OP1_VAL_SL(str_op);
 	}
 					
 	if (f > Z_STRLEN(str) || (f < 0 && -f > Z_STRLEN(str))) {
 		optimize_fcall_to_false(prev, op, (*op)->extended_value, cb, Ts);

 		zval_dtor(&__OP1_VAL(str_op));
 		SET_TO_NOP(str_op);
 		if ((*op)->extended_value == 3) {
 			zval_dtor(&__OP1_VAL(f_op));
 			SET_TO_NOP(f_op);
 		}
 						
 		OPTIMIZER_STATS_ADD("Optimized call to substr to false");
 	} else if (l < 0 && (l + Z_STRLEN(str) - f) < 0) {
 		optimize_fcall_to_false(prev, op, (*op)->extended_value, cb, Ts);
 		
 		zval_dtor(&__OP1_VAL(str_op));
 		SET_TO_NOP(str_op);
 		if ((*op)->extended_value == 3) {
 			zval_dtor(&__OP1_VAL(f_op));
 			SET_TO_NOP(f_op);
 		}	
 					
 		OPTIMIZER_STATS_ADD("Optimized call to substr to false");
 	} else {
 		/* if 'from' position is negative */
 		if (f < 0) {
 			f = Z_STRLEN(str) + f;
 			if (f < 0) {
 				f = 0;
 			}
 		}
 						
 		/* if 'length' position is negative */
 		if (l < 0) {
 			l = (Z_STRLEN(str) - f) + l;
 			if (l < 0) {
 				l = 0;
 			}
 		}
 						
 		if (f >= Z_STRLEN(str)) {
 			optimize_fcall_to_false(prev, op, (*op)->extended_value, cb, Ts);
 							
 			zval_dtor(&__OP1_VAL(str_op));
 			SET_TO_NOP(str_op);
 			if ((*op)->extended_value == 3) {
 				zval_dtor(&__OP1_VAL(f_op));
 				SET_TO_NOP(f_op);
 			}
 							
 			OPTIMIZER_STATS_ADD("Optimized call to substr to false");
 		} else {
 			if ((f + l) > Z_STRLEN(str)) {
 				l = Z_STRLEN(str) - f;
 			}

			char* s = Z_STRVAL(str);
			s += f;
			s = estrndup((const char*)s, (size_t)l);
 			optimize_fcall_to_string(s, l, prev, op, (*op)->extended_value, cb, Ts);
 			efree(s);
 			
 			zval_dtor(&__OP1_VAL(str_op));
 			SET_TO_NOP(str_op);
 			if ((*op)->extended_value == 3) {
 				zval_dtor(&__OP1_VAL(f_op));
 				SET_TO_NOP(f_op);
 			}
 			
 			OPTIMIZER_STATS_ADD("Optimized out call to substr()");
 		}
 	}
}


/*
 * Optimize call to trim(), ltrim(), or rtrim()
 *   Modes:
 *     mode 1 : trim left
 *     mode 2 : trim right
 *     mode 3 : trim left and right
 */
static void optimize_trim(int mode, zend_op** prev, zend_op** op, CB* cb, zend_op** Ts OPTIMIZER_STATS_PARAM TSRMLS_DC)
{
	if ((*op)->extended_value == 1 || (*op)->extended_value == 2) {
		char* r;
		zend_op* arg2 = *prev;
		zend_op* arg1 = NULL;
		
		convert_to_string(&__OP1_VAL(arg2));
		
		if ((*op)->extended_value == 2) {
			arg1 = PREV_OP(arg2); 
			convert_to_string(&__OP1_VAL(arg1));
			r = php_trim(OP1_VAL_S(arg1), OP1_VAL_SL(arg1), OP1_VAL_S(arg2), OP1_VAL_SL(arg2), NULL, mode TSRMLS_CC);
		} else {
			r = php_trim(OP1_VAL_S(arg2), OP1_VAL_SL(arg2), NULL, 0, NULL, mode TSRMLS_CC);
		}
		
		optimize_fcall_to_string(r, strlen(r), prev, op, (*op)->extended_value, cb, Ts);
		efree(r);
		
		if ((*op)->extended_value == 2) {
			zval_dtor(&__OP1_VAL(arg1));
			SET_TO_NOP(arg1);
		}
		
		OPTIMIZER_STATS_ADD("Optimized out call to trim(), rtrim(), or ltrim()");
	}
}


/*
 * Optimize file system functions when possible (when optimize.file_ops=1 in ini)
 *   This means that the results of these functions will be pre-calculated, this could possibly be a problem on a server where files change frequently. It is usually recomended that this is not used. 
 */
static void optimize_file_op(int type, zend_op** a, zend_op** b, CB* cb, zend_op** Ts OPTIMIZER_STATS_PARAM TSRMLS_DC) 
{
	zval ret;
	zend_op *prev = *a;
	zend_op *op = *b;

	if (op->extended_value != 1 || !OPTIMIZER_G(optimize_file_ops)) {
		if (op->extended_value == 1) {
			optimize_fullpath(prev, NULL, 0 TSRMLS_CC);
		}
		return;
	}

	convert_to_string(&__OP1_VAL(prev));
	php_stat(OP1_VAL_S(prev), (php_stat_len) OP1_VAL_SL(prev), type, &ret TSRMLS_CC);

	switch (Z_TYPE(ret)) {
		case IS_BOOL:
			OPTIMIZE_FCALL_TO_BOOL(Z_LVAL(ret), 1);
			OPTIMIZER_STATS_ADD("Optimized file op\n");
			break;
		case IS_LONG:
			optimize_fcall_to_long(Z_LVAL(ret), a, b, 1, cb, Ts);
			OPTIMIZER_STATS_ADD("Optimized file op\n");
			break;
		case IS_STRING:
			optimize_fcall_to_string(Z_STRVAL(ret), Z_STRLEN(ret), a, b, 1, cb, Ts);
			OPTIMIZER_STATS_ADD("Optimized file op\n");
			break;
	}
}


static void optimize_printf(zend_op** a, zend_op** b, CB* cb, zend_op** Ts OPTIMIZER_STATS_PARAM TSRMLS_DC) 
{
	zend_op* op = *a;
	zend_op* prev = *b;
	
	if (op->extended_value == 1) {
		/* Just printing... optimize to echo */
		int g = get_next_znode_pos(op, 1);
		
		if (g == 0) {
			prev->opcode = ZEND_ECHO;
			zval_dtor(&__OP1_VAL(op));
			SET_TO_NOP(op);
		} else {
			convert_to_string(&__OP1_VAL(prev));

			char* c = OP1_VAL_S(prev);
			int l = OP1_VAL_SL(prev);
			OP1_VAL_S(prev) = OP1_VAL_S(op);
			OP1_VAL_SL(prev) = OP1_VAL_SL(op);
			OP1_VAL_S(op) = c;
			OP1_VAL_SL(op) = l;
			
			op->opcode = ZEND_ECHO;
			RES_TYPE(op) = IS_UNUSED;
			OP2_TYPE(op) = IS_UNUSED;
			zval_dtor(&__OP1_VAL(prev));
			SET_TO_NOP(prev);
			
			/* NOT optimize_fcall_to_long as we need to keep op for the ZEND_ECHO */
			optimize_to_long(l, &prev, &op, cb, Ts);
		}
		OPTIMIZER_STATS_ADD();
	} //else {
		/* real printf */
		/* We will optimize this later... */
	//}
}


#if 0
static void optimize_pow(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	if (OP1_TYPE((op-2)) == IS_CONST && OP1_TYPE((op-1)) == IS_CONST) {
		/* we have two constant expressions */
	} else if (OP1_TYPE((op-1)) == IS_CONST && OP1_VALT((op-1)) == ZEND_LONG && OP1_VAL((op-1)) <= 5) {
		int exp = OP1_VAL((op-1));
  	    int r_eq_b = 0;
  	
  	
		switch () {
  			case 0:
  				/* pow($a, 0) is always going to equal 1 */
  				break;
  			case 1:
  				/* pow($a, 1) is always going to equal $a (if we have $a = pow($a, 1) we can eliminate these ops all together as we are basicalyl saying $a = $a) */
  				if (r_eq_b) {
  					/* $a = pow($a, 1); -> NOP */
  				} else {
  					/* $v = pow($a, 1); -> $v = $a; */
  				}
  				break;
  			case 2:
  				if (r_eq_b) {
  					/* $a = pow($a, 1); -> $a *= $a; */
  				} else {
  					/* $v = pow($a, 1); -> $v = $a*$a; */
  				}
  				break;
  			case 3:
  				/* $v = pow($a, 3); -> !tmp = $a*$a; $v = !tmp*$a */
  				break;
  			case 4:
  				break;
  			case 5:
  				break;	
		};
	}
}
#endif
/*
static void optimize_abs(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) {}
static void optimize_min(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) {}
static void optimize_max(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) {}
static void optimize_rad2deg(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) {}
static void optimize_deg2rad(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) {}
*/

void fcr_table_ctor(HashTable* fh)
{
	int i = 0;

	zend_hash_init(fh, FN_BASIC, NULL, NULL, 0);
	for (i = 0; i < FN_BASIC; i++) {
		fcr_func_entry* fe = (fcr_func_entry*)&fn[i];

#if PHP_VERSION_ID < 50300
		zend_bool d = 1;
		
		/* If we do not have acosh, asinh, or atanh then we dont want the info 
		 * about the funcs as they wont be in 5.2 builds on windows 
		 */
#  ifndef HAVE_ACOSH
		d = (fe->fid == FN_ACOSH ? 0 : 1);
#  endif
#  ifndef HAVE_ASINH
		d = (fe->fid == FN_ASINH ? 0 : 1);
#  endif
#  ifndef HAVE_ATANH
		d = (fe->fid == FN_ATANH ? 0 : 1);
#  endif

		if (d) {
			zend_hash_update(fh, (char *)fe->name, strlen(fe->name) + 1, (void**)fe, sizeof(fcr_func_entry), NULL);
		}
#else
		zend_hash_update(fh, (char *)fe->name, strlen(fe->name) + 1, (void**)fe, sizeof(fcr_func_entry), NULL);
#endif
	}
}

/*
 * Determines if a function call appears to be able to be optimized out.
 *   passing this test does NOT garuntee that a fucntion will be a cenidate
 *   for optimizing out.
 *
 * Return values:
 *   0: Unable to optimize the function call
 *   1: OP1 is a var on the next op
 *   2: OP2 is a var on the next op
 *   3: 
 *   4:
 */
int can_optimize_fcall(zend_op* op) 
{
	int i;
	zend_op* x = op;

	for (i = 0; i < op->extended_value; i++) {
		x = PREV_OP(x);

		if ((x->opcode != ZEND_SEND_VAL || OP1_TYPE(x) != IS_CONST) && x->opcode != ZEND_NOP) {
			return 0;
		}
	}

	return get_next_znode_pos(op, 1);
}



void optimize_fcall(zend_op** a, zend_op** b, CB* cb, zend_op_array* op_array, zend_op** Ts, HashTable* fh OPTIMIZER_STATS_PARAM TSRMLS_DC) 
{
	zend_op* op = *a;
	zend_op* prev = *b;

	if (op->opcode == ZEND_DO_FCALL && OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_STRING) {
		zend_bool od = 0;
		
		if (OPTIMIZER_G(optimize_out_funcs_l) >= OP1_VAL_SL(op)) {
			/*
			 * Optimize out functions based on a list provided in the ini
			 * This is probably a bit buggy and doesnt clean up after itself... 
			 *   thus it might leave a lot of dead code. :(
			 *
			 * TODO: Use some sort of expression tree to fully understand the fcall we are NOPing
			 */
			zend_bool do_oof = 1;
			char *oof = OPTIMIZER_G(optimize_out_funcs);
			char *e = oof+OPTIMIZER_G(optimize_out_funcs_l);
			char *p = php_memnstr(oof, OP1_VAL_S(op), OP1_VAL_SL(op), e);

			if (!p) {
				do_oof = 0;
			}
			if (do_oof && p > oof && *(p-1) != ' ') {
				do_oof = 0;
			}
			if (do_oof && p+OP1_VAL_SL(op) < e && *(p+OP1_VAL_SL(op)) != ' ') {
				do_oof = 0;
			}
				
			if (do_oof) {
				zend_bool ret_used = fcall_return_value_used(op);
				int send_vars_left = op->extended_value;
				int i = 1;
				
				while (send_vars_left > 0) {
					zend_op* arg = op-i;
					
					switch (arg->opcode) {
						case ZEND_SEND_VAL:
						case ZEND_SEND_VAR:
						case ZEND_SEND_VAR_NO_REF:
								if (OP1_TYPE(arg) == IS_CONST) {
									zval_dtor(&__OP1_VAL(arg));
								}
					
								if (OP2_TYPE(arg) == IS_CONST) {
									zval_dtor(&__OP2_VAL(arg));
								}
							
								SET_TO_NOP(arg);
								send_vars_left--;
							break;
						case ZEND_DO_FCALL:
						case ZEND_DO_FCALL_BY_NAME:
							send_vars_left += arg->extended_value;
							break;
					}
					
					i++;
				}
				
				zval_dtor(&__OP1_VAL(op));
				SET_TO_NOP(op);
				
				if (ret_used) {
					optimize_to_null(&prev, &op, cb, Ts);
				}

				od = 1;
			}
		}
		
		
		/* get function information from the table of optimizable functions... */
		fcr_func_entry* fe = NULL;
		if (zend_hash_find(fh, OP1_VAL_S(op), OP1_VAL_SL(op)+1, (void**)&fe) != SUCCESS) {
			/* set optimization done to 1 so that we dont do any optimizations */
			od = 1;
			fe = NULL;
		}
		
		/* Optimize out silence blocks */
		if (!od && fe->flags & FN_OPTIMIZE_SILENCE) {
			zend_op* se = NEXT_OP(op);
			
			if (se->opcode == ZEND_END_SILENCE) {
				int i = 0;
				zend_bool optimize_sblock = 1;
				zend_op* sb = op;
				
				do {
					i++;
					sb = op-i;
					
					if (sb->opcode == ZEND_DO_FCALL && OP1_TYPE(sb) == IS_CONST && OP1_VALT(sb) == IS_STRING) {
						/* 
						 * We have another fcall insde the silence block for the func we 
						 *   are optimizing. So, we need to check to see if this 
						 *   function can have silence blocks removed from it as 
						 *   well... if not we finish our search here.
						 */
						fcr_func_entry* sb_fe = NULL;
						if (zend_hash_find(fh, OP1_VAL_S(sb), OP1_VAL_SL(sb)+1, (void**)&sb_fe) != SUCCESS) {
							optimize_sblock = 0;
						} else if (!(sb_fe->flags & FN_OPTIMIZE_SILENCE)) {
							optimize_sblock = 0;
						}
					}
				} while(optimize_sblock == 1 && sb->opcode != ZEND_BEGIN_SILENCE && sb != op_array->opcodes);
				
				if (optimize_sblock == 1 && sb != NULL && sb->opcode == ZEND_BEGIN_SILENCE) {
					SET_TO_NOP(sb);
					SET_TO_NOP(se);
				}
			}
		}
		
		
		int cofc = 0;
		if (!od && fe->flags & FN_OPTIMIZABLE_STATIC && (cofc = can_optimize_fcall(op)) != 0) {
			/* Optimize only when we have all static parameters */
			optimize_fcall_fcr(op, prev, cb, op_array, Ts, fh, fe, cofc, &od OPTIMIZER_STATS_PARAM_PASS TSRMLS_DC);
			od = 1;
		}
		
		if (!od && fe->flags & FN_OPTIMIZABLE && !(fe->flags & FN_OPTIMIZABLE_STATIC_ONLY_MASK)) {
			/* Optimize when we do not have all static parameters */
			switch (fe->fid) {
				case FN_PRINTF:
					/* printf() optimizations */
					optimize_printf(&op, &prev, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
					break;
				case FN_DEG2RAD:
				case FN_RAD2DEG:
					if (op->extended_value == 1) {
						/* EXPERIMENTAL */
						zend_uint cv = OP1_VR(prev);
						
						SET_TO_NOP_EX(op);
						SET_TO_NOP_EX(prev);
						
						op->opcode = ZEND_MUL;
						RES_TYPE(op) = IS_TMP_VAR;
						RES_VR(op) = optimizer_fetch_free_tmp_var(op_array);
						OP1_TYPE(op) = IS_CV;
						OP1_VR(op) = cv;
						OP2_TYPE(op) = IS_CONST;
						OP2_VALT(op) = IS_DOUBLE;
						OP2_VAL_D(op) = (fe->fid == FN_DEG2RAD ? (M_PI/180.0) : (180.0/M_PI));
					
						optimize_to_var(RES_VR(op), IS_TMP_VAR, &prev, &op, cb, Ts);
					}
					break;
				case FN_FOPEN:
				case FN_READFILE:
				case FN_FILE:
				case FN_FILE_GET_CONTENTS:
				case FN_FILE_PUT_CONTENTS:
					/* optimize to full pathname */
					if (op->extended_value >= 1 && (op-op->extended_value)->opcode == ZEND_SEND_VAL && 
					    OP1_TYPE((op-op->extended_value)) == IS_CONST && OP1_VALT((op - op->extended_value)) == IS_STRING &&
					    !IS_ABSOLUTE_PATH(OP1_VAL_S((op - op->extended_value)), OP1_VAL_SL((op - op->extended_value))) &&
					    !php_memnstr(OP1_VAL_S((op - op->extended_value)), "://", sizeof("://") - 1, OP1_VAL_S((op - op->extended_value)) + OP1_VAL_SL((op - op->extended_value)))
					) {
						optimize_fullpath((op - op->extended_value), NULL, 0 TSRMLS_CC);
					}
					break;
			}
		}
	}
}



/*
 * Optimizes a function call.
 *   Optimize out calls with all static params
 * 
 */
static void optimize_fcall_fcr(zend_op* op, zend_op* prev, CB* cb, zend_op_array* op_array, zend_op** Ts, HashTable* fh, fcr_func_entry* fe, int cofc, zend_bool* optimization_done OPTIMIZER_STATS_PARAM TSRMLS_DC) 
{
#if OPTIMIZER_STATS_COMPILE && OPTIMIZER_DEBUG_COMPILE
	/* Make a copy of the fname so we can use this in logging later */
	char* stats_fname = estrdup(OP1_VAL_S(op));
#endif

	/* This allows us to not have to call can_optimize_fcall and zend_hash_find multiple times when comming from optimize_fcall */
	zend_bool can_optimize = 0;
	if (fe == NULL || cofc < 0 || cofc > 4) {
		if (can_optimize_fcall(op) && zend_hash_find(fh, OP1_VAL_S(op), OP1_VAL_SL(op)+1, (void**)&fe) == SUCCESS) {
			can_optimize = 1;
		}
	} else if (fe != NULL && cofc != 0) {
		can_optimize = 1;
	}

	prev = PREV_OP(op);

	if (can_optimize) {
  		switch(fe->fid) {
  			case FN_DEFINED: /* defined() */
  				if (op->extended_value == 1) {
					optimize_defined(&prev, &op, cb, Ts TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				}
  				break;
  			case FN_EXTENSION_LOADED: /* exension_loaded() */
  				if (op->extended_value == 1) {
					optimize_extension_loaded(&prev, &op, cb, Ts TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				}
  				break;
  			case FN_FUNCTION_EXISTS: /* function_exists() */
  			case FN_IS_CALLABLE: /* is_callable() */
  				if (op->extended_value == 1) {
					optimize_function_exists(&prev, &op, cb, Ts TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				}
  				break;
  			case FN_CLASS_EXISTS: /* class_exists() */
  				if (op->extended_value == 1) {
					optimize_class_exists(&prev, &op, cb, Ts TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				} else if (op->extended_value == 2) {
					zend_op *x = prev - 1;
					if (optimize_class_exists(&x, &op, cb, Ts TSRMLS_CC)) {
						SET_TO_NOP(prev);
					}
				}
  				break;
  			case FN_INTERFACE_EXISTS: /* interface_exists() */
  				if (op->extended_value == 1) {
					optimize_interface_exists(&prev, &op, cb, Ts TSRMLS_CC);
				} else if (op->extended_value == 2) {
					zend_op *x = prev - 1;
					if (optimize_interface_exists(&x, &op, cb, Ts TSRMLS_CC)) {
						SET_TO_NOP(prev);
					}
				}
  				break;
  			case FN_STRLEN: /* strlen() */
  				if (op->extended_value == 1) {
					convert_to_string(&__OP1_VAL(prev));
					optimize_fcall_to_long((long)OP1_VAL_SL(prev), &prev, &op, 1, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
  				break;
  			case FN_GETLASTMOD: /* getlastmod() */
  				if (op->extended_value == 0) {
					long lm = php_getlastmod(TSRMLS_C);
					if (lm >= 0) {
						optimize_fcall_to_long(lm, &prev, &op, 0, cb, Ts);
						OPTIMIZER_STATS_ADD_FCR();
					}
				}
  				break;
  			case FN_GETMYUID: /* getmyuid() */
  				if (op->extended_value == 0) {
					php_statpage(TSRMLS_C);
					if (BG(page_uid) >= 0) {
						optimize_fcall_to_long(BG(page_uid), &prev, &op, 0, cb, Ts);	
						OPTIMIZER_STATS_ADD_FCR();
					}
				}
  				break;
  			case FN_GETMYGID: /* getmygid() */
  				if (op->extended_value == 0) {
					php_statpage(TSRMLS_C);
					if (BG(page_gid) >= 0) {
						optimize_fcall_to_long(BG(page_gid), &prev, &op, 0, cb, Ts);
						OPTIMIZER_STATS_ADD_FCR();
					}
				}
 				break;
 			case FN_GETMYINODE: /* getmyinode() */
 				if (op->extended_value == 0) {
					php_statpage(TSRMLS_C);
					if (BG(page_inode) >= 0) {
						optimize_fcall_to_long(BG(page_inode), &prev, &op, 0, cb, Ts);
						OPTIMIZER_STATS_ADD_FCR();
					}
				}
 				break;
 			case FN_GET_CURRENT_USER: /* get_current_user() */
 				if (op->extended_value == 0) {
					char *user = php_get_current_user();
					optimize_fcall_to_string(user, strlen(user), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_PHPVERSION: /* phpversion() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_string(PHP_VERSION, strlen(PHP_VERSION), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				} else if (op->extended_value == 1) { 
					/* module name is provided */
					const char *v;

					convert_to_string(&__OP1_VAL(prev));
					if ((v = zend_get_module_version(OP1_VAL_S(prev)))) {
						optimize_fcall_to_string((char*)v, strlen(v), &prev, &op, 1, cb, Ts);	
						OPTIMIZER_STATS_ADD_FCR();
					}
				}
 				break;
 			case FN_ZEND_VERSION: /* zend_version() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_string(ZEND_VERSION, strlen(ZEND_VERSION), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_PHP_SAPI_NAME: /* php_sapi_name() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_string(sapi_module.name, strlen(sapi_module.name), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_PHP_LOGO_GUIDE: /* php_logo_guide() */
 			case FN_PHP_REAL_LOGO_GUIDE: /* php_real_logo_guide() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_string(PHP_LOGO_GUID, strlen(PHP_LOGO_GUID), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_ZEND_LOGO_GUIDE: /* zend_logo_guide() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_string(ZEND_LOGO_GUID, strlen(ZEND_LOGO_GUID), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_PHP_UNAME: /* php_uname() */
 				if (op->extended_value == 0) {
					char *u = php_get_uname('a');
					optimize_fcall_to_string(u, strlen(u), &prev, &op, 0, cb, Ts);
					efree(u);
				} else if (op->extended_value == 1) {
					char *u;
					convert_to_string(&__OP1_VAL(prev));
					u = php_get_uname(OP1_VAL_S(prev)[0]);
					optimize_fcall_to_string(u, strlen(u), &prev, &op, 1, cb, Ts);
					efree(u);
				}
 				break;
 			case FN_PHP_INI_SCANNED_FILES: /* php_ini_scanned_files() */
 				if (op->extended_value == 0 && strlen(PHP_CONFIG_FILE_SCAN_DIR) && php_ini_scanned_files) {
					optimize_fcall_to_string(php_ini_scanned_files, strlen(php_ini_scanned_files), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_MD5: /* md5() */
 				if (op->extended_value == 1) {
					optimize_md5(&prev, &op, cb, Ts, 0 TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				} else if (op->extended_value == 2) {
					zend_op *p2 = prev - 1;
					convert_to_boolean(&__OP1_VAL(prev));
					optimize_md5(&p2, &prev, cb, Ts, Z_BVAL(__OP1_VAL(prev)) TSRMLS_CC);
					SET_TO_NOP(prev);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_SHA1: /* sha1() */
 				if (op->extended_value == 1) {
					optimize_sha1(&prev, &op, cb, Ts, 0 TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				} else if (op->extended_value == 2) {
					zend_op *p2 = prev - 1;
					convert_to_boolean(&__OP1_VAL(prev));
					optimize_sha1(&p2, &prev, cb, Ts, Z_BVAL(__OP1_VAL(prev)) TSRMLS_CC);
					SET_TO_NOP(prev);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_CRC32: /* crc32() */
 				if (op->extended_value == 1) {
					convert_to_string(&__OP1_VAL(prev));
					optimize_crc32(&prev, &op, cb, Ts TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_PI: /* pi() */
 				if (op->extended_value == 0) {
 					optimize_fcall_to_double(M_PI, &prev, &op, 0, cb, Ts);
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_GETRANDMAX: /* getrandmax() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_long(PHP_RAND_MAX, &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_MT_GETRANDMAX: /* mt_getrandmax() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_long(PHP_MT_RAND_MAX, &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_GET_MAGIC_QUOTES_GPC: /* get_magic_quotes_gpc() */
 				if (op->extended_value == 0) {
					optimize_fcall_to_long(PG(magic_quotes_gpc), &prev, &op, 0, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_ORD: /* ord() */
 				if (op->extended_value == 1) {
					convert_to_string(&__OP1_VAL(prev));
					optimize_fcall_to_long((unsigned char) OP1_VAL_S(prev)[0], &prev, &op, 1, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_CHR: /* chr() */
 				if (op->extended_value == 1) {
					char c[2];
					convert_to_long(&__OP1_VAL(prev));
					c[0] = (char) OP1_VAL_L(prev);
					c[1] = '\0';
					optimize_fcall_to_string(c, 1, &prev, &op, 1, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_STRTOUPPER: /* strtoupper() */
 				if (op->extended_value == 1) {
					convert_to_string(&__OP1_VAL(prev));
					php_strtoupper(OP1_VAL_S(prev), OP1_VAL_SL(prev));
					optimize_fcall_to_string(OP1_VAL_S(prev), OP1_VAL_SL(prev), &prev, &op, 1, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_STRTOLOWER: /* strtolower() */
 				if (op->extended_value == 1) {
					convert_to_string(&__OP1_VAL(prev));
					php_strtolower(OP1_VAL_S(prev), OP1_VAL_SL(prev));
					optimize_fcall_to_string(OP1_VAL_S(prev), OP1_VAL_SL(prev), &prev, &op, 1, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_STRCMP: /* strcmp() */
 				if (op->extended_value == 2) {
 					zend_op* x = PREV_OP(prev);
					convert_to_string(&__OP1_VAL((prev)));
					convert_to_string(&__OP1_VAL(x));

					optimize_fcall_to_long(zend_binary_zval_strcmp(&__OP1_VAL(x), &__OP1_VAL(prev)), &prev, &op, 2, cb, Ts);

					zval_dtor(&__OP1_VAL(x));
					SET_TO_NOP(x);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_ADDSLASHES: /* addslashes() */
 				if (op->extended_value == 1) {
					char *p;
					int len;
					convert_to_string(&__OP1_VAL(prev));
					p = php_addslashes(OP1_VAL_S(prev), OP1_VAL_SL(prev), &len, 0 TSRMLS_CC);
					optimize_fcall_to_string(p, len, &prev, &op, 1, cb, Ts);
					efree(p);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_HTMLSPECIALCHARS: /* htmlspecialchars() */
 			case FN_HTMLENTITIES: /* htmlentities() */
 				if (op->extended_value == 1 || op->extended_value == 2 || op->extended_value == 3) {
					char *charset = NULL;
					long quote_style = ENT_COMPAT;
					char *p, *str = NULL;
					int len=0, str_len=0;

					zend_op* arg3 = prev;
					zend_op* arg2 = prev;
					zend_op* arg1 = prev;

					switch (op->extended_value) {
						case 3:
							convert_to_string(&__OP1_VAL(arg3));
							charset = OP1_VAL_S(arg3);
						case 2:
							arg2 = PREV_OP(arg3);
							convert_to_long(&__OP1_VAL(arg2));
							quote_style = OP1_VAL_L((arg2));
						case 1:
							arg1 = PREV_OP(arg2);
							convert_to_string(&__OP1_VAL(arg1));
							str = OP1_VAL_S((arg1));
							str_len = OP1_VAL_SL((arg1));
					}

					if (str) {
						p = php_escape_html_entities((unsigned char *)str, str_len, &len, (fe->fid != FN_HTMLSPECIALCHARS), quote_style, charset TSRMLS_CC);
						optimize_fcall_to_string(p, len, &prev, &op, op->extended_value, cb, Ts);
						efree(p);
				
						if (op->extended_value == 3) {
							zval_dtor(&__OP1_VAL(arg1));
							SET_TO_NOP((arg1));
						}
						if (op->extended_value >= 2) {
							zval_dtor(&__OP1_VAL(arg2));
							SET_TO_NOP(arg2);
						}
					}
				}
 				break;
 			case FN_DIRNAME: /* dirname() */
 				if (op->extended_value == 1) {
					size_t new_len;
					convert_to_string(&__OP1_VAL(prev));
					new_len = php_dirname(OP1_VAL_S(prev), OP1_VAL_SL(prev));
					optimize_fcall_to_string(OP1_VAL_S(prev), new_len, &prev, &op, 1, cb, Ts);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_BASENAME: /* basename() */
 				do {
					char *string, *suffix = NULL, *ret;
					int string_len, suffix_len = 0;
					size_t ret_len;

					zend_op* arg2;
					zend_op* arg1;

					if (op->extended_value == 1) {
						arg1 = prev;
						convert_to_string(&__OP1_VAL(arg1));
						string = OP1_VAL_S(arg1);
						string_len = OP1_VAL_SL(prev);
					} else if (op->extended_value == 2) {
						arg2 = prev;
						arg1 = PREV_OP(arg2);
						convert_to_string(&__OP1_VAL(arg2));
						convert_to_string(&__OP1_VAL(arg1));
						suffix = OP1_VAL_S(arg2);
						suffix_len = OP1_VAL_SL(arg2);
						string = OP1_VAL_S(arg1);
						string_len = OP1_VAL_SL(arg1);
					} else {
						break;
					}

					php_basename(string, string_len, suffix, suffix_len, &ret, &ret_len TSRMLS_CC);
					optimize_fcall_to_string(ret, ret_len, &prev, &op, op->extended_value, cb, Ts);
					efree(ret);
					if (op->extended_value == 2) {
						zval_dtor(&__OP1_VAL(arg1));
						SET_TO_NOP(arg1);
					}
					break;
				} while (1);
 				break;
 			case FN_REALPATH: /* realpath() */
 				if (op->extended_value == 1) {
					char resolved_path_buff[MAXPATHLEN];
					convert_to_string(&__OP1_VAL(prev));
					if (VCWD_REALPATH(OP1_VAL_S(prev), resolved_path_buff)) {
#ifdef ZTS
						if (!VCWD_ACCESS(resolved_path_buff, F_OK)) {
#endif
							optimize_fcall_to_string(resolved_path_buff, strlen(resolved_path_buff), &prev, &op, 1, cb, Ts);
							OPTIMIZER_STATS_ADD_FCR();
#ifdef ZTS
						}
#endif
					}
				}
 				break;
#ifdef HAVE_SYMLINK
 			case FN_READLINK: /* readlink() */
 				if (op->extended_value == 1 && OPTIMIZER_G(optimize_file_ops)) {
					int ret;
					char buff[MAXPATHLEN];

					convert_to_string(&__OP1_VAL(prev));

					if (PG(safe_mode) && !php_checkuid(OP1_VAL_S(prev), NULL, CHECKUID_CHECK_FILE_AND_DIR)) {
						break;
					}
					if (php_check_open_basedir(OP1_VAL_S(prev) TSRMLS_CC)) {
						break;
					}
					if ((ret = readlink(OP1_VAL_S(prev), buff, MAXPATHLEN-1)) == -1) {
						break;
					}
					buff[ret] = '\0';
					
					optimize_fcall_to_string(buff, ret, &prev, &op, 1, cb, Ts);
				} else if (op->extended_value == 1) {
					optimize_fullpath(prev, NULL, 0 TSRMLS_CC);
				}
 				break;
#endif
 			case FN_FILE_EXISTS: /* file_exists() */
 				optimize_file_op(FS_EXISTS, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_IS_LINK: /* is_link() */
 				optimize_file_op(FS_IS_LINK, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_IS_DIR: /* is_dir() */
 				optimize_file_op(FS_IS_DIR, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_IS_FILE: /* is_file() */
 				optimize_file_op(FS_IS_FILE, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_IS_EXECUTABLE: /* is_executable() */
 				optimize_file_op(FS_IS_X, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_IS_READABLE: /* is_readable() */
 				optimize_file_op(FS_IS_R, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_IS_WRITEABLE: /* is_writeable() */
 				optimize_file_op(FS_IS_W, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILEATIME: /* fileatime() */
 				optimize_file_op(FS_ATIME, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILECTIME: /* filectime() */
 				optimize_file_op(FS_CTIME, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILEGROUP: /* filegroup() */
 				optimize_file_op(FS_GROUP, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILEINODE: /* fileinode() */
 				optimize_file_op(FS_INODE, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILEMTIME: /* filemtime() */
 				optimize_file_op(FS_MTIME, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILEOWNER: /* fileowner() */
 				optimize_file_op(FS_OWNER, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILEPERMS: /* fileperms() */
 				optimize_file_op(FS_PERMS, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILESIZE: /* filesize() */
 				optimize_file_op(FS_SIZE, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_FILETYPE: /* filetype() */
 				optimize_file_op(FS_TYPE, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_STRNCMP: /* strncmp() */
 				if (op->extended_value == 3) {
 					zend_op* arg3 = prev;
 					zend_op* arg2 = PREV_OP(arg3);
 					zend_op* arg1 = PREV_OP(arg2);
 					
					convert_to_long(&__OP1_VAL(arg3));
					convert_to_string(&__OP1_VAL(arg2));
					convert_to_string(&__OP1_VAL(arg1));

					if (OP1_VAL_L(prev) > 0) {
						long val = zend_binary_zval_strncmp(&__OP1_VAL(arg1), &__OP1_VAL(arg2), &__OP1_VAL(arg3));
						optimize_fcall_to_long(val, &prev, &op, 3, cb, Ts);
					} else {
						optimize_fcall_to_false(&prev, &op, 3, cb, Ts);
					}
					zval_dtor(&__OP1_VAL(arg2));
					SET_TO_NOP(arg2);
					zval_dtor(&__OP1_VAL(arg1));
					SET_TO_NOP(arg1);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_STRCASECMP: /* strcasecmp() */
 				if (op->extended_value == 2) {
 					zend_op* arg2 = prev;
 					zend_op* arg1 = PREV_OP(prev);
 					
					convert_to_string(&__OP1_VAL(arg2));
					convert_to_string(&__OP1_VAL(arg1));

					optimize_fcall_to_long(zend_binary_zval_strcasecmp(&__OP1_VAL(arg1), &__OP1_VAL(arg2)), &prev, &op, 2, cb, Ts);

					zval_dtor(&__OP1_VAL(arg1));
					SET_TO_NOP(arg1);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_STRNCASECMP: /* strncasecmp() */
 				if (op->extended_value == 3) {
 					zend_op* arg3 = prev;
 					zend_op* arg2 = PREV_OP(arg3);
 					zend_op* arg1 = PREV_OP(arg2);
 				
					convert_to_long(&__OP1_VAL(arg3));
					convert_to_string(&__OP1_VAL(arg2));
					convert_to_string(&__OP1_VAL(arg1));

					if (OP1_VAL_L(prev) > 0) {
						long val = zend_binary_zval_strncasecmp(&__OP1_VAL(arg1), &__OP1_VAL(arg2), &__OP1_VAL(arg3));
						optimize_fcall_to_long(val, &prev, &op, 3, cb, Ts);
					} else {
						optimize_fcall_to_false(&prev, &op, 3, cb, Ts);
					}
					zval_dtor(&__OP1_VAL(arg2));
					SET_TO_NOP(arg2);
					zval_dtor(&__OP1_VAL(arg1));
					SET_TO_NOP(arg1);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_VERSION_COMPARE: /* version_compare() */
 				if (op->extended_value == 3 || op->extended_value == 2) {
					zend_op* arg3 = prev;
					zend_op* arg2 = prev;
					zend_op* arg1 = prev;
					
					if (op->extended_value == 3) {
						arg2 = PREV_OP(arg3);
						convert_to_string(&__OP1_VAL(arg3));
					}
					arg1 = PREV_OP(arg2);
					
					convert_to_string(&__OP1_VAL(arg2));
					convert_to_string(&__OP1_VAL(arg1));
					
					if (op->extended_value == 2) {
						optimize_fcall_to_long(php_version_compare(OP1_VAL_S(arg1), OP1_VAL_S(arg2)), &prev, &op, 2, cb, Ts);
					} else { /* 3 */
						int i;
						char *s = OP1_VAL_S(arg3);
						int s_len = OP1_VAL_SL(arg3);

						i = php_version_compare(OP1_VAL_S(arg1), OP1_VAL_S(arg2));
						if (!strncmp(s, "<", s_len) || !strncmp(s, "lt", s_len)) {
							OPTIMIZE_FCALL_TO_BOOL(i == -1, 3);
						} else if (!strncmp(s, "<=", s_len) || !strncmp(s, "le", s_len)) {
							OPTIMIZE_FCALL_TO_BOOL(i != 1, 3);
						} else if (!strncmp(s, ">", s_len) || !strncmp(s, "gt", s_len)) {
							OPTIMIZE_FCALL_TO_BOOL(i == 1, 3);
						} else if (!strncmp(s, ">=", s_len) || !strncmp(s, "ge", s_len)) {
							OPTIMIZE_FCALL_TO_BOOL(i != -1, 3);
						} else if (!strncmp(s, "==", s_len) || !strncmp(s, "=", s_len) || !strncmp(s, "eq", s_len)) {
							OPTIMIZE_FCALL_TO_BOOL(i == 0, 3);
						} else if (!strncmp(s, "!=", s_len) || !strncmp(s, "<>", s_len) || !strncmp(s, "ne", s_len)) {
							OPTIMIZE_FCALL_TO_BOOL(i != 0, 3);
						} else { /* NULL */
							optimize_fcall_to_null(&prev, &op, op->extended_value, cb, Ts);
						}
					}
					
					if (op->extended_value == 3) {
						zval_dtor(&__OP1_VAL(arg2));
						SET_TO_NOP(arg2);
					}
					zval_dtor(&__OP1_VAL(arg1));
					SET_TO_NOP(arg1);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_DECHEX: /* dechex() */
 				if (op->extended_value == 1) {
					char *result;
					convert_to_long(&__OP1_VAL(prev));

					result = _php_math_longtobase(&__OP1_VAL(prev), 16);
					optimize_fcall_to_string(result, strlen(result), &prev, &op, 1, cb, Ts);
					efree(result);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_HEXDEC: /* hexdec() */
 				if (op->extended_value == 1) {
					zval ret;
					convert_to_string(&__OP1_VAL(prev));

					if (_php_math_basetozval(&__OP1_VAL(prev), 16, &ret) == SUCCESS) {
						if (Z_TYPE(ret) == IS_LONG) {
							optimize_fcall_to_long(Z_LVAL(ret), &prev, &op, 1, cb, Ts);
						} else { /* IS_DOUBLE */
							optimize_fcall_to_double(Z_DVAL(ret), &prev, &op, 1, cb, Ts);
						}
					} else {
						optimize_fcall_to_false(&prev, &op, 1, cb, Ts);
					}
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_DECBIN: /* decbin() */
 				if (op->extended_value == 1) {
					char *result;
					convert_to_long(&__OP1_VAL(prev));

					result = _php_math_longtobase(&__OP1_VAL(prev), 2);
					optimize_fcall_to_string(result, strlen(result), &prev, &op, 1, cb, Ts);
					efree(result);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_DECOCT: /* decoct() */
 				if (op->extended_value == 1) {
					char *result;
					convert_to_long(&__OP1_VAL(prev));

					result = _php_math_longtobase(&__OP1_VAL(prev), 8);
					optimize_fcall_to_string(result, strlen(result), &prev, &op, 1, cb, Ts);
					efree(result);
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_OCTDEC: /* octdec() */
 				if (op->extended_value == 1) {
					zval ret;
					convert_to_string(&__OP1_VAL(prev));

					if (_php_math_basetozval(&__OP1_VAL(prev), 8, &ret) == SUCCESS) {
						if (Z_TYPE(ret) == IS_LONG) {
							optimize_fcall_to_long(Z_LVAL(ret), &prev, &op, 1, cb, Ts);
						} else { /* IS_DOUBLE */
							optimize_fcall_to_double(Z_DVAL(ret), &prev, &op, 1, cb, Ts);
						}
					} else {
						optimize_fcall_to_false(&prev, &op, 1, cb, Ts);
					}
					
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_STR_REPEAT: /* str_repeat() */
 				if (op->extended_value == 2) {
					optimize_str_repeat(&prev, &op, cb, Ts TSRMLS_CC);
					OPTIMIZER_STATS_ADD_FCR();
				}
 				break;
 			case FN_INTVAL: /* intval() */
 				do {
					long base = 10;
					long res;
					zend_op *x = prev;
					if (op->extended_value == 2) {
						convert_to_long(&__OP1_VAL(prev));
						base = OP1_VAL_L(prev);
						x = prev - 1;
					}
					/* one arg or 2 args (x != prev) */
					if (op->extended_value == 1 || x != prev) {
						convert_to_long_base(&__OP1_VAL(x), base);
						res = OP1_VAL_L(x);
						optimize_fcall_to_long(res, &prev, &op, op->extended_value, cb, Ts);

						if (x != prev) {
							zval_dtor(&__OP1_VAL(op - 2));
							SET_TO_NOP(op - 2);
						}
					}
				} while (0);
 				break;
 			case FN_SUBSTR: /* substr */
 				if (op->extended_value == 2 || op->extended_value == 3) {
 					optimize_substr(&prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				}
 				break;
 			case FN_ACOS:  /* acos() */
 			case FN_ACOSH: /* acosh() */
 			case FN_ASIN:  /* asin() */
 			case FN_ASINH: /* asinh() */
 			case FN_ATAN:  /* atan() */
 			case FN_ATANH: /* atanh() */
 			case FN_COS:   /* cos() */
 			case FN_COSH:  /* cosh() */
 			case FN_SIN:   /* sin() */
 			case FN_SINH:  /* sinh() */
 			case FN_TAN:   /* tan() */
 			case FN_TANH:  /* tanh() */
 			case FN_EXP:   /* exp() */
 			case FN_LOG10: /* log10() */
 			case FN_SQRT:  /* sqrt() */
 				if (op->extended_value == 1) {
 					double d;
 					convert_to_double(&__OP1_VAL(prev));
 					d = OP1_VAL_D(prev);
 					
 					switch (fe->fid) {
 						case FN_ACOS:  d = acos(d);  break;
 						case FN_ACOSH: d = optimizer_acosh(d); break;
 						case FN_ASIN:  d = asin(d);  break;
 						case FN_ASINH: d = optimizer_asinh(d); break;
 						case FN_ATAN:  d = atan(d);  break;
 						case FN_ATANH: d = optimizer_atanh(d); break;
 						case FN_COS:   d = cos(d);   break;
 						case FN_COSH:  d = cosh(d);  break;
 						case FN_SIN:   d = sin(d);   break;
 						case FN_SINH:  d = sinh(d);  break;
 						case FN_TAN:   d = tan(d);   break;
 						case FN_TANH:  d = tanh(d);  break;
 						case FN_EXP:   d = exp(d);   break;
 						case FN_LOG10: d = log10(d); break;
 						case FN_SQRT:  d = sqrt(d);  break;
 					}
 					
 					optimize_fcall_to_double(d, &prev, &op, 1, cb, Ts);
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_ATAN2: /* atan2() */
 				if (op->extended_value == 2) {
 					zend_op* arg2 = prev;
 					zend_op* arg1 = PREV_OP(arg2);
 					
 					convert_to_double(&__OP1_VAL(arg2));
 					convert_to_double(&__OP1_VAL(arg1));
 					optimize_fcall_to_double(atan2(OP1_VAL_D(arg1), OP1_VAL_D(arg2)), &prev, &op, 2, cb, Ts);
 					SET_TO_NOP(arg1);
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_CEIL:  /* ceil() */
 			case FN_FLOOR: /* floor() */
 				if (op->extended_value == 1) {
 					convert_scalar_to_number(&__OP1_VAL(prev) TSRMLS_CC);
 					
 					if (OP1_VALT(prev) == IS_DOUBLE) {
 						if (fe->fid == FN_CEIL) {
 							optimize_fcall_to_double(ceil(OP1_VAL_D(prev)), &prev, &op, 1, cb, Ts);
 						} else {
 							optimize_fcall_to_double(floor(OP1_VAL_D(prev)), &prev, &op, 1, cb, Ts);
 						}
 					} else if (OP1_VALT(prev) == IS_LONG) {
 						convert_to_double(&__OP1_VAL(prev));
 						optimize_fcall_to_double(OP1_VAL_D(prev), &prev, &op, 1, cb, Ts);
 					} else {
 						optimize_fcall_to_false(&prev, &op, 1, cb, Ts);
 					}
 					
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_FMOD: /* fmod() */
 				if (op->extended_value == 2) {
 					zend_op* arg2 = prev; 
 					zend_op* arg1 = PREV_OP(arg2);
 				
 					convert_to_double(&__OP1_VAL(arg2));
 					convert_to_double(&__OP1_VAL(arg1));
 					
 					optimize_fcall_to_double(fmod(OP1_VAL_D(arg1), OP1_VAL_D(arg2)), &prev, &op, 2, cb, Ts);
 					OPTIMIZER_STATS_ADD_FCR();
 					SET_TO_NOP(arg1);
 				}
 				break;
 			case FN_INI_GET: /* ini_get() */
 				if (op->extended_value == 1) {
 					zend_ini_entry* ini_entry;
 					convert_to_string(&__OP1_VAL(prev));
 					
 					if (zend_hash_find(EG(ini_directives), OP1_VAL_S(prev), OP1_VAL_SL(prev)+1, (void**)&ini_entry) == SUCCESS) {
 						if (ini_entry->modifiable == ZEND_INI_SYSTEM && ini_entry->value && ini_entry->value_length > 0) {
 							optimize_fcall_to_string(ini_entry->value, ini_entry->value_length, &prev, &op, 1, cb, Ts);
 							OPTIMIZER_STATS_ADD_FCR();
 						}
 					}
 				}
 				break;
 			case FN_IP2LONG: /* ip2long() */
 				if (op->extended_value == 1) {
 					unsigned long int ip;
 					convert_to_string(&__OP1_VAL(prev));
 					
 					if (OP1_VAL_SL(prev) == 0 || (ip = inet_addr(OP1_VAL_S(prev))) == INADDR_NONE) {
 						if (OP1_VAL_SL(prev) == sizeof("255.255.255.255")-1 && !memcmp(OP1_VAL_S(prev), "255.255.255.255", sizeof("255.255.255.255")-1)) {
 							optimize_fcall_to_long(0xFFFFFFFF, &prev, &op, 1, cb, Ts);
 						} else {
 							optimize_fcall_to_false(&prev, &op, 1, cb, Ts);
 						}
 					} else {
 						optimize_fcall_to_long(ntohl(ip), &prev, &op, 1, cb, Ts);
 					}
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_LONG2IP: /* long2ip() */
 				if (op->extended_value == 1) {
 					unsigned long n;
 					char* ip;
 					struct in_addr myaddr;
 					convert_to_string(&__OP1_VAL(prev));
 					
 					n = strtoul(OP1_VAL_S(prev), NULL, 0);
 					myaddr.s_addr = htonl(n);
 					ip = inet_ntoa(myaddr);
 					optimize_fcall_to_string(ip, strlen(ip), &prev, &op, 1, cb, Ts);
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_TRIM: /* trim() */
 				optimize_trim(3, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_CHOP:  /* chop()  */
 			case FN_RTRIM: /* rtrim() */
 				optimize_trim(2, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_LTRIM: /* ltrim() */
 				optimize_trim(1, &prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
 				break;
 			case FN_RAD2DEG: /* rad2deg() */
 				if (op->extended_value == 1) {
 					convert_to_double(&__OP1_VAL(prev));
 					optimize_fcall_to_double(((OP1_VAL_D(prev)/M_PI)*180.0), &prev, &op, 1, cb, Ts);
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_DEG2RAD: /* deg2rad() */
 				if (op->extended_value == 1) {
 					convert_to_double(&__OP1_VAL(prev));
 					optimize_fcall_to_double(((OP1_VAL_D(prev)/180.0)*M_PI), &prev, &op, 1, cb, Ts);
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_ABS: /* abs() */
 				if (op->extended_value == 1) {
 					convert_scalar_to_number(&__OP1_VAL(prev) TSRMLS_CC);
 					
 					if (OP1_VALT(prev) == IS_DOUBLE) {
 						optimize_fcall_to_double(fabs(OP1_VAL_D(prev)), &prev, &op, 1, cb, Ts);
 					} else if (OP1_VALT(prev) == IS_LONG) {
 						if (OP1_VAL_L(prev) == LONG_MIN) {
 							optimize_fcall_to_double((-(double)LONG_MIN), &prev, &op, 1, cb, Ts);
 						} else {
 							optimize_fcall_to_long((OP1_VAL_L(prev) < 0 ? -OP1_VAL_L(prev) : OP1_VAL_L(prev)), &prev, &op, 1, cb, Ts);
 						}
 					} else {
 						optimize_fcall_to_false(&prev, &op, 1, cb, Ts);
 					}
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
 			case FN_GET_CFG_VAR:
 				if (op->extended_value == 1) {
 					char* value = NULL;
 					convert_to_string(&__OP1_VAL(prev));
 					
 					if (cfg_get_string(OP1_VAL_S(prev), &value) == FAILURE) {
 						optimize_fcall_to_false(&prev, &op, 1, cb, Ts);
 					} else {
 						optimize_fcall_to_string(value, strlen(value), &prev, &op, 1, cb, Ts);
 					}
 					OPTIMIZER_STATS_ADD_FCR();
 				}
 				break;
  		}
	}
	
#if OPTIMIZER_STATS_COMPILE && OPTIMIZER_DEBUG_COMPILE
	efree(stats_fname);
#endif
}


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
