/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2008 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Graham Kelly <graham@php.net>                               |
  +----------------------------------------------------------------------+
*/
/* $Id: optimizer_debug.c,v 1.13 2008/08/07 05:23:04 graham Exp $ */


#include "php_optimizer.h"
#include "optimize.h"
#include "optimizer_debug.h"

#ifndef PHP_WIN32
#if OPTIMIZER_DEBUG_COMPILE
#include <dlfcn.h>
typedef void (*vld_dump_oparray_f) (zend_op_array* TSRMLS_DC);
typedef void vld_dump_op_f;
#endif
#endif


/*
 * Dump an op array
 * 
 */
#if OPTIMIZER_DEBUG_COMPILE
void _optimizer_dump_op_array(zend_op_array* op_array, char* file, int line TSRMLS_DC) 
{
#ifndef PHP_WIN32
	if (OPTIMIZER_G(optimizer_verbose_debug)) {
		vld_dump_oparray_f dump_f = dlsym(NULL, "vld_dump_oparray");
		
		printf("VLD OP-ARRAY DUMP [%s:%d]\n", file, line);
		
		if (dump_f) {
			/* USE the Vulcan Logic Dissasembler (VLD) to do the printing! */
			vld_dump_oparray(op_array TSRMLS_CC);
		} else {
			printf("VLD is not installed.\n");
		}
	}
#endif
}
#endif


#if OPTIMIZER_DEBUG_COMPILE
void _optimizer_dump_op(zend_op* op, char* file, int line TSRMLS_DC)
{
	if (OPTIMIZER_G(optimizer_verbose_debug)) {
		printf("VLD OP DUMP [%s:%d]\n", file, line);
		vld_dump_op(0, op, 1, 1 TSRMLS_CC);
	}
}
#endif


/*
 * Dump the contents of the control flow graph
 *  
 */ 
void optimizer_dump_cfg(CB* cb, zend_op_array* op_array TSRMLS_DC) 
{
	if (OPTIMIZER_G(optimizer_verbose_debug)) {
		php_printf("Optimizer: Control Flow Graph DUMP\n");
		php_printf("function: %s()\n", (op_array->function_name == NULL ? "main" : op_array->function_name));
		php_printf("------------------------------------------------------------\n"); /* 60 */
		
		CB* p = cb;
		while (p != NULL) {
			php_printf("zend_op* start=%p\n", p->start);
			p = p->next;
		}
	}
}



/*
 * Outputs the Control Flow Graph in DOT format. This format can then be run 
 *   through Graphviz to create a visual representation of the graph.
 *
 *   http://www.graphviz.org (This worked using version 2.16.1)
 *
 * TODO: At the moment this outputs parts to both stdout and stderr (VLD)
 *         I would Line to make it output to the same place... also, the HTML 
 *         encounteres an error on jump lines dure to the '>' character being
 *         used by VLD... this means that basically any script with jmp lines 
 *         needs to have -> converted to -&gt; by hand... this should just be 
 *         a simple text substituation though.
 * TODO: Add support for pchains
 * TODO: Fix the lines when rendering TB (cosmetic change only)
 */
void optimizer_dump_cfg_graphviz(CB* cb, zend_op_array* op_array TSRMLS_DC)
{
#if OPTIMIZER_DEBUG_COMPILE
	php_printf("/*\n");
	php_printf(" * APC Optimizer Control Flow Graph\n");
	php_printf(" *\n");
	php_printf(" * op_array: %s()\n", (op_array->function_name == NULL ? "main" : op_array->function_name));
	php_printf(" * file: %s\n", op_array->filename);
	php_printf(" */\n\n");
	
	php_printf("digraph cgf {\n");
	php_printf("	size=\"100,100\";\n");
	
	php_printf("	graph [\n");
	php_printf("		rankdir=\"LR\"\n");
	php_printf("		bgcolor=\"#f7f7f7\"\n");
	php_printf("		label=\"Control Flow Graph for %s() in %s\"\n", (op_array->function_name == NULL ? "main" : op_array->function_name), op_array->filename);
	php_printf("		labeljust=\"l\"");
	php_printf("	];\n");
	php_printf("	node [\n");
	php_printf("		fontname=\"Helvetica\"\n");
	php_printf("		fontsize=\"11\"");
	php_printf("		shape=\"plaintext\"");
	php_printf("	];\n");
	php_printf("	edge []\n");
	
	CB* p = cb;
	int n = 0;
	int opline = 0;
	
	while (p != NULL) {
		int i;
		php_printf("	\"node%p\" [\n", p);
		php_printf("		label=<\n");
	
		php_printf("			<TABLE BORDER=\"2\" CELLBORDER=\"1\" CELLSPACING=\"0\" BGCOLOR=\"#ffffff\">\n");
		php_printf("			<TR><TD PORT=\"h0\" BGCOLOR=\"#9bab96\" ALIGN=\"CENTER\">CB %d (%p)</TD></TR>\n", n, p);
		php_printf("			<TR><TD PORT=\"h1\" BGCOLOR=\"#bed1b8\" ALIGN=\"LEFT\">used: %s</TD></TR>\n", (p->used ? "true" : "false"));
		php_printf("			<TR><TD PORT=\"h2\" BGCOLOR=\"#bed1b8\" ALIGN=\"LEFT\">exn: %s</TD></TR>\n", (p->exn ? "true" : "false"));
		php_printf("			<TR><TD PORT=\"jmp_1\" BGCOLOR=\"#f4f7da\" ALIGN=\"LEFT\">jmp_1: %p</TD></TR>\n", p->jmp_1);
		php_printf("			<TR><TD PORT=\"jmp_2\" BGCOLOR=\"#f4f7da\" ALIGN=\"LEFT\">jmp_2: %p</TD></TR>\n", p->jmp_2);
		php_printf("			<TR><TD PORT=\"jmp_ext\" BGCOLOR=\"#f4f7da\" ALIGN=\"LEFT\">jmp_ext: %p</TD></TR>\n", p->jmp_ext);
		php_printf("			<TR><TD PORT=\"follow\" BGCOLOR=\"#f4f7da\" ALIGN=\"LEFT\">follow: %p</TD></TR>\n", p->follow);
		php_printf("			<TR><TD PORT=\"next\" BGCOLOR=\"#f4f7da\" ALIGN=\"LEFT\">next: %p</TD></TR>\n", p->next);
		php_printf("			<TR><TD PORT=\"pchain\" BGCOLOR=\"#f4f7da\" ALIGN=\"LEFT\">pchain: %p</TD></TR>\n", p->pchain);
	
	
		zend_uint base_address = (zend_uint)((intptr_t)&(p->start[0]));
		for (i = 0; i < p->len; i++, opline++) {
			php_printf("			<TR><TD PORT=\"o%d\" ALIGN=\"LEFT\">", i);
			vld_dump_op(i, p->start, base_address, 1 TSRMLS_CC);
			php_printf("</TD></TR>\n");
		}
	
		php_printf("</TABLE>>\n");
		php_printf("	];\n");
	
	
		if (p->jmp_1 != NULL) {
			php_printf("	\"node%p\":jmp_1->\"node%p\":o0;\n", p, p->jmp_1);
		}
	
		if (p->jmp_2 != NULL) {
			php_printf("	\"node%p\":jmp_2->\"node%p\":o0;\n", p, p->jmp_2);
		}
	
		if (p->jmp_ext != NULL) {
			php_printf("	\"node%p\":jmp_ext->\"node%p\":o0;\n", p, p->jmp_ext);
		}
	
		if (p->follow != NULL) {
			php_printf("	\"node%p\":follow->\"node%p\":o0;\n", p, p->follow);
		}
	
		if (p->next != NULL) {
			php_printf("	\"node%p\":next->\"node%p\":o0;\n", p, p->next);
		}
		
		p = p->next;
		n++;
	}
	
	php_printf("\n}\n");
#endif
}



/*
 * Extended information stats add function (does optimization logging)
 *
 */
void optimizer_stats_add_ex(optimizer_stats* stats, zend_op* op, char* log_message TSRMLS_DC) 
{
	/* For now... */
	php_printf("%s\n", log_message);
}


/*
 * Dumps the statistics information for an op-array.
 *
 */
void optimizer_dump_op_array_stats(optimizer_stats* stats, zend_op_array* op_array TSRMLS_DC) 
{
#if OPTIMIZER_DEBUG_COMPILE
	if (OPTIMIZER_G(optimizer_verbose_debug)) {
		static const char* inc = "increase";
	    static const char* dec = "decrease";
	    static const char* nc  = "no change";
		double op_row_percent = 0, vars_percent = 0, temp_percent = 0;
		
#define STATS_PERCENT_CHANGE(pre, post) ((((double)post-(double)pre)/(pre == 0 ? 1 : (double)pre))*100)
#define ABS(n) (n < 0 ? -n : n)


		op_row_percent = STATS_PERCENT_CHANGE(stats->pre.op_row_count, stats->post.op_row_count);
		vars_percent   = STATS_PERCENT_CHANGE(stats->pre.var_count   , stats->post.var_count   );
		temp_percent   = STATS_PERCENT_CHANGE(stats->pre.temp_count  , stats->post.temp_count  );
		
		php_printf("APC Optimizer Statistics for %s() in file %s\n", (op_array->function_name == NULL ? "main" : op_array->function_name), op_array->filename);
		php_printf("----------------------------------------------------------------------\n");
		php_printf("Optimization Count: %d\n", stats->optimization_count);
		
		php_printf("Pre Optimization Stats: \n");
		php_printf("--------------------------------------\n");
		php_printf("Op-rows: %d\n", stats->pre.op_row_count);
		php_printf("Vars: %d\n", stats->pre.var_count);
		php_printf("Temp vars: %d\n", stats->pre.temp_count);
		
		php_printf("Post Optimization Stats: \n");
		php_printf("--------------------------------------\n");
		php_printf("Op-rows: %d (%4.2f%% %s)\n"    , stats->post.op_row_count, ABS(op_row_percent), (op_row_percent < 0 ? dec : (op_row_percent == 0 ? nc : inc)));
		php_printf("Vars: %d (%4.2f%% %s)\n"       , stats->post.var_count   , ABS(vars_percent)  , (vars_percent < 0   ? dec : (vars_percent == 0   ? nc : inc)));
		php_printf("Temp vars: %d (%4.2f%% %s)\n\n", stats->post.temp_count  , ABS(temp_percent)  , (temp_percent < 0   ? dec : (temp_percent == 0   ? nc : inc)));
	}
#endif
}


#define O_FLAG_OPLINE  1
#define O_FLAG_OPNUM   2
#define O_FLAG_CLASS   4


#if OPTIMIZER_TEST_COMPILE
static void optimizer_test_dump_znode(znode* node, zend_uint base_address, zend_bool flags TSRMLS_DC)
{
	printf("%04d ", node->op_type);
	
	if (flags & O_FLAG_OPLINE) {
		printf("%010ld ", ((node->u.opline_num - base_address) / sizeof(zend_op)));
	} else if (flags & O_FLAG_OPNUM) {
		printf("%010d ", node->u.opline_num);
	} else if (flags & O_FLAG_CLASS) {
		printf("%010ld ", node->u.var / sizeof(temp_variable));
	} else {
		switch (node->op_type) {
			case IS_CONST:
				printf("%010d ", node->u.constant.type);
				break;
			case IS_VAR:
			case IS_TMP_VAR:
				printf("%010ld ", node->u.var / sizeof(temp_variable));
				break;
			case IS_CV:
				printf("%010d ", node->u.var);
				break;
			case IS_UNUSED:
				printf("---------- ");
				break;
		}
	}
}
#endif


/*
 * Dumps an op-array in a format that is used in the phpt files
 *
 * This format was never intended to be like VLD... I know it would be nice to 
 *   be able to view test values text like in VLD... HOWEVER, I wanted a concise
 *   format that wasnt too large for this, thus a numeric dump. All values in 
 *   zvals are ignored in this dump... this is to focus only on the opcodes. 
 *
 *
 * op-row format:
 *  op# opcode ext fetch res-type res-var/const op1-type op1-var/const op2-type op2-var/const
 */
void optimizer_test_dump_op_array(zend_op_array* op_array TSRMLS_DC)
{
#if OPTIMIZER_TEST_COMPILE
	zend_uint base_address = (zend_uint)((intptr_t)&(op_array->opcodes[0]));

	printf("%s\n", (op_array->function_name == NULL ? "main" : op_array->function_name));
	printf("%d\n%d\n", op_array->last, op_array->T);
	
	unsigned int i;
	for (i = 0; i < op_array->last; i++) {
		zend_op* op = op_array->opcodes+i;
		zend_bool op1_flags = 0;
		zend_bool op2_flags = 0;
		zend_bool res_flags = 0;
		zend_bool is_fetch  = 0;
		
		/* Opcode */
		printf("%04d %04d %08ld ", i, op->opcode, op->extended_value);

		switch (op->opcode) {
			case ZEND_JMP:
				op1_flags = O_FLAG_OPLINE;
				break;
			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
			case ZEND_JMPNZ_EX:
				op2_flags = O_FLAG_OPLINE;
				break;
			case ZEND_JMPZNZ:
				op2_flags = O_FLAG_OPNUM;
				break;
#ifdef ZEND_JMP_SET
			case ZEND_JMP_SET:
				op2_flags = O_FLAG_OPLINE;
				break;
#endif
			case ZEND_FETCH_R:
			case ZEND_FETCH_W:
			case ZEND_FETCH_RW:
			case ZEND_FETCH_FUNC_ARG:
				is_fetch = 1;
				break;
			case ZEND_FE_FETCH:
			case ZEND_FE_RESET:
				op2_flags = O_FLAG_OPNUM;
				break;
			case ZEND_FETCH_CLASS:
				res_flags = O_FLAG_CLASS;
				break;
			case ZEND_NEW:
				op1_flags = O_FLAG_CLASS;
				break;
		}
		
		if (is_fetch) {
			printf("%04d ", op->op2.u.EA.type);
		} else {
			printf("---- ");
		}

		/* RES */
		optimizer_test_dump_znode(&op->result, base_address, res_flags TSRMLS_CC);
		
		/* OP1 */
		optimizer_test_dump_znode(&op->op1, base_address, op1_flags TSRMLS_CC);
		
		/* OP2 */
		optimizer_test_dump_znode(&op->op2, base_address, op2_flags TSRMLS_CC);
		
		printf("\n");
	}
	
	fflush(stdout);
#endif
}


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
 
