/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2008 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Ilia Alshanetsky <iliaa@php.net>                            |
  |          Graham Kelly <graham@php.net>                               |
  |          Dmitry Stogov (Truck MMcache)                               |
  |          Derick Rethans <derick@php.net>                             |
  +----------------------------------------------------------------------+
*/
/*
Portions of the code below are based on the Turck MMCache by Dmitry Stogov

The components derived from Turck MMCache are licensed under the BSD license 
for the purposes of this optimizer. All other components of Turck MMCache 
retain their existing GPL license.
*/

/* $Id: optimize.c,v 1.18 2008/08/07 11:37:19 graham Exp $ */

#include "zend.h"
#include "zend_API.h"
#include "zend_constants.h"
#include "zend_compile.h"
#include "zend_extensions.h"
#include "zend_ini.h"

#include "SAPI.h" 
#include "php.h"

#include "ext/standard/basic_functions.h"
#include "ext/standard/pageinfo.h"
#include "ext/standard/info.h"
#include "ext/standard/md5.h"
#include "ext/standard/sha1.h"
#include "ext/standard/crc32.h"
#include "ext/standard/php_string.h"
#include "ext/standard/php_math.h"
#include "ext/standard/php_rand.h"
#include "ext/standard/html.h"
#include "ext/standard/php_versioning.h"

#include "php_optimizer.h"
#include "optimize.h"
#include "optimizer_debug.h"
#include "optimize_fcr.h"



unsigned int zops[] = {
/* ZEND_NOP*/ EXT_UNUSED | OP1_UNUSED | OP2_UNUSED | RES_UNUSED,
/* ZEND_ADD*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_SUB*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_MUL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_DIV*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_MOD*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_SL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_SR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_CONCAT*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_BW_OR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_BW_AND*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_BW_XOR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_BW_NOT*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_TMP,
/* ZEND_BOOL_NOT*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_BOOL_XOR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IS_IDENTICAL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IS_NOT_IDENTICAL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IS_EQUAL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IS_NOT_EQUAL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IS_SMALLER*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IS_SMALLER_OR_EQUAL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_CAST*/ EXT_CAST| OP1_STD| OP2_UNUSED | RES_TMP,
/* ZEND_QM_ASSIGN*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_TMP,
/* ZEND_ASSIGN_ADD*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_SUB*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_MUL*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_DIV*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_MOD*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_SL*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_SR*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_CONCAT*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_BW_OR*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_BW_AND*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_BW_XOR*/ EXT_ASSIGN | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_PRE_INC*/ EXT_UNUSED | OP1_VAR| OP2_UNUSED | RES_VAR,
/* ZEND_PRE_DEC*/ EXT_UNUSED | OP1_VAR| OP2_UNUSED | RES_VAR,
/* ZEND_POST_INC*/ EXT_UNUSED | OP1_VAR| OP2_UNUSED | RES_TMP,
/* ZEND_POST_DEC*/ EXT_UNUSED | OP1_VAR| OP2_UNUSED | RES_TMP,
/* ZEND_ASSIGN*/ EXT_UNUSED | OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_REF*/ EXT_UNUSED | OP1_VAR| OP2_VAR| RES_VAR,
/* ZEND_ECHO*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_PRINT*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_TMP,
/* ZEND_JMP*/ EXT_UNUSED | OP1_JMPADDR| OP2_UNUSED | RES_UNUSED,
/* ZEND_JMPZ*/ EXT_UNUSED | OP1_STD| OP2_JMPADDR| RES_UNUSED,
/* ZEND_JMPNZ*/ EXT_UNUSED | OP1_STD| OP2_JMPADDR| RES_UNUSED,
/* ZEND_JMPZNZ*/ EXT_OPLINE | OP1_STD| OP2_OPLINE | RES_UNUSED,
/* ZEND_JMPZ_EX*/ EXT_UNUSED | OP1_STD| OP2_JMPADDR| RES_TMP,
/* ZEND_JMPNZ_EX*/ EXT_UNUSED | OP1_STD| OP2_JMPADDR| RES_TMP,
/* ZEND_CASE*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_SWITCH_FREE*/ EXT_BIT| OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_BRK*/ EXT_UNUSED | OP1_BRK| OP2_STD| RES_UNUSED,
/* ZEND_CONT*/ EXT_UNUSED | OP1_CONT| OP2_STD| RES_UNUSED,
/* ZEND_BOOL*/ EXT_UNUSED | OPS_STD| OP2_UNUSED | RES_TMP,
/* ZEND_INIT_STRING*/ EXT_UNUSED | OP1_UNUSED | OP2_UNUSED | RES_TMP,
/* ZEND_ADD_CHAR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_ADD_STRING*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_ADD_VAR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_BEGIN_SILENCE*/ EXT_UNUSED | OP1_UNUSED | OP2_UNUSED | RES_TMP,
/* ZEND_END_SILENCE*/ EXT_UNUSED | OP1_TMP| OP2_UNUSED | RES_UNUSED,
/* ZEND_INIT_FCALL_BY_NAME*/ EXT_INIT_FCALL | OP1_STD| OP2_STD| RES_UNUSED,
/* ZEND_DO_FCALL*/ EXT_FCALL | OP1_STD| OP2_OPLINE | RES_VAR,
/* ZEND_DO_FCALL_BY_NAME*/ EXT_FCALL | OP1_STD| OP2_OPLINE | RES_VAR,
/* ZEND_RETURN*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_RECV*/ EXT_UNUSED | OP1_ARG| OP2_UNUSED | RES_VAR,
/* ZEND_RECV_INIT*/ EXT_UNUSED | OP1_ARG| OP2_STD| RES_VAR,
/* ZEND_SEND_VAL*/ EXT_SEND| OP1_STD| OP2_ARG| RES_UNUSED,
/* ZEND_SEND_VAR*/ EXT_SEND| OP1_VAR| OP2_ARG| RES_UNUSED,
/* ZEND_SEND_REF*/ EXT_SEND| OP1_VAR| OP2_ARG| RES_UNUSED,
/* ZEND_NEW*/ EXT_UNUSED | OP1_CLASS | OP2_UNUSED | RES_VAR,
/* ZEND_JMP_NO_CTOR*/ EXT_UNUSED | OP1_STD| OP2_OPLINE | RES_UNUSED,
/* ZEND_FREE*/ EXT_UNUSED | OP1_TMP| OP2_UNUSED | RES_UNUSED,
/* ZEND_INIT_ARRAY*/ EXT_BIT| OP1_STD| OP2_STD| RES_TMP,
/* ZEND_ADD_ARRAY_ELEMENT*/ EXT_BIT| OP1_STD| OP2_STD| RES_TMP,
/* ZEND_INCLUDE_OR_EVAL*/ EXT_UNUSED | OP1_STD| OP2_INCLUDE| RES_VAR,
/* ZEND_UNSET_VAR*/ EXT_UNUSED | OP1_STD| OP2_FETCH | RES_UNUSED,
/* ZEND_UNSET_DIM*/ EXT_STD| OP1_STD| OP2_STD| RES_UNUSED,
/* ZEND_UNSET_OBJ*/ EXT_STD| OP1_STD| OP2_STD| RES_UNUSED,
/* ZEND_FE_RESET*/ EXT_BIT| OP1_STD| OP2_OPLINE | RES_VAR,
/* ZEND_FE_FETCH*/ EXT_FE | OP1_STD| OP2_OPLINE | RES_TMP,
/* ZEND_EXIT*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_FETCH_R*/ EXT_UNUSED | OP1_STD| OP2_FETCH | RES_VAR,
/* ZEND_FETCH_DIM_R*/ EXT_FETCH | OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_FETCH_OBJ_R*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_W*/ EXT_UNUSED | OP1_STD| OP2_FETCH | RES_VAR,
/* ZEND_FETCH_DIM_W*/ EXT_UNUSED | OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_FETCH_OBJ_W*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_RW*/ EXT_UNUSED | OP1_STD| OP2_FETCH | RES_VAR,
/* ZEND_FETCH_DIM_RW*/ EXT_UNUSED | OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_FETCH_OBJ_RW*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_IS*/ EXT_UNUSED | OP1_STD| OP2_FETCH | RES_VAR,
/* ZEND_FETCH_DIM_IS*/ EXT_UNUSED | OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_FETCH_OBJ_IS*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_FUNC_ARG*/ EXT_ARG| OP1_STD| OP2_FETCH | RES_VAR,
/* ZEND_FETCH_DIM_FUNC_ARG*/ EXT_ARG| OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_FETCH_OBJ_FUNC_ARG*/ EXT_ARG| OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_UNSET*/ EXT_UNUSED | OP1_STD| OP2_FETCH | RES_VAR,
/* ZEND_FETCH_DIM_UNSET*/ EXT_UNUSED | OP1_VAR| OP2_STD| RES_VAR,
/* ZEND_FETCH_OBJ_UNSET*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_DIM_TMP_VAR*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_FETCH_CONSTANT*/ EXT_UNUSED | OP1_UCLASS | OP2_STD| RES_TMP,
/* ZEND_DECLARE_FUNCTION_OR_CLASS*/ EXT_DECLARE| OP1_STD| OP2_STD| RES_UNUSED,
/* ZEND_EXT_STMT*/ EXT_STD| OP1_STD| OP2_STD| RES_STD,
/* ZEND_EXT_FCALL_BEGIN*/ EXT_STD| OP1_STD| OP2_STD| RES_STD,
/* ZEND_EXT_FCALL_END*/ EXT_STD| OP1_STD| OP2_STD| RES_STD,
/* ZEND_EXT_NOP*/ EXT_UNUSED | OP1_UNUSED | OP2_UNUSED | RES_UNUSED,
/* ZEND_TICKS*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_SEND_VAR_NO_REF*/ EXT_SEND_NOREF| OP1_VAR | OP2_ARG| RES_UNUSED,
/* ZEND_CATCH*/ EXT_OPLINE | OP1_CLASS | OP2_STD| RES_UNUSED,
/* ZEND_THROW*/ EXT_UNUSED | OP1_STD| OP2_OPLINE | RES_UNUSED,
/* ZEND_FETCH_CLASS*/ EXT_FCLASS | OP1_STD| OP2_STD| RES_CLASS,
/* ZEND_CLONE*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_VAR,
/* ZEND_INIT_CTOR_CALL*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_INIT_METHOD_CALL*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_INIT_STATIC_METHOD_CALL*/ EXT_UNUSED | OP1_UCLASS | OP2_STD| RES_UNUSED,
/* ZEND_ISSET_ISEMPTY_VAR*/ EXT_ISSET | OP1_STD| OP2_FETCH | RES_TMP,
/* ZEND_ISSET_ISEMPTY_DIM_OBJ*/ EXT_ISSET | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_IMPORT_FUNCTION*/ EXT_UNUSED | OP1_CLASS | OP2_STD| RES_UNUSED,
/* ZEND_IMPORT_CLASS*/ EXT_UNUSED | OP1_CLASS | OP2_STD| RES_UNUSED,
/* ZEND_IMPORT_CONST*/ EXT_UNUSED | OP1_CLASS | OP2_STD| RES_UNUSED,
/* ZEND_OP_119*/ EXT_STD| OP1_STD| OP2_STD| RES_STD,
/* ZEND_OP_120*/ EXT_STD| OP1_STD| OP2_STD| RES_STD,
/* ZEND_ASSIGN_ADD_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_SUB_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_MUL_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_DIV_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_MOD_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_SL_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_SR_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_CONCAT_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_BW_OR_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_BW_AND_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ASSIGN_BW_XOR_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_PRE_INC_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_PRE_DEC_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_POST_INC_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_POST_DEC_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_ASSIGN_OBJ*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_OP_DATA*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_STD,
/* ZEND_INSTANCEOF*/ EXT_UNUSED | OP1_STD| OP2_CLASS | RES_TMP,
/* ZEND_DECLARE_CLASS*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_CLASS,
/* ZEND_DECLARE_INHERITED_CLASS*/ EXT_CLASS | OP1_STD| OP2_STD| RES_CLASS,
/* ZEND_DECLARE_FUNCTION*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_UNUSED,
/* ZEND_RAISE_ABSTRACT_ERROR*/ EXT_UNUSED | OP1_UNUSED | OP2_UNUSED | RES_UNUSED,
/* ZEND_START_NAMESPACE*/ EXT_UNUSED | OP1_STD| OP2_UNUSED | RES_UNUSED,
/* ZEND_ADD_INTERFACE*/ EXT_IFACE | OP1_CLASS | OP2_CLASS | RES_UNUSED,
/* ZEND_VERIFY_INSTANCEOF*/ EXT_UNUSED | OP1_CLASS | OP2_STD| RES_UNUSED,
/* ZEND_VERIFY_ABSTRACT_CLASS*/ EXT_UNUSED | OP1_CLASS | OP2_UNUSED | RES_UNUSED,
/* ZEND_ASSIGN_DIM*/ EXT_UNUSED | OP1_STD| OP2_STD| RES_VAR,
/* ZEND_ISSET_ISEMPTY_PROP_OBJ*/ EXT_ISSET | OP1_STD| OP2_STD| RES_TMP,
/* ZEND_HANDLE_EXCEPTION*/ EXT_STD| OP1_UNUSED | OP2_UNUSED | RES_STD, 
/* ZEND_USER_OPCODE*/ EXT_STD| OP1_UNUSED | OP2_UNUSED | RES_STD,
};



zend_op *optimize_next_op(zend_op *op)
{
	zend_op *x = op + 1;

	while (x && x->opcode == ZEND_NOP) {
		x++;
	}

	return x;
}


zend_op *optimize_prev_op(zend_op *op)
{
	zend_op* x = op-1;
	
	while (x && x->opcode == ZEND_NOP) {
		x--;
	}
	
	return x;
}


zend_uint optimizer_fetch_free_tmp_var(zend_op_array* op_array)
{
	op_array->T++;
	return op_array->T;
}

void optimizer_parse_level(TSRMLS_DC)
{
	zend_uint level = OPTIMIZER_G(optimization_level);
	
	if (level > 0) {
		level = (level > 6 ? 6 : level);
		
		switch (level) {
			case 1:
				/* BASIC optimizations ONLY */
				/*
				 * 2 passes of the peephole optimizer
				 * variable optimization
				 * jump optimization (done in peephole)
				 */
				 OPTIMIZER_G(peephole_passes) = 2;
				break;
			case 2:
				/* BASIC with Data Flow Analysis */
				/* 
				 * 2 passes of the peephole optimizer WITH data flow analysis on
				 * variable optimization
				 * jump optimization (done with peephole)
				 */
				OPTIMIZER_G(peephole_passes) = 2;
				break;
			case 3:
				/* BASIC, DFA and global optimizations */
				/*
				 * Once the optimizer is able to gain information about an 
				 *   entire sorce file level 3 might be where this should 
				 *   actually happen... Basically at this point things like 
				 *   inling and whatnot can start to happen as we will gather 
				 *   better global info about the scope of a op-array
				 *
				 */
				OPTIMIZER_G(peephole_passes) = 2;
				break;
			case 4:
				/* BASIC, DFA, global, CSE and Loop? */
				/* Not sure yet... but basically a lot of optimizations... */
				OPTIMIZER_G(peephole_passes) = 2;
				break;
			case 5:
				/* Full out optimization */
				OPTIMIZER_G(peephole_passes) = 3;
				/* Possible future settings */
				/* OPTIMIZER_G(do_dfa) = 1; */
				/* OPTIMIZER_G(do_cse) = 1; */
				/* OPTIMIZER_G(do_scheduling) = 1; */
				/* OPTIMIZER_G(do_loops) = 1; */
				break;
			case 6:
				/* EVERYTHING AND experimental optimizations */
				/*
				 * Everything from level 6 PLUS any experimental optimizations 
				 *   that *might* not be finished or have all the bugs and 
				 *   quarks worked out.  
				 */
				OPTIMIZER_G(peephole_passes) = 3;
				break;
		}
	}
}


static void get_var(CB* cb, zend_op_array* op_array, char* global)
{
	CB* p = cb;
	char *used = do_alloca(op_array->T * sizeof(char));

	if (!used) {
		return;
	}

	memset(global, 0, op_array->T * sizeof(char));

	if (p != NULL && p->next != NULL) {
		int cb_count = 0;
		char *def = do_alloca(op_array->T * sizeof(char));

		if (!def) {
			return;
		}

		while (p != NULL) {
			zend_op* op = p->start;
			zend_op* end = op + p->len;

			memset(def, 0, op_array->T * sizeof(char));

			while (op < end) {
				if ((OP1_TYPE(op) == IS_VAR || OP1_TYPE(op) == IS_TMP_VAR) && !def[VAR_NUM(OP1_VR(op))] && !global[VAR_NUM(OP1_VR(op))]) {
					global[VAR_NUM(OP1_VR(op))] = 1;
				}
				if ((OP2_TYPE(op) == IS_VAR || OP2_TYPE(op) == IS_TMP_VAR) && !def[VAR_NUM(OP2_VR(op))] && !global[VAR_NUM(OP2_VR(op))]) {
					if (op->opcode != ZEND_OP_DATA) {
						global[VAR_NUM(OP2_VR(op))] = 1;
					}
				}
				if (op->opcode == ZEND_DECLARE_INHERITED_CLASS && !def[VAR_NUM(op->extended_value)] && !global[VAR_NUM(op->extended_value)]) {
					global[VAR_NUM(op->extended_value)] = 1;
				}
				if ((RES_TYPE(op) == IS_VAR && (op->opcode == ZEND_RECV || op->opcode == ZEND_RECV_INIT || !USED_RESULT(op))) || (RES_TYPE(op) == IS_TMP_VAR)) {
					if (!def[VAR_NUM(RES_VR(op))] && !global[VAR_NUM(RES_VR(op))]) {
						switch (op->opcode) {
							case ZEND_RECV:
							case ZEND_RECV_INIT:	
							case ZEND_ADD_ARRAY_ELEMENT:
								global[VAR_NUM(RES_VR(op))] = 1;	
								break;
						}
					}
					def[VAR_NUM(RES_VR(op))] = 1;
				}
				op++;
			}
			p = p->next;
			cb_count++;
		}
		free_alloca(def);
	}

	p = cb;
	while (p != NULL) {
		zend_op *op = p->start;
		zend_op *end = op + p->len;

		memset(used, 0, op_array->T * sizeof(char));

		while (op < end) {
			end--;

			if (((RES_TYPE(end) == IS_VAR && (end->opcode == ZEND_RECV || end->opcode == ZEND_RECV_INIT || !USED_RESULT(end))) || (RES_TYPE(end) == IS_TMP_VAR)) && !global[VAR_NUM(RES_VR(end))] && !used[VAR_NUM(RES_VR(end))]) {
				switch(end->opcode) {
					case ZEND_JMPZ_EX:
						end->opcode = ZEND_JMPZ;
						RES_TYPE(end) = IS_UNUSED;
						break;
					case ZEND_JMPNZ_EX:
						end->opcode = ZEND_JMPNZ;
						RES_TYPE(end) = IS_UNUSED;
						break;
					case ZEND_ASSIGN_ADD:
					case ZEND_ASSIGN_SUB:
					case ZEND_ASSIGN_MUL:
					case ZEND_ASSIGN_DIV:
					case ZEND_ASSIGN_MOD:
					case ZEND_ASSIGN_SL:
					case ZEND_ASSIGN_SR:
					case ZEND_ASSIGN_CONCAT:
					case ZEND_ASSIGN_BW_OR:
					case ZEND_ASSIGN_BW_AND:
					case ZEND_ASSIGN_BW_XOR:
					case ZEND_PRE_INC:
					case ZEND_PRE_DEC:
					case ZEND_POST_INC:
					case ZEND_POST_DEC:
					case ZEND_ASSIGN:
					case ZEND_ASSIGN_REF:
					case ZEND_DO_FCALL:
					case ZEND_DO_FCALL_BY_NAME:
						if (RES_TYPE(end) == IS_VAR) {
							RESULT_TYPE(end) |= EXT_TYPE_UNUSED;
						}
						break;
					case ZEND_UNSET_VAR:
					case ZEND_UNSET_DIM:
					case ZEND_UNSET_OBJ:
						RES_TYPE(end) = IS_UNUSED;
						break;
					case ZEND_RECV:
					case ZEND_RECV_INIT:
					case ZEND_INCLUDE_OR_EVAL:
					case ZEND_NEW:
					case ZEND_FE_FETCH:
					case ZEND_PRINT:
					case ZEND_INIT_METHOD_CALL:
					case ZEND_INIT_STATIC_METHOD_CALL:
					case ZEND_ASSIGN_DIM:
					case ZEND_ASSIGN_OBJ:
					case ZEND_DECLARE_CLASS:
					case ZEND_DECLARE_INHERITED_CLASS:
						break;
					default:
						if (OP1_TYPE(end) == IS_CONST) {
							zval_dtor(&__OP1_VAL(end));
						}
						if (OP2_TYPE(end) == IS_CONST) {
							zval_dtor(&__OP2_VAL(end));
						}
						SET_TO_NOP(end);
				}
			} else if (RES_TYPE(end) == IS_VAR && USED_RESULT(end) && end->opcode != ZEND_RECV && end->opcode != ZEND_RECV_INIT && used[VAR_NUM(RES_VR(end))]) {
				RESULT_TYPE(end) &= ~EXT_TYPE_UNUSED;
			}

			if ((RES_TYPE(end) == IS_VAR && (end->opcode == ZEND_RECV || end->opcode == ZEND_RECV_INIT || !USED_RESULT(end))) || (RES_TYPE(end) == IS_TMP_VAR)) {
				switch (end->opcode) {
					case ZEND_RECV:
					case ZEND_RECV_INIT:
					case ZEND_ADD_ARRAY_ELEMENT:
						used[VAR_NUM(RES_VR(end))] = 1;
						break;
					default:
						used[VAR_NUM(RES_VR(end))] = 0;
				}
			}

			if (OP1_TYPE(end) == IS_VAR || OP1_TYPE(end) == IS_TMP_VAR) {
				used[VAR_NUM(OP1_VR(end))] = 1;
			}

			if (OP2_TYPE(end) == IS_VAR || OP2_TYPE(end) == IS_TMP_VAR) {
				used[VAR_NUM(OP2_VR(end))] = 1;
			}

			if (end->opcode == ZEND_DECLARE_INHERITED_CLASS) {
				used[VAR_NUM(end->extended_value)] = 1;
			}
		}
		p = p->next;
	}
	free_alloca(used);
}


/* Set FROM as the parent of TO */
#define CB_ADD_PRED(TO,FROM) {	    \
	CBchain *c = (TO)->pchain;	    \
	while (c != NULL) {	            \
		if (c->cb == (FROM)) {	    \
			break;	                \
		}	                        \
		c = c->next;	            \
	}	                            \
	if (c == NULL) {	            \
		c = emalloc(sizeof(*c));	\
		c->cb = (FROM);	            \
		c->next = (TO)->pchain;	    \
		(TO)->pchain = c;	        \
	}	                            \
}                                   \


/* Remove FROM from parent of TO */
#define CB_DEL_PRED(TO,FROM) {	              \
	CBchain *c = (TO)->pchain;	              \
	if (c != NULL) {	                      \
		if (c->cb == (FROM)) {	              \
			(TO)->pchain = c->next;	          \
			efree(c);	                      \
		} else {	                          \
			while (c->next != NULL) {	      \
				if (c->next->cb == (FROM)) {  \
					CBchain *cc = c->next;	  \
					c->next = c->next->next;  \
					efree(cc);	              \
					break;	                  \
				}	                          \
				c = c->next;	              \
			}	                              \
		}	                                  \
	}	                                      \
}                                             \



#define RM_CB(p) do {	                 \
	if (p->pchain == NULL && p != cb) {	 \
		rm_cb(p);	                     \
	}	                                 \
} while (0)                              \



static void mark_used_cb(CB* cb)
{
	if (cb->used) {
		return;
	}

	cb->used = 1;

	if (cb->jmp_1 != NULL) {
		mark_used_cb(cb->jmp_1);
		CB_ADD_PRED(cb->jmp_1, cb);
	}
	if (cb->jmp_2 != NULL) {
		mark_used_cb(cb->jmp_2);
		CB_ADD_PRED(cb->jmp_2, cb);
	}
	if (cb->jmp_ext != NULL) {
		mark_used_cb(cb->jmp_ext);
		CB_ADD_PRED(cb->jmp_ext, cb);
	}
	if (cb->follow != NULL) {
		mark_used_cb(cb->follow);
		CB_ADD_PRED(cb->follow, cb);
	}
}


static void mark_used_cb2(CB* cb)
{
	if (cb->used) {
		return;
	}

	cb->used = 1;

	if (cb->jmp_1 != NULL) {
		mark_used_cb2(cb->jmp_1);
	}
	if (cb->jmp_2 != NULL) {
		mark_used_cb2(cb->jmp_2);
	}
	if (cb->jmp_ext != NULL) {
		mark_used_cb2(cb->jmp_ext);
	}
	if (cb->follow != NULL) {
		mark_used_cb2(cb->follow);
	}
}


static void rm_cb(CB* cb)
{
	if (cb->used == 0) {
		return;
	}

	cb->used = 0;

	if (cb->jmp_1 != NULL) {
		CB_DEL_PRED(cb->jmp_1, cb);
	}
	if (cb->jmp_2 != NULL) {
		CB_DEL_PRED(cb->jmp_2, cb);
	}
	if (cb->jmp_ext != NULL) {
		CB_DEL_PRED(cb->jmp_ext, cb);
	}
	if (cb->follow != NULL) {
		CB_DEL_PRED(cb->follow, cb);
	}
}


static void del_cb(CB* cb)
{
	zend_op* op = cb->start;
	zend_op* end = op + cb->len;

	rm_cb(cb);
	while (op < end) {
		--end;
		if (OP1_TYPE(end) == IS_CONST) {
			zval_dtor(&__OP1_VAL(end));
		}
		if (OP2_TYPE(end) == IS_CONST) {
			zval_dtor(&__OP2_VAL(end));
		}
		SET_TO_NOP(end);
	}
	cb->len = 0;
	cb->used = 0;
}


static void replace_cb(CB* src, CB* dst)
{
	CBchain* p = src->pchain;
	while (p != NULL) {
		CBchain* q = p->next;
		if (p->cb->jmp_1 == src) {
			p->cb->jmp_1 = dst;
			CB_ADD_PRED(dst,p->cb);
		}
		if (p->cb->jmp_2 == src) {
			p->cb->jmp_2 = dst;
			CB_ADD_PRED(dst,p->cb);
		}
		if (p->cb->jmp_ext == src) {
			p->cb->jmp_ext = dst;
			CB_ADD_PRED(dst,p->cb);
		}
		if (p->cb->follow == src) {
			p->cb->follow = dst;
			CB_ADD_PRED(dst,p->cb);
		}
		efree(p);
		p = q;
	}
	src->pchain = NULL;
}



static void optimize_jmp(CB* cb, zend_op_array* op_array, int *jmp_lines)
{
	CB *p;
	zend_op *op, *oe;
	int i, line_num;

	while(1) {
		zend_bool ok = 1;

		/* Remove Unused Blocks */
		p = cb;
		while (p->next != NULL) {
			if (p->next->used && p->next->pchain) {
				p = p->next;
			} else {
				del_cb(p->next);
				p->next = p->next->next;
				ok = 0;
			}
		}

		/* Clear leading NOP(s) */
		p = cb;
		while (p != NULL) {
			if (p->used) {
				while (p->len > 0 && p->start->opcode ==ZEND_NOP) {
					p->start++;
					p->len--;	
				}
				if (p->len == 0 && p != cb) {
					if (p->follow) {
						replace_cb(p, p->follow);
					}
					rm_cb(p);
					ok = p->used = 0;
				}
			}
			p = p->next;
		}

		/* JMP optimization */
		p = cb;
		while (p != NULL) {
			while (p->next != NULL && (!p->next->used || p->next->pchain == NULL)) {
				del_cb(p->next);
				p->next = p->next->next;
				ok = 0;
			}
			if (p->used && p->len > 0) {
				zend_op* op = &p->start[p->len-1];
				switch (op->opcode) {
					case ZEND_JMP:
jmp:

						/* L1: JMP L1+1  => NOP */
						if (p->jmp_1 == p->next) {
							if (p->follow) { /* this is needed for optimized out BRK/JMP that already have a follow */
								CB_DEL_PRED(p->follow, p);
							}
							p->follow = p->jmp_1;
							p->jmp_1 = NULL;
							SET_TO_NOP(op);
							--(p->len);
							ok = 0;
							break;
						}
						/*	JMP L1  =>  JMP L2
							...         ...
							L1: JMP L2      JMP L2
						*/
						while (p->jmp_1->len == 1 && p->jmp_1->start->opcode == ZEND_JMP && p->jmp_1 != p) {
							CB* x_p = p->jmp_1;
							CB_DEL_PRED(p->jmp_1, p);
							RM_CB(x_p);
							p->jmp_1 = x_p->jmp_1;
							CB_ADD_PRED(p->jmp_1, p);
							ok = 0;
						}
						break;
					case ZEND_JMPZNZ:
jmp_znz:
						/* JMPZNZ  ?,L1,L1  =>  JMP L1 */
						if (p->jmp_ext == p->jmp_2) {
							op->opcode = ZEND_JMP;
							op->extended_value = 0;
							OP1_TYPE(op) = IS_UNUSED;
							OP2_TYPE(op) = IS_UNUSED;
							p->jmp_1 = p->jmp_2;
							p->jmp_2 = NULL;
							p->jmp_ext = NULL;
							ok = 0;
							goto jmp;
						} else if (OP1_TYPE(op) == IS_CONST) {
							/* JMPZNZ  0,L1,L2  =>  JMP L1 */
							if (!zend_is_true(&__OP1_VAL(op))) {
								op->opcode = ZEND_JMP;
								op->extended_value = 0;
								OP1_TYPE(op) = IS_UNUSED;
								OP2_TYPE(op) = IS_UNUSED;
								if (p->jmp_ext != p->jmp_2) {
									CB_DEL_PRED(p->jmp_ext, p);
									RM_CB(p->jmp_ext);
								}
								p->jmp_1 = p->jmp_2;
								p->jmp_2 = NULL;
								p->jmp_ext = NULL;
								p->follow  = NULL;
								ok = 0;
								goto jmp;
								/* JMPZNZ  1,L1,L2  =>  JMP L2 */
							} else {
								op->opcode = ZEND_JMP;
								op->extended_value = 0;
								OP1_TYPE(op) = IS_UNUSED;
								OP2_TYPE(op) = IS_UNUSED;
								if (p->jmp_ext != p->jmp_2) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_1 = p->jmp_ext;
								p->jmp_2 = NULL;
								p->jmp_ext = NULL;
								p->follow = NULL;
								ok = 0;
								goto jmp;
							}
							/* L1: JMPZNZ ?,L2,L1+1  => JMPZ ?,L2 */
						} else if (p->jmp_ext == p->next) {
							op->opcode = ZEND_JMPZ;
							op->extended_value = 0;
							p->follow = p->jmp_ext;
							p->jmp_ext = NULL;
							ok = 0;
							goto jmp_z;
							/* L1: JMPZNZ ?,L1+1,L2  => JMPNZ ?,L2 */
						} else if (p->jmp_2 == p->next) {
							op->opcode = ZEND_JMPNZ;
							op->extended_value = 0;
							p->follow = p->jmp_2;
							p->jmp_2  = p->jmp_ext;
							p->jmp_ext = NULL;
							ok = 0;
							goto jmp_nz;
						} else if (p->jmp_2->len == 1 && OP1_TYPE(op) == IS_TMP_VAR) {
							/*	JMPZNZ $x,L1,L2  =>  JMPZNZ $x,L3,L2
								...                  ...
								L1: JMPZ   $x,L3         JMPZ   $x,L3
							*/
							/*	JMPZNZ $x,L1,L2  =>  JMPZNZ $x,L3,L2
								...                  ...
								L1: JMPZNZ $x,L3,L4      JMPZNZ $x,L3,L4
							*/
							if ((p->jmp_2->start->opcode == ZEND_JMPZ || p->jmp_2->start->opcode == ZEND_JMPZNZ) && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->jmp_ext) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_2;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_znz;
								/*     JMPZNZ $x,L1,L2  =>  JMPZNZ $x,L1+1,L2
									...                  ...
									L1: JMPNZ  $x,L3         JMPNZ  $x,L3
								*/
							} else if (p->jmp_2->start->opcode == ZEND_JMPNZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->jmp_ext) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_znz;
								/*     JMPZNZ $x,L1,L2  =>  JMPZNZ $x,L1,L3
									...                  ...
									L2: JMPNZ  $x,L3         JMPNZ  $x,L3
								*/
							} else if (p->jmp_ext->start->opcode == ZEND_JMPNZ && OP1_TYPE(p->jmp_ext->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_ext->start)) {
								if (p->jmp_2 != p->jmp_ext) {
									CB_DEL_PRED(p->jmp_ext, p);
									RM_CB(p->jmp_ext);
								}
								p->jmp_ext = p->jmp_ext->jmp_2;
								CB_ADD_PRED(p->jmp_ext, p);
								ok = 0;
								goto jmp_znz;
								/*     JMPZNZ $x,L1,L2  =>  JMPZNZ $x,L1,L4
									...                  ...
									L2: JMPZNZ $x,L3,L4      JMPZNZ $x,L3,L4
								*/
							} else if (p->jmp_ext->start->opcode == ZEND_JMPZNZ && OP1_TYPE(p->jmp_ext->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_ext->start)) {
								if (p->jmp_2 != p->jmp_ext) {
									CB_DEL_PRED(p->jmp_ext, p);
									RM_CB(p->jmp_ext);
								}
								p->jmp_ext = p->jmp_ext->jmp_ext;
								CB_ADD_PRED(p->jmp_ext, p);
								ok = 0;
								goto jmp_znz;
								/*     JMPZNZ $x,L1,L2  =>  JMPZNZ $x,L1,L2+1
									...                  ...
									L2: JMPZ   $x,L3         JMPZ   $x,L3
								*/
							} else if (p->jmp_ext->start->opcode == ZEND_JMPZ && OP1_TYPE(p->jmp_ext->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_ext->start)) {
								if (p->jmp_2 != p->jmp_ext) {
									CB_DEL_PRED(p->jmp_ext, p);
									RM_CB(p->jmp_ext);
								}
								p->jmp_ext = p->jmp_ext->follow;
								CB_ADD_PRED(p->jmp_ext, p);
								ok = 0;
								goto jmp_znz;
							}
						}
						while (p->jmp_2->len == 1 && p->jmp_2->start->opcode == ZEND_JMP) {
							CB* x_p = p->jmp_2;
							if (p->jmp_2 != p->jmp_ext) {
								CB_DEL_PRED(p->jmp_2, p);
								RM_CB(x_p);
							}
							p->jmp_2 = x_p->jmp_1;
							CB_ADD_PRED(p->jmp_2, p);
							ok = 0;
						}
						while (p->jmp_ext->len == 1 && p->jmp_ext->start->opcode == ZEND_JMP) {
							CB* x_p = p->jmp_ext;
							if (p->jmp_2 != p->jmp_ext) {
								CB_DEL_PRED(p->jmp_ext, p);
								RM_CB(x_p);
							}
							p->jmp_ext = x_p->jmp_1;
							CB_ADD_PRED(p->jmp_ext, p);
							ok = 0;
						}
						break;
					case ZEND_JMPZ:
jmp_z:
						/* L1: JMPZ  ?,L1+1  =>  NOP */
						if (p->follow == p->jmp_2) {
							p->jmp_2   = NULL;
							SET_TO_NOP(op);
							--(p->len);
							ok = 0;
							break;
						} else if (OP1_TYPE(op) == IS_CONST) {
							/* JMPZ  0,L1  =>  JMP L1 */
							if (!zend_is_true(&__OP1_VAL(op))) {
								op->opcode = ZEND_JMP;
								OP1_TYPE(op) = IS_UNUSED;
								OP2_TYPE(op) = IS_UNUSED;

								if (p->follow != p->jmp_2) {
									CB_DEL_PRED(p->follow, p);
									RM_CB(p->follow);
								}
								p->jmp_1  = p->jmp_2;
								p->jmp_2  = NULL;
								p->follow = NULL;
								ok = 0;
								goto jmp;
								/* JMPZ  1,L1  =>  NOP */
							} else {
								if (p->follow != p->jmp_2) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2   = NULL;
								SET_TO_NOP(op);
								--(p->len);
								ok = 0;
								break;
							}
							/*	JMPZ ?,L1  =>  JMPZNZ  ?,L1,L2
								JMP  L2        JMP     L2
							*/
						} else if (p->follow->len == 1 && p->follow->start->opcode == ZEND_JMP) {
							CB* x_p = p->follow;
							
							/* handler for when JMPZ and JMP have the same destination address */
							if (x_p->jmp_1 == p->jmp_2) {
								p->jmp_2   = NULL;
								SET_TO_NOP(op);
								--(p->len);
								ok = 0;
								break;
							}

							op->opcode = ZEND_JMPZNZ;
							if (p->jmp_2 != p->follow) {
								CB_DEL_PRED(p->follow, p);
								RM_CB(x_p);
							}
							p->follow = NULL;
							p->jmp_ext = x_p->jmp_1;
							CB_ADD_PRED(p->jmp_ext, p);
							ok = 0;
							goto jmp_znz;
						} else if (p->jmp_2->len == 1 && OP1_TYPE(op) == IS_TMP_VAR) {
							/*	JMPZ $x,L1  =>  JMPZ $x,L2
								...             ...
								L1: JMPZ $x,L2      JMPZ $x,L2
								----------------------------------------
								JMPZ   $x,L1     =>  JMPZ  $x,L2
								...                   ...
								L1: JMPZNZ $x,L2,L3      JMPZNZ $x,L2,L3
							*/
							if ((p->jmp_2->start->opcode == ZEND_JMPZ || p->jmp_2->start->opcode == ZEND_JMPZNZ) && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == p->jmp_2->start->op1.u.var) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_2;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_z;
								/*	JMPZ  $x,L1  =>  JMPZ  $x,L1+1
									...              ...
									L1: JMPNZ $x,L2      JMPNZ $x,L2
								*/
							} else if (p->jmp_2->start->opcode == ZEND_JMPNZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_z;
							}
						}
						goto jmp_2;
					case ZEND_JMPNZ:
jmp_nz:
						/* L1: JMPNZ  ?,L1+1  =>  NOP */
						if (p->follow == p->jmp_2) {
							p->jmp_2 = NULL;
							SET_TO_NOP(op);
							--(p->len);
							ok = 0;
							break;
						} else if (OP1_TYPE(op) == IS_CONST) {
							/* JMPNZ  1,L1  =>  JMP L1 */
							if (zend_is_true(&__OP1_VAL(op))) {
								op->opcode = ZEND_JMP;
								OP1_TYPE(op) = IS_UNUSED;
								OP2_TYPE(op) = IS_UNUSED;
								if (p->follow != p->jmp_2) {
									CB_DEL_PRED(p->follow, p);
									RM_CB(p->follow);
								}
								p->jmp_1 = p->jmp_2;
								p->jmp_2 = NULL;
								p->follow = NULL;
								ok = 0;
								goto jmp;
								/* JMPNZ  0,L1  =>  NOP */
							} else {
								if (p->follow != p->jmp_2) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = NULL;
								SET_TO_NOP(op);
								--(p->len);
								ok = 0;
								break;
							}
							/*	JMPNZ ?,L1  =>  JMPZNZ  ?,L2,L1
								JMP   L2        JMP     L2
							*/
						} else if (p->follow->len == 1 && p->follow->start->opcode == ZEND_JMP) {
							CB* x_p = p->follow;

							op->opcode = ZEND_JMPZNZ;
							if (p->jmp_2 != p->follow) {
								CB_DEL_PRED(p->follow, p);
								RM_CB(p->follow);
							}
							p->follow = NULL;
							p->jmp_ext = p->jmp_2;
							p->jmp_2 = x_p->jmp_1;
							CB_ADD_PRED(p->jmp_2, p);
							ok = 0;
							goto jmp_znz;
							/*	JMPNZ $x,L1  =>  JMPNZ $x,L2
								...              ...
								L1: JMPNZ $x,L2      JMPNZ $x,L2
							*/
						} else if (p->jmp_2->len == 1 && OP1_TYPE(op) == IS_TMP_VAR) {
							if (p->jmp_2->start->opcode == ZEND_JMPNZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_2;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz;
								/*	JMPNZ  $x,L1  =>  JMPNZ  $x,L1+1
									...               ...
									L1: JMPZ   $x,L2      JMPZ $x,L2
								*/
							} else if (p->jmp_2->start->opcode == ZEND_JMPZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz;
								/*	JMPNZ  $x,L1     =>  JMPNZ  $x,L3
									...                   ...
									L1: JMPZNZ $x,L2,L3      JMPZNZ $x,L2,L3
								*/
							} else if (p->jmp_2->start->opcode == ZEND_JMPZNZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && OP1_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_ext;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz;
							}
						}
						goto jmp_2;
					case ZEND_JMPZ_EX:
jmp_z_ex:
						/* L1: JMPZ_EX  $x,L1+1,$x  =>  NOP */
						if (p->follow == p->jmp_2 && OP1_TYPE(op) == IS_TMP_VAR && RES_TYPE(op) == IS_TMP_VAR && OP1_VR(op) == RES_VR(op)) {
							p->jmp_2   = NULL;
							SET_TO_NOP(op);
							--(p->len);
							ok = 0;
							break;
							/* L1: JMPZ_EX  $x,L1+1,$y  =>  BOOL $x,$y */
						} else if (p->follow == p->jmp_2) {
							p->jmp_2   = NULL;
							op->opcode = ZEND_BOOL;
							OP2_TYPE(op) = IS_UNUSED;
							ok = 0;
							break;
						} else if (p->jmp_2->len == 1 && RES_TYPE(op) == IS_TMP_VAR) {
							/*	JMPZ_EX ?,L1,$x  =>  JMPZ_EX ?,L2,$x
								...                  ...
								L1: JMPZ    $x,L2        JMPZ    $x,L2
								------------------------------------------
								JMPZ_EX ?,L1,$x  =>  JMPZ_EX ?,L2,$x
								...                  ...
								L1: JMPZNZ  $x,L2,L3     JMPZNZ  $x,L2,L3
								------------------------------------------
								JMPZ_EX ?,L1,$x  =>  JMPZ_EX ?,L2,$x
								...                  ...
								L1: JMPZ_EX $x,L2,$x     JMPZ_EX $x,L2,$x
							*/
							if ((
									(p->jmp_2->start->opcode == ZEND_JMPZ || p->jmp_2->start->opcode == ZEND_JMPZNZ) &&
									OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) ||
									(p->jmp_2->start->opcode == ZEND_JMPZ_EX && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
									RES_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start) &&
									RES_VR(op) == p->jmp_2->start->result.u.var)) {

								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_2;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_z_ex;
								/*	JMPZ_EX ?,L1,$x   =>  JMPZ_EX ?,L2+1,$x
							                ...                   ...
									L1: JMPNZ    $x,L2        JMPNZ    $x,L2
									------------------------------------------
									JMPZ_EX ?,L1,$x   =>  JMPZ_EX  ?,L2+1,$x
									...                   ...
									L1: JMPNZ_EX $x,L2,$x     JMPNZ_EX $x,L2,$x
								*/
							} else if ((p->jmp_2->start->opcode == ZEND_JMPNZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
									RES_VR(op) == OP1_VR(p->jmp_2->start)) ||
									(p->jmp_2->start->opcode == ZEND_JMPNZ_EX && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
									RES_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start) &&
									RES_VR(op) == p->jmp_2->start->result.u.var)) {
	
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_z_ex;
								/*	JMPZ_EX ?,L1,$x   =>  JMPZ_EX ?,L1+1,$y
									...                   ...
									L1: BOOL    $x,$y         BOOL    $x,$y
								*/
							} else if (p->jmp_2->start->opcode == ZEND_BOOL && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
								memcpy(&op->result, &p->jmp_2->start->result, sizeof(zval));
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_z_ex;
								/*	JMPZ_EX ?,L1,$x   =>  JMPZ    ?,L1+1
									...                   ...
									L1: FREE    $x            FREE    $x
								*/
							} else if (p->jmp_2->start->opcode == ZEND_FREE && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
								op->opcode = ZEND_JMPZ;
								RES_TYPE(op) = IS_UNUSED;
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_z;
							}
							/*	JMPZ_EX ?,L1,$x   =>  JMPZ ?,L1+1
								...                   ...
								L1: FREE    $x            FREE $x
							*/
						} else if (RES_TYPE(op) == IS_TMP_VAR && p->jmp_2->start->opcode == ZEND_FREE && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
							if (p->jmp_2->len > 1) {
								/* splitting */
								CB* new_cb = (p->jmp_2+1);
								new_cb->used = 1;
								new_cb->start = p->jmp_2->start+1;
								new_cb->len = p->jmp_2->len-1;
								p->jmp_2->len = 1;
								new_cb->next = p->jmp_2->next;
								p->jmp_2->next = new_cb;
								new_cb->pchain = NULL;
	
								if (p->jmp_2->jmp_1) {
									new_cb->jmp_1 = p->jmp_2->jmp_1;
									CB_ADD_PRED(new_cb->jmp_1, new_cb);
									CB_DEL_PRED(new_cb->jmp_1, p->jmp_2);
									p->jmp_2->jmp_1 = NULL;
								}
								if (p->jmp_2->jmp_2) {
									new_cb->jmp_2 = p->jmp_2->jmp_2;
									CB_ADD_PRED(new_cb->jmp_2, new_cb);
									CB_DEL_PRED(new_cb->jmp_2, p->jmp_2);
									p->jmp_2->jmp_2 = NULL;
								}
								if (p->jmp_2->jmp_ext) {
									new_cb->jmp_ext = p->jmp_2->jmp_ext;
									CB_ADD_PRED(new_cb->jmp_ext, new_cb);
									CB_DEL_PRED(new_cb->jmp_ext, p->jmp_2);
									p->jmp_2->jmp_ext   = NULL;
								}
								op->opcode = ZEND_JMPZ;
								RES_TYPE(op) = IS_UNUSED;

								if (p->jmp_2->follow) {
									new_cb->follow = p->jmp_2->follow;
									CB_ADD_PRED(new_cb->follow, new_cb); 
									CB_DEL_PRED(new_cb->follow, p->jmp_2);
									p->jmp_2->follow = NULL;
								}
								p->jmp_2->follow = new_cb;
								CB_ADD_PRED(p->jmp_2->follow, p->jmp_2);
							}
							if (p->jmp_2 != p->follow) {
								CB_DEL_PRED(p->jmp_2, p);
								RM_CB(p->jmp_2);
							}
							p->jmp_2 = p->jmp_2->follow;
							CB_ADD_PRED(p->jmp_2, p);
							ok = 0;
							goto jmp_z;
						}
						goto jmp_2;
					case ZEND_JMPNZ_EX:
jmp_nz_ex:
				            /* L1: JMPNZ_EX  $x,L1+1,$x  =>  NOP */
						if (p->follow == p->jmp_2 && OP1_TYPE(op) == IS_TMP_VAR && RES_TYPE(op) == IS_TMP_VAR && OP1_VR(op) == RES_VR(op)) {
							p->jmp_2   = NULL;
							SET_TO_NOP(op);
							--(p->len);
							ok = 0;
							break;
							/* L1: JMPNZ_EX  $x,L1+1,$y  =>  BOOL $x,$y */
						} else if (p->follow == p->jmp_2) {
							p->jmp_2   = NULL;
							op->opcode = ZEND_BOOL;
							OP2_TYPE(op) = IS_UNUSED;
							ok = 0;
							break;
						} else if (p->jmp_2->len == 1 && RES_TYPE(op) == IS_TMP_VAR) {
							/*	JMPNZ_EX ?,L1,$x  =>  JMPNZ_EX ?,L2,$x
								...                   ...
								L1: JMPNZ    $x,L2        JMPNZ    $x,L2
								------------------------------------------
								JMPNZ_EX ?,L1,$x  =>  JMPNZ_EX ?,L2,$x
								...                   ...
								L1: JMPNZ_EX $x,L2,$x     JMPNZ_EX $x,L2,$x
							*/
							if ((p->jmp_2->start->opcode == ZEND_JMPNZ &&
									OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
									RES_VR(op) == OP1_VR(p->jmp_2->start)) ||
									(p->jmp_2->start->opcode == ZEND_JMPNZ_EX &&
									OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
									RES_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
									RES_VR(op) == OP1_VR(p->jmp_2->start) &&
									RES_VR(op) == p->jmp_2->start->result.u.var)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_2;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz_ex;
							/*	JMPNZ_EX ?,L1,$x   =>  JMPNZ_EX ?,L3,$x
								...                    ...
						               L1: JMPZNZ   $x,L2,L3      JMPZNZ   $x,L2,L3
							*/
							} else if (p->jmp_2->start->opcode == ZEND_JMPZNZ && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->jmp_ext;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz_ex;
								/*	JMPNZ_EX ?,L1,$x   =>  JMPNZ_EX ?,L1+1,$x
									...                    ...
									L1: JMPZ    $x,L2          JMPZ    $x,L2
									------------------------------------------
									JMPNZ_EX ?,L1,$x   =>  JMPNZ_EX  ?,L1+1,$x
									...                    ...
									L1: JMPZ_EX $x,L2,$x      JMPZ_EX $x,L2,$x
								*/
							} else if ((p->jmp_2->start->opcode == ZEND_JMPZ &&
										OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
										RES_VR(op) == OP1_VR(p->jmp_2->start)) ||
										(p->jmp_2->start->opcode == ZEND_JMPZ_EX &&
										OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
										RES_TYPE(p->jmp_2->start) == IS_TMP_VAR &&
										RES_VR(op) == OP1_VR(p->jmp_2->start) &&
										RES_VR(op) == p->jmp_2->start->result.u.var)) {
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
									p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz_ex;
								/*	JMPNZ_EX ?,L1,$x   =>  JMPNZ_EX ?,L1+1,$y
									...                   ...
									L1: BOOL    $x,$y         BOOL    $x,$y
								*/
							} else if (p->jmp_2->start->opcode == ZEND_BOOL && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
								memcpy(&op->result, &p->jmp_2->start->result, sizeof(zval));
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz_ex;
								/*	JMPNZ_EX ?,L1,$x   =>  JMPNZ ?,L1+1
									...                    ...
									L1: FREE    $x             FREE    $x
								*/
							} else if (p->jmp_2->start->opcode == ZEND_FREE && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
								op->opcode = ZEND_JMPNZ;
								RES_TYPE(op) = IS_UNUSED;
								if (p->jmp_2 != p->follow) {
									CB_DEL_PRED(p->jmp_2, p);
									RM_CB(p->jmp_2);
								}
								p->jmp_2 = p->jmp_2->follow;
								CB_ADD_PRED(p->jmp_2, p);
								ok = 0;
								goto jmp_nz;
							}
							/*	JMPNZ_EX ?,L1,$x   =>  JMPNZ_EX ?,L1+1,$x
								...                    ...
								L1: FREE    $x             FREE    $x
							*/
						} else if (RES_TYPE(op) == IS_TMP_VAR && p->jmp_2->start->opcode == ZEND_FREE && OP1_TYPE(p->jmp_2->start) == IS_TMP_VAR && RES_VR(op) == OP1_VR(p->jmp_2->start)) {
							if (p->jmp_2->len > 1) {
								/* splitting */
								CB* new_cb = (p->jmp_2+1);
								new_cb->used = 1;
								new_cb->start = p->jmp_2->start+1;
								new_cb->len = p->jmp_2->len-1;
								p->jmp_2->len = 1;
								new_cb->next = p->jmp_2->next;
								p->jmp_2->next = new_cb;
								new_cb->pchain   = NULL;

								if (p->jmp_2->jmp_1) {
									new_cb->jmp_1 = p->jmp_2->jmp_1;
									CB_ADD_PRED(new_cb->jmp_1, new_cb);
									CB_DEL_PRED(new_cb->jmp_1, p->jmp_2);
									p->jmp_2->jmp_1 = NULL;
								}
								if (p->jmp_2->jmp_2) {
									new_cb->jmp_2 = p->jmp_2->jmp_2;
									CB_ADD_PRED(new_cb->jmp_2, new_cb);
									CB_DEL_PRED(new_cb->jmp_2, p->jmp_2);
									p->jmp_2->jmp_2 = NULL;
								}
								if (p->jmp_2->jmp_ext) {
									new_cb->jmp_ext = p->jmp_2->jmp_ext;
									CB_ADD_PRED(new_cb->jmp_ext, new_cb);
									CB_DEL_PRED(new_cb->jmp_ext, p->jmp_2);
									p->jmp_2->jmp_ext = NULL;
								}
								if (p->jmp_2->follow) {
									new_cb->follow = p->jmp_2->follow;
									CB_ADD_PRED(new_cb->follow, new_cb);
									CB_DEL_PRED(new_cb->follow, p->jmp_2);
									p->jmp_2->follow = NULL;
								}
								p->jmp_2->follow = new_cb;
								CB_ADD_PRED(p->jmp_2->follow, p->jmp_2);
							}
							op->opcode = ZEND_JMPNZ;
							RES_TYPE(op) = IS_UNUSED;
							if (p->jmp_2 != p->follow) {
								CB_DEL_PRED(p->jmp_2, p);
								RM_CB(p->jmp_2);
							}
							p->jmp_2 = p->jmp_2->follow;
							CB_ADD_PRED(p->jmp_2, p);
							ok = 0;
							goto jmp_nz;
						}
						goto jmp_2;
					case ZEND_NEW:
					case ZEND_FE_FETCH:
jmp_2:
						while (p->jmp_2->len == 1 && p->jmp_2->start->opcode == ZEND_JMP) {
							CB* x_p = p->jmp_2;
							if (p->jmp_2 != p->follow) {
								CB_DEL_PRED(p->jmp_2, p);
								RM_CB(x_p);
							}
							p->jmp_2 = x_p->jmp_1;
							CB_ADD_PRED(p->jmp_2, p);
							ok = 0;
						}
				}
			}
			

			/* Merge Blocks */
			if (
				p->used && p->pchain != NULL && p->pchain->cb->used && p->pchain->next == NULL &&
				p->pchain->cb->follow == p && p->pchain->cb->next == p && p->pchain->cb->jmp_1 == NULL &&
				p->pchain->cb->jmp_2 == NULL && p->pchain->cb->jmp_ext == NULL && !p->exn
			) {
				CB* x = p->pchain->cb;
				CB_DEL_PRED(p, x);
				x->len = &p->start[p->len] - x->start;

				if (p->jmp_1 != NULL) {
					x->jmp_1 = p->jmp_1;
					CB_DEL_PRED(p->jmp_1, p);
					CB_ADD_PRED(p->jmp_1, x);
				}

				if (p->jmp_2 != NULL) {
					x->jmp_2 = p->jmp_2;
					CB_DEL_PRED(p->jmp_2, p);
					CB_ADD_PRED(p->jmp_2, x);
				}

				if (p->jmp_ext != NULL) {
					x->jmp_ext = p->jmp_ext;
					CB_DEL_PRED(p->jmp_ext, p);
					CB_ADD_PRED(p->jmp_ext, x);
				}
				x->follow = p->follow;

				if (p->follow != NULL) {
					CB_DEL_PRED(p->follow, p);
					CB_ADD_PRED(p->follow, x);
				}
				p->used = 0;
				p->len = 0;
				ok = 0;
			}

			p = p->next;
		}

		if (ok) {
			/* Eliminate JMP to RETURN or EXIT */
			p = cb;
			while (p != NULL) {
				if (p->used && p->len > 0) {
					zend_op* op = &p->start[p->len-1];

					if (op->opcode == ZEND_JMP && p->jmp_1->len == 1 && (p->jmp_1->start->opcode == ZEND_RETURN || p->jmp_1->start->opcode == ZEND_EXIT)) {
						CB_DEL_PRED(p->jmp_1, p);
						RM_CB(p->jmp_1);
						memcpy(op, p->jmp_1->start, sizeof(zend_op));

						if (OP1_TYPE(op) == IS_CONST)  {
							zval_copy_ctor(&__OP1_VAL(op));
						}
						p->jmp_1 = NULL;
						ok = 0;
					}
				}
				p = p->next;
			}
		}
		if (ok) {
			break;
		}
	}


	op = op_array->opcodes;
	oe = op + op_array->last;

	memset(jmp_lines, 0, sizeof(int) * op_array->last);

    /*
    int _debug_i = 0;
	for (line_num = 0; op < oe; op++, line_num++, _debug_i++) {
		printf("%d: jmp op_array 4: %p\n", _debug_i, op_array);
		switch (op->opcode) {
			case ZEND_JMP:
				for (i = line_num; i < op->op1.u.opline_num; i++) {
					jmp_lines[i] = 1;
				}
				break;
			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
				for (i = line_num; i < op->op2.u.opline_num; i++) {
					jmp_lines[i] = 1;
				}
				break;
		}
	}
	*/
	
	/* Safe jmp_lines (this one has bounds checking for jump locations...) */
	/*
	 * Were setting any line within a JMP instruction to 1 in this case. This
	 *   will indicate later on in the program that something is nested within 
	 *   a conditional block (or anything else with a JMP). 
	 * This is a VERY simple system that will exclude a LOT of information from 
	 *   being optimized. A better system might need to be devised. 
	 *
	 */
	for (line_num = 0; op < oe; op++, line_num++) {
		switch(op->opcode) {
			case ZEND_JMP:
				for (i = line_num; i < op->op1.u.opline_num && i < op_array->last; i++) {
					jmp_lines[i] = 1;
				}
				break;
			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
				for (i = line_num; i < op->op2.u.opline_num && i < op_array->last; i++) {
					jmp_lines[i] = 1;
				}
				break;
		}
	}
}


/*
 * Returns 1 if the zerver var in key can be optimized out
 *
 */
static zend_bool optimizeable_server_var(char *ops, char *key, int key_len)
{
	char *e = ops + strlen(ops);
	char *p = php_memnstr(ops, key, key_len, e);
	
	if (!p) {
		return 0;
	}
	if (p > ops && *(p - 1) != ' ') {
		return 0;
	}
	if (p + key_len < e && *(p + key_len) != ' ') {
		return 0;
	}
	return 1;
}


int optimize_get_constant(const char *name, int name_len, const char *cname, int cname_len, zval *r TSRMLS_DC)
{
	int retval;

	if (cname == NULL) { 
		/* basic constant */
		zend_constant *c = NULL;

		if (zend_hash_find(EG(zend_constants), (char *)name, name_len + 1, (void **)&c) == SUCCESS) {
			if (r) {
				*r = c->value;
			}
			retval = 1;
		} else {
			char *lookup_name = do_alloca(name_len + 1);

			memcpy(lookup_name, name, name_len);
			lookup_name[name_len] = '\0';

			zend_str_tolower(lookup_name, name_len);

			if (zend_hash_find(EG(zend_constants), lookup_name, name_len + 1, (void **)&c) == SUCCESS) {
				if ((c->flags & CONST_CS) && (memcmp(c->name, name, name_len) != 0)) {
					retval = 0;
				} else {
					retval = 1;
					if (r) {
						*r = c->value;
					}
				}
			} else {
				retval = 0;
			}
			free_alloca(lookup_name);
		}
	} else { 
		/* class constant */
		zend_class_entry **ce = NULL;
		zval **c;
		char *tcname = do_alloca(cname_len + 1);

		memcpy(tcname, cname, cname_len);
		tcname[cname_len] = '\0';
		zend_str_tolower(tcname, cname_len);

		if (zend_hash_find(EG(class_table), (char *)tcname, cname_len + 1, (void **)&ce) == FAILURE) {
			retval = 0;
		} else if (zend_hash_find(&(*ce)->constants_table, (char *)name, name_len + 1, (void **)&c) == FAILURE) {
			retval = 0;
		} else {
			if (r) {
				*r = **c;
			}
			retval = 1;
		}
		free_alloca(tcname);
	}

	return retval;
}

#define UNDEFINE_OP(op) Ts[VAR_NUM((op).u.var)] = NULL;
#define DEFINE_OP(op)   Ts[VAR_NUM((op)->result.u.var)] = (op);
#define IS_DEFINED(op)    (Ts[VAR_NUM((op).u.var)] != NULL)
#define DEFINED_OP(op)    (Ts[VAR_NUM((op).u.var)])


void optimize_fullpath(zend_op *op, char *base, zend_bool use_include_path TSRMLS_DC)
{
	char *realpath;

	if (
		OP1_VALT(op) != IS_STRING || 
		IS_ABSOLUTE_PATH(OP1_VAL_S(op), OP1_VAL_SL(op)) ||
		!php_memnstr(OP1_VAL_S(op), "://", sizeof("://") - 1, OP1_VAL_S(op) + OP1_VAL_SL(op))
	) {
		return;
	}

	switch (use_include_path) {
		case ZEND_INCLUDE_ONCE:
		case ZEND_REQUIRE_ONCE:
		case ZEND_INCLUDE:
		case ZEND_REQUIRE:
			{
				char *opath, *pathbuf;
				FILE *fp;
				char *exec_fname = base;
				int exec_fname_length = strlen(exec_fname);

				while ((--exec_fname_length >= 0) && !IS_SLASH(exec_fname[exec_fname_length]));
				if ((exec_fname && exec_fname[0] == '[') || exec_fname_length<=0) {
					/* [no active file] or no path */
					pathbuf = estrdup(OP1_VAL_S(op));
				} else {
					pathbuf = (char *) emalloc(exec_fname_length + OP1_VAL_SL(op) +1 +1);
					memcpy(pathbuf, OP1_VAL_S(op), OP1_VAL_SL(op));
					pathbuf[OP1_VAL_SL(op)] = DEFAULT_DIR_SEPARATOR;
					memcpy(pathbuf+OP1_VAL_SL(op)+1, exec_fname, exec_fname_length);
					pathbuf[OP1_VAL_SL(op) + exec_fname_length +1] = '\0';
				}				
				fp = php_fopen_with_path(OP1_VAL_S(op), "rb", (const char *)pathbuf, &opath TSRMLS_CC);
				efree(pathbuf);
				if (fp) {
					fclose(fp);
					efree(OP1_VAL_S(op));
					OP1_VAL_S(op) = opath;
					OP1_VAL_SL(op) = strlen(opath);
				} else {
					return;
				}
			}
			break;
	}

	if ((realpath = expand_filepath(OP1_VAL_S(op), NULL TSRMLS_CC)) != NULL) {
		efree(OP1_VAL_S(op));
		OP1_VAL_S(op) = realpath;
		OP1_VAL_SL(op) = strlen(realpath);
	}
}


/*
 * Attempt to optimize the defining of a constant 
 * 
 */
static void optimize_define(zend_op** a, zend_op** b, CB* cb, zend_op** Ts TSRMLS_DC) 
{
	return;
	
	zend_op *prev = *a;
	zend_op *op = *b;
	zend_constant c;
	zend_op *val, *name;

	if (op->extended_value == 3) {
		convert_to_long(&__OP1_VAL(prev));
		c.flags = OP1_VAL_L(prev) ? 0 : CONST_CS;
		val = op -	 2;
		name = op - 3;
	} else {
		c.flags = CONST_CS;
		val = prev;
		name = op - 2;
	}
	convert_to_string(&__OP1_VAL(name));

	c.name = zend_strndup(OP1_VAL_S(name), OP1_VAL_SL(name));
	c.name_len = OP1_VAL_SL(name) + 1;
	c.module_number = 0;
	c.value.type = OP1_VAL_L(val);
	c.value = __OP1_VAL(val);

	/* need this hack to avoid crashes due to hackish implementation of STD* constants in CLI */
	if (strcmp(sapi_module.name, "cli")) {
		c.flags |= CONST_PERSISTENT;
	}
	zval_copy_ctor(&c.value);

	if (zend_register_constant(&c TSRMLS_CC) == SUCCESS) {
		int k = 0;

		if (!USED_RESULT(op)) {
			optimize_to_true(&prev, &op, cb, Ts TSRMLS_CC);
			k += 2;
		} else {
			CB_DEL_PRED(cb->jmp_2, cb);
			cb->jmp_2 = NULL;
		}
		while (k <= op->extended_value) {
			zval_dtor(&__OP1_VAL((op-k)));
			SET_TO_NOP(op-k);
			k++;
		}
	} else if (c.flags & CONST_PERSISTENT) {
		zval_dtor(&c.value);
	}
}



int get_next_znode_pos(zend_op* op, zend_bool var_only)
{
	zend_op *x;
	
	x = NEXT_OP(op);

	switch (x->opcode) {
		case ZEND_ADD:
		case ZEND_SUB:
		case ZEND_MUL:
		case ZEND_DIV:
		case ZEND_MOD:
		case ZEND_BOOL_NOT:
		case ZEND_BOOL_XOR:
		case ZEND_BOOL:
		case ZEND_IS_IDENTICAL:
		case ZEND_IS_NOT_IDENTICAL:
		case ZEND_IS_EQUAL:
		case ZEND_IS_NOT_EQUAL:
		case ZEND_IS_SMALLER:
		case ZEND_IS_SMALLER_OR_EQUAL:
		case ZEND_CONCAT:
			if (OP1_TYPE(x) == IS_VAR && OP1_VR(x) == RES_VR(op)) {
				return 1;
			} else if (OP2_TYPE(x) == IS_VAR && OP2_VR(x) == RES_VR(op)) {
				return 2;
			} else {
				return 0;
			}
			break;
		case ZEND_QM_ASSIGN:
		case ZEND_SEND_VAR:
		case ZEND_SEND_VAL:
		case ZEND_ECHO:
		case ZEND_JMPZ:
		case ZEND_JMPNZ:
		case ZEND_RETURN:
			if (OP1_VR(x) == RES_VR(op) && ((var_only && OP1_TYPE(x) == IS_VAR) || (!var_only && OP1_IS_VART(x)))) {
			//if ((OP1_TYPE(x) == IS_VAR || OP1_TYPE(x) == IS_CV || OP1_TYPE(x) == IS_TMP_VAR) && OP1_VR(x) == RES_VR(op)) {
				return 1;
			} else {
				return 0;
			}
		case ZEND_ASSIGN:
		case ZEND_ASSIGN_CONCAT:
			if (OP2_VR(x) == RES_VR(op) && ((var_only && OP2_TYPE(x) == IS_VAR) || (!var_only && OP2_IS_VART(x)))) {
			//if ((OP2_TYPE(x) == IS_VAR || OP2_TYPE(x) == IS_CV || OP2_TYPE(x) == IS_TMP_VAR) && OP2_VR(x) == RES_VR(op)) {
				return 2;
			} else {
				return 0;
			}
		case ZEND_ASSIGN_DIM:
			if (NEXT_OP(x)->opcode == ZEND_OP_DATA && OP1_VR(NEXT_OP(x)) == RES_VR(op)) {
				return 3;				
			}
		case ZEND_FETCH_W:
			if (
					OP1_TYPE(x) == IS_CONST && OP1_VALT(x) == IS_STRING && 
					!strcmp(OP1_VAL_S(x), "GLOBALS") &&
					NEXT_OP(NEXT_OP(x))->opcode == ZEND_OP_DATA && 
					OP1_VR(NEXT_OP(NEXT_OP(x))) == RES_VR(op)
			) {
				return 4;				
			}
		case ZEND_SEND_VAR_NO_REF:
			if ((x->extended_value & ZEND_ARG_COMPILE_TIME_BOUND) && !(x->extended_value & ZEND_ARG_SEND_BY_REF)) {
				x->opcode = ZEND_SEND_VAR;
				x->extended_value = ZEND_DO_FCALL;
				
				if (OP1_VR(x) == RES_VR(op) && ((var_only && OP1_TYPE(x) == IS_VAR) || (!var_only && OP1_IS_VART(x)))) {
				//if ((OP1_TYPE(x) == IS_VAR || OP1_TYPE(x) == IS_CV || OP1_TYPE(x) == IS_TMP_VAR) && OP1_VR(x) == RES_VR(op)) {
					return 1;
				} else {
					return 0;
				}
			} else {
				return 0;
			}
		default:
			return 0;
	}
}


#define OPTIMIZE_TO_HEADER                                \
	zend_op* prev = NULL;                                 \
	if (a == NULL) {                                      \
		flags = flags & (0xFF ^ OPTIMIZE_TO_DEL_PREV);    \
	} else {                                              \
		prev = (a != b ? PREV_OP(*b) : *a);               \
	}                                                     \
	zend_op* op   = (b != a ? *b : NULL);                 \
	zend_op* next = NEXT_OP(*b);                          \
	znode* n;                                             \
	int nn;                                               \
	if (flags & OPTIMIZE_TO_FCALL) {                      \
		nn = can_optimize_fcall(*b);                      \
	} else {                                              \
		nn = get_next_znode_pos(*b, 0);                   \
	}                                                     \
	switch(nn) {                                          \
		case 1: n = &next->op1; break;                    \
		case 2: n = &next->op2; break;                    \
		case 3: n = &NEXT_OP(next)->op1; break;           \
		case 4: n = &NEXT_OP(NEXT_OP(next))->op1; break;  \
		default: return;                                  \
	}                                                     \
	if (next->opcode == ZEND_ASSIGN) {	                  \
		UNDEFINE_OP(next->op2);	                          \
	} else if (next->opcode == ZEND_SEND_VAR) {	          \
		next->opcode = ZEND_SEND_VAL;	                  \
	}                                                     \


#define OPTIMIZE_TO_FOOTER                               \
	if (flags & OPTIMIZE_TO_DEL_PREV) {                  \
		SET_TO_NOP_EX(prev);                             \
	}                                                    \
	if (flags & OPTIMIZE_TO_DEL_OP) {                    \
		if (op && op->opcode == ZEND_FETCH_DIM_R) {	     \
			if (op) {	                                 \
				zval_dtor(&__OP2_VAL(op));      	     \
				SET_TO_NOP(op);	                         \
			}	                                         \
		} else {	                                     \
			if (op) {	                                 \
				zval_dtor(&__OP1_VAL(op));	             \
				SET_TO_NOP(op);	                         \
			}	                                         \
			CB_DEL_PRED(cbl->jmp_2, cbl);       	     \
			cbl->jmp_2 = NULL;	                         \
		}                                                \
	}                                                    \



void optimize_to_bool_ex(zend_constant* c, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	OPTIMIZE_TO_HEADER

	n->op_type = IS_CONST;
	n->u.constant = c->value;
	zval_copy_ctor(&n->u.constant);
	INIT_PZVAL(&n->u.constant);

	OPTIMIZE_TO_FOOTER
}


void optimize_to_true_ex(zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	zend_constant *c = NULL;

	/* we shouldn't have PHP without "true" constant */
	if (zend_hash_find(EG(zend_constants), "true", sizeof("true"), (void **) &c) != SUCCESS) {
		assert(0);
	}

	optimize_to_bool_ex(c, a, b, cbl, Ts, flags TSRMLS_CC);
}


void optimize_to_false_ex(zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	zend_constant *c = NULL;

	/* we shouldn't have PHP without "false" constant */
	if (zend_hash_find(EG(zend_constants), "false", sizeof("false"), (void **) &c) != SUCCESS) {
		assert(0);
	}

	optimize_to_bool_ex(c, a, b, cbl, Ts, flags TSRMLS_CC);
}


void optimize_to_null_ex(zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	zend_constant *c = NULL;

	/* we shouldn't have PHP without "null" constant */
	if (zend_hash_find(EG(zend_constants), "null", sizeof("null"), (void **) &c) != SUCCESS) {
		assert(0);
	}

	optimize_to_bool_ex(c, a, b, cbl, Ts, flags TSRMLS_CC);
}


void optimize_to_string_ex(char* str, int str_len, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	OPTIMIZE_TO_HEADER

	n->op_type = IS_CONST;
	INIT_PZVAL(&n->u.constant);
	ZVAL_STRINGL(&n->u.constant, str, str_len, 1);

	OPTIMIZE_TO_FOOTER
}


void optimize_to_long_ex(long val, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	OPTIMIZE_TO_HEADER

	n->op_type = IS_CONST;
	INIT_PZVAL(&n->u.constant);
	ZVAL_LONG(&n->u.constant, val);

	OPTIMIZE_TO_FOOTER
}


void optimize_to_double_ex(double val, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC)
{
	OPTIMIZE_TO_HEADER

	n->op_type = IS_CONST;
	INIT_PZVAL(&n->u.constant);
	ZVAL_DOUBLE(&n->u.constant, val);

	OPTIMIZE_TO_FOOTER
}


void optimize_to_var_ex(zend_uint var, int var_type,  zend_op **a, zend_op **b, CB *cbl, zend_op **Ts, zend_uchar flags TSRMLS_DC)
{
	OPTIMIZE_TO_HEADER
	
	/* Undo ZEND_SEND_VAR -> ZEND_SEND_VAL */
	if (next->opcode == ZEND_SEND_VAL) {
		next->opcode = ZEND_SEND_VAR;
	}
	
	n->op_type = var_type;
	n->u.var = var;
	
	OPTIMIZE_TO_FOOTER
}


/*
 * Returns 1 if the op is known to return a numeric result, 0 otherwise
 * 
 */
static int is_numeric_result(zend_op* x)
{
	switch (x->opcode) {
		case ZEND_ADD:
		case ZEND_SUB:
		case ZEND_MUL:
		case ZEND_DIV:
		case ZEND_MOD:
		case ZEND_SL:
		case ZEND_SR:

		case ZEND_ASSIGN_ADD:
		case ZEND_ASSIGN_SUB:
		case ZEND_ASSIGN_MUL:
		case ZEND_ASSIGN_DIV:
		case ZEND_ASSIGN_MOD:
		case ZEND_ASSIGN_SL:
		case ZEND_ASSIGN_SR:		

		case ZEND_BOOL_NOT:
		case ZEND_BOOL_XOR:
		case ZEND_BOOL:

		case ZEND_IS_IDENTICAL:
		case ZEND_IS_NOT_IDENTICAL:
		case ZEND_IS_EQUAL:
		case ZEND_IS_NOT_EQUAL:
		case ZEND_IS_SMALLER:
		case ZEND_IS_SMALLER_OR_EQUAL:

		case ZEND_PRE_DEC:
		case ZEND_PRE_INC:
		case ZEND_POST_INC:
		case ZEND_POST_DEC:
			return 1;

		case ZEND_CAST:
			return (x->extended_value == IS_BOOL || x->extended_value == IS_LONG || x->extended_value == IS_DOUBLE);

		case ZEND_DO_FCALL:

#define OPTIMIZER_INT_FUNC_CHECK(name) !strcmp(OP1_VAL_S(x), name)

			if (OP1_TYPE(x) == IS_CONST && OP1_VALT(x) == IS_STRING && 
				(
					OPTIMIZER_INT_FUNC_CHECK("count") ||
					OPTIMIZER_INT_FUNC_CHECK("extract") ||
					OPTIMIZER_INT_FUNC_CHECK("array_push") ||
					OPTIMIZER_INT_FUNC_CHECK("array_unshift") ||
					OPTIMIZER_INT_FUNC_CHECK("assert") ||
					OPTIMIZER_INT_FUNC_CHECK("ip2long") ||
					OPTIMIZER_INT_FUNC_CHECK("get_magic_quotes_runtime") ||
					OPTIMIZER_INT_FUNC_CHECK("get_magic_quotes_gpc") ||
					OPTIMIZER_INT_FUNC_CHECK("connection_aborted") ||
					OPTIMIZER_INT_FUNC_CHECK("connection_status") ||
					OPTIMIZER_INT_FUNC_CHECK("ignore_user_abort") ||
					OPTIMIZER_INT_FUNC_CHECK("getservbyname") ||
					OPTIMIZER_INT_FUNC_CHECK("getprotobyname") ||
					OPTIMIZER_INT_FUNC_CHECK("dl") ||
					OPTIMIZER_INT_FUNC_CHECK("dns_check_record") ||
					OPTIMIZER_INT_FUNC_CHECK("system") ||
					OPTIMIZER_INT_FUNC_CHECK("file_put_contents") ||
					OPTIMIZER_INT_FUNC_CHECK("pclose") ||
					OPTIMIZER_INT_FUNC_CHECK("fwrite") ||
					OPTIMIZER_INT_FUNC_CHECK("ftell") ||
					OPTIMIZER_INT_FUNC_CHECK("fseek") ||
					OPTIMIZER_INT_FUNC_CHECK("mkdir") ||
					OPTIMIZER_INT_FUNC_CHECK("readfile") ||
					OPTIMIZER_INT_FUNC_CHECK("umask") ||
					OPTIMIZER_INT_FUNC_CHECK("fpassthru") ||
					OPTIMIZER_INT_FUNC_CHECK("fputcsv") ||
					OPTIMIZER_INT_FUNC_CHECK("fileperms") ||
					OPTIMIZER_INT_FUNC_CHECK("fileinode") ||
					OPTIMIZER_INT_FUNC_CHECK("filesize") ||
					OPTIMIZER_INT_FUNC_CHECK("fileowner") ||
					OPTIMIZER_INT_FUNC_CHECK("filegroup") ||
					OPTIMIZER_INT_FUNC_CHECK("fileatime") ||
					OPTIMIZER_INT_FUNC_CHECK("filemtime") ||
					OPTIMIZER_INT_FUNC_CHECK("filectime") ||
					OPTIMIZER_INT_FUNC_CHECK("printf") ||
					OPTIMIZER_INT_FUNC_CHECK("vprintf") ||
					OPTIMIZER_INT_FUNC_CHECK("fprintf") ||
					OPTIMIZER_INT_FUNC_CHECK("vfprintf") ||
					OPTIMIZER_INT_FUNC_CHECK("ftok") ||
					OPTIMIZER_INT_FUNC_CHECK("levenshtein") ||
					OPTIMIZER_INT_FUNC_CHECK("linkinfo") ||
					OPTIMIZER_INT_FUNC_CHECK("symlink") ||
					OPTIMIZER_INT_FUNC_CHECK("link") ||
					OPTIMIZER_INT_FUNC_CHECK("ezmlm_hash") ||
					OPTIMIZER_INT_FUNC_CHECK("mail") ||
					OPTIMIZER_INT_FUNC_CHECK("abs") ||
					OPTIMIZER_INT_FUNC_CHECK("bindec") ||
					OPTIMIZER_INT_FUNC_CHECK("hexdec") ||
					OPTIMIZER_INT_FUNC_CHECK("octdec") ||
					OPTIMIZER_INT_FUNC_CHECK("getmyuid") ||
					OPTIMIZER_INT_FUNC_CHECK("getmygid") ||
					OPTIMIZER_INT_FUNC_CHECK("getmypid") ||
					OPTIMIZER_INT_FUNC_CHECK("getmyinode") ||
					OPTIMIZER_INT_FUNC_CHECK("getlastmod") ||
					OPTIMIZER_INT_FUNC_CHECK("proc_terminate") ||
					OPTIMIZER_INT_FUNC_CHECK("proc_close") ||
					OPTIMIZER_INT_FUNC_CHECK("rand") ||
					OPTIMIZER_INT_FUNC_CHECK("mt_rand") ||
					OPTIMIZER_INT_FUNC_CHECK("getrandmax") ||
					OPTIMIZER_INT_FUNC_CHECK("mt_getrandmax") ||
					OPTIMIZER_INT_FUNC_CHECK("ereg") ||
					OPTIMIZER_INT_FUNC_CHECK("eregi") ||
					OPTIMIZER_INT_FUNC_CHECK("stream_select") ||
					OPTIMIZER_INT_FUNC_CHECK("stream_set_write_buffer") ||
					OPTIMIZER_INT_FUNC_CHECK("stream_socket_enable_crypto") ||
					OPTIMIZER_INT_FUNC_CHECK("strspn") ||
					OPTIMIZER_INT_FUNC_CHECK("strcspn") ||
					OPTIMIZER_INT_FUNC_CHECK("strcoll") ||
					OPTIMIZER_INT_FUNC_CHECK("strpos") ||
					OPTIMIZER_INT_FUNC_CHECK("stripos") ||
					OPTIMIZER_INT_FUNC_CHECK("strrpos") ||
					OPTIMIZER_INT_FUNC_CHECK("strripos") ||
					OPTIMIZER_INT_FUNC_CHECK("ord") ||
					OPTIMIZER_INT_FUNC_CHECK("similar_text") ||
					OPTIMIZER_INT_FUNC_CHECK("strnatcmp") ||
					OPTIMIZER_INT_FUNC_CHECK("strnatcasecmp") ||
					OPTIMIZER_INT_FUNC_CHECK("substr_count") ||
					OPTIMIZER_INT_FUNC_CHECK("substr_compare") ||
					OPTIMIZER_INT_FUNC_CHECK("intval") ||
					OPTIMIZER_INT_FUNC_CHECK("memory_get_usage") ||
					OPTIMIZER_INT_FUNC_CHECK("memory_get_peak_usage") ||
					OPTIMIZER_INT_FUNC_CHECK("version_compare")
				)
			) {
				return 1;
			}
			return 0;
	}
	return 0;
}


/*
 * Optimize binary (math) operation where one of the opperands is a constant 
 *   value and the other is a variable. The are basic math identities being 
 *   optimized.
 */
static void optimize_binary_op_1const(zend_op** a, zend_op** b, CB* cb, zend_op** Ts OPTIMIZER_STATS_PARAM TSRMLS_DC)
{
	zend_op* op = *b;
	zend_op* prev = *a;
	int rl = -1;
	long v;
	zend_uint var;
	int var_type;
			
	if (OP2_TYPE(op) == IS_CONST && (OP1_TYPE(op) == IS_TMP_VAR || OP1_TYPE(op) == IS_CV || OP1_TYPE(op) == IS_VAR)) {
		/* right opperand (op2) */
		convert_scalar_to_number(&__OP2_VAL(op) TSRMLS_CC);
				
		if (OP2_VALT(op) == IS_LONG && (OP2_VAL_L(op) == 0 || OP2_VAL_L(op) == 1)) {
			v = OP2_VAL_L(op);
			var = op->op1.u.var;
			var_type = OP1_TYPE(op);
			rl = 0;
		}
	} else if (OP1_TYPE(op) == IS_CONST && (OP2_TYPE(op) == IS_TMP_VAR || OP2_TYPE(op) == IS_CV || OP2_TYPE(op) == IS_VAR)) {
		/* left opperand (op1) */
		convert_scalar_to_number(&__OP1_VAL(op) TSRMLS_CC);
				
		if (OP1_VALT(op) == IS_LONG && (OP1_VAL_L(op) == 0 || OP1_VAL_L(op) == 1)) {
			v = OP1_VAL_L(op);
			var = op->op2.u.var;
			var_type = OP2_TYPE(op);
			rl = 1;
		}
	}
			
	if (rl >= 0) {
		switch (op->opcode) {
			case ZEND_ADD:
			case ZEND_BW_OR:
				if (v == 0) {
					/* ZEND_ADD: $a+0 -> $a or 0+$a -> $a*/
					/* ZEND_BW_OR: $a|0 -> $a or 0|$a -> $a */
					optimize_to_var(var, var_type, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_SUB:
				if (rl == 0 && v == 0) {
					/* $a-0 -> $a */
					optimize_to_var(var, var_type, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_MUL:
				if (v == 0) {
					/* $a*0 -> 0 or 0*$a -> 0 */
					optimize_to_long(0, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				} else if (v == 1) {
					/* $a*1 -> $a or 1*$a -> $a */
					optimize_to_var(var, var_type, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_DIV:
				if (rl == 0 && v == 1) {
					/* $a/1 -> $a */
					optimize_to_var(var, var_type, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				} else if (rl == 1 && v == 0) {
					/* 0/$a -> 0 */
					optimize_to_long(0, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_MOD:
				if (rl == 1 && v == 0) {
					/* 0%$a -> 0 */
					optimize_to_long(0, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				} else if (rl == 0 && v == 1) {
					/* $a%1 -> 0 */
					optimize_to_long(0, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_BW_AND:
				if (v == 0) {
					/* $a&0 -> 0 or 0&$a -> 0 */
					optimize_to_long(0, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_SR:
			case ZEND_SL:
				if (rl == 0 && v == 0) {
					/* ZEND_SR: $a >> 0 -> $a */
					/* ZEND_SL: $a << 0 -> $a */
					optimize_to_var(var, var_type, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				} else if (rl == 1 && v == 0) {
					/* ZEND_SR: 0 >> $a -> 0 */
					/* ZEND_SL: 0 << $a -> 0 */
					optimize_to_long(0, &prev, &op, cb, Ts);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_ASSIGN_ADD:
			case ZEND_ASSIGN_SUB:
			case ZEND_ASSIGN_BW_OR:
			case ZEND_ASSIGN_SR:
			case ZEND_ASSIGN_SL:
				if (v == 0) {
					/* ZEND_ASSIGN_ADD: $a += 0 -> NOP */
					/* ZEND_ASSIGN_SUB: $a -= 0 -> NOP */
					/* ZEND_ASSIGN_BW_OR: $a |= 0 -> NOP */
					/* ZEND_ASSIGN_SR: $a >>= 0 -> NOP */
					/* ZEND_ASSIGN_SL: $a <<= 0 -> NOP */
					SET_TO_NOP(op);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_ASSIGN_MUL:
				if (v == 0) {
					/* $a *= 0 -> $a=0 */
					op->opcode = ZEND_ASSIGN;
					OP2_VALT(op) = IS_LONG;
					OP2_VAL_L(op) = 0;
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				} else if (v == 1) {
					/* $a *= 1 -> NOP */
					SET_TO_NOP(op);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_ASSIGN_DIV:
				if (v == 1) {
					/* $a /= 1 -> NOP */
					SET_TO_NOP(op);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_ASSIGN_MOD:
				if (v == 1) {
					/* $a %= 1 -> 0 */
					op->opcode = ZEND_ASSIGN;
					OP2_VALT(op) = IS_LONG;
					OP2_VAL_L(op) = 0;
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
			case ZEND_ASSIGN_BW_AND:
				if (v == 0) {
					/* $a &= 0 -> 0 */
					op->opcode = ZEND_ASSIGN;
					OP2_VALT(op) = IS_LONG;
					OP2_VAL_L(op) = 0;
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				break;
		}
	}
}


/*
 * Performs peephole optimization on a code block.
 * 
 */
static void optimize_code_block(CB *cb, zend_op_array *op_array, char *global, int pass, HashTable fh, int *jmp_lines OPTIMIZER_STATS_PARAM TSRMLS_DC)
{
	zend_op* prev = NULL;
	zend_op* op = cb->start;
	zend_op* end = op + cb->len;
	char *server_op = NULL;

	HashTable assigns;
	HashTable fetch_dim;

	zend_op** Ts = do_alloca(sizeof(zend_op*) * op_array->T);
	memset(Ts, 0, sizeof(zend_op*) * op_array->T);

	zend_hash_init(&assigns, 0, NULL, NULL, 0);
	zend_hash_init(&fetch_dim, 0, NULL, NULL, 0);

	server_op = OPTIMIZER_G(optimize_inline_server);
	if (*server_op == '\0') {
		server_op = NULL;
	}

	while (op < end) {
		/* Constant Folding */
		if (OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_QM_ASSIGN && OP1_TYPE(DEFINED_OP(op->op1)) == IS_CONST) {
			zend_op *x = DEFINED_OP(op->op1);
			if (op->opcode != ZEND_CASE) {
				UNDEFINE_OP(op->op1);
				memcpy(&op->op1, &x->op1, sizeof(znode));
				SET_TO_NOP(x);
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
		}
		
		if (OP2_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op2) && DEFINED_OP(op->op2)->opcode == ZEND_QM_ASSIGN && OP1_TYPE(DEFINED_OP(op->op2)) == IS_CONST) {
			zend_op *x = DEFINED_OP(op->op2);
			UNDEFINE_OP(op->op2);
			memcpy(&op->op2, &x->op1, sizeof(znode));
			SET_TO_NOP(x);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		
		
		/* simplify binary operations */
		if (op->opcode == ZEND_IS_EQUAL) {
			if (OP1_TYPE(op) == IS_CONST && (OP1_VALT(op) == IS_BOOL && OP1_VAL_L(op) == 0)) {
				op->opcode = ZEND_BOOL_NOT;
				memcpy(&op->op1, &op->op2, sizeof(znode));
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_BOOL && OP1_VAL_L(op) == 1) {
				op->opcode = ZEND_BOOL;
				memcpy(&op->op1, &op->op2, sizeof(znode));
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_BOOL && OP2_VAL_L(op) == 0) {
				op->opcode = ZEND_BOOL_NOT;
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_BOOL && OP2_VAL_L(op) == 1) {
				op->opcode = ZEND_BOOL;
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_LONG && OP2_VAL_L(op) == 0 && (OP1_TYPE(op) == IS_TMP_VAR || OP1_TYPE(op) == IS_VAR) && IS_DEFINED(op->op1) && is_numeric_result(DEFINED_OP(op->op1))) {
				op->opcode = ZEND_BOOL_NOT;
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
		} else if (op->opcode == ZEND_IS_NOT_EQUAL) {
			if (OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_BOOL && OP1_VAL_L(op) == 0) {
				op->opcode = ZEND_BOOL;
				memcpy(&op->op1, &op->op2, sizeof(znode));
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_BOOL && OP1_VAL_L(op) == 1) {
				op->opcode = ZEND_BOOL_NOT;
				memcpy(&op->op1, &op->op2, sizeof(znode));
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_BOOL && OP2_VAL_L(op) == 0) {
				op->opcode = ZEND_BOOL;
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_BOOL && OP2_VAL_L(op) == 1) {
				op->opcode = ZEND_BOOL_NOT;
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_LONG && OP2_VAL_L(op) == 0 && (OP1_TYPE(op) == IS_TMP_VAR || OP1_TYPE(op) == IS_VAR) && IS_DEFINED(op->op1) && is_numeric_result(DEFINED_OP(op->op1))) {
				op->opcode = ZEND_BOOL;
				OP2_TYPE(op) = IS_UNUSED;
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
		}

		
		/* Optimize binary ops with one constant operand and one variable operand using some basic identities */
		if ((op->opcode == ZEND_ADD || op->opcode == ZEND_SUB || op->opcode == ZEND_MUL || 
		    op->opcode == ZEND_DIV || op->opcode == ZEND_MOD || op->opcode == ZEND_BW_AND || 
		    op->opcode == ZEND_BW_OR || op->opcode == ZEND_ASSIGN_ADD || op->opcode == ZEND_ASSIGN_SUB || 
		    op->opcode == ZEND_ASSIGN_MUL || op->opcode == ZEND_ASSIGN_DIV || op->opcode == ZEND_ASSIGN_MOD ||
		    op->opcode == ZEND_SL || op->opcode == ZEND_SR || op->opcode == ZEND_ASSIGN_BW_AND || 
		    op->opcode == ZEND_ASSIGN_BW_OR || op->opcode == ZEND_ASSIGN_SR || op->opcode == ZEND_ASSIGN_SL) &&    
		    ((OP2_TYPE(op) == IS_CONST && (OP1_TYPE(op) == IS_TMP_VAR || OP1_TYPE(op) == IS_CV || OP1_TYPE(op) == IS_VAR)) || 
		    (OP1_TYPE(op) == IS_CONST && (OP2_TYPE(op) == IS_TMP_VAR || OP2_TYPE(op) == IS_CV || OP2_TYPE(op) == IS_VAR))) &&
		    NEXT_OP(op)->opcode != ZEND_OP_DATA)
		{
			optimize_binary_op_1const(&prev, &op, cb, Ts OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
		}


		/* output block optimization */
		if (op->opcode == ZEND_INIT_STRING && RES_TYPE(op) == IS_TMP_VAR) {
			zend_op *x = NEXT_OP(op);
			/* validation check */
			while (x->opcode == ZEND_ADD_STRING || x->opcode == ZEND_ADD_VAR) {
				x++;
			}
			if (x->opcode == ZEND_ECHO) {
				zend_op *s = NEXT_OP(op);
				UNDEFINE_OP(op->op1);
				SET_TO_NOP(op);

				while (s < x) {
					s->opcode = ZEND_ECHO;
					RES_TYPE(s) = IS_UNUSED;
					memcpy(&s->op1, &s->op2, sizeof(s->op2));
					OP2_TYPE(s) = IS_UNUSED;
					s++;
				}

				UNDEFINE_OP(x->op1);
				SET_TO_NOP(x);
			}
		}		

		
		/* Optimize function calls */
		if (op->opcode == ZEND_DO_FCALL && OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_STRING) {
			optimize_fcall(&op, &prev, cb, op_array, Ts, &fh OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
		}	
		

		/* optimize out echo ''; lines (common issue with Smarty templates) */
		if (op->opcode == ZEND_ECHO && OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_STRING && OP1_VAL_SL(op) == 0) {
			efree(OP1_VAL_S(op));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out known constants */
		else if (op->opcode == ZEND_FETCH_CONSTANT && RES_TYPE(op) == IS_TMP_VAR && OP1_TYPE(op) == IS_UNUSED && OP2_TYPE(op) == IS_CONST && !op->extended_value) {
			zval c;
			if (optimize_get_constant(OP2_VAL_S(op), OP2_VAL_SL(op), NULL, 0, &c TSRMLS_CC)) {
				UNDEFINE_OP(op->op1);
				zval_dtor(&__OP2_VAL(op));
				op->opcode = ZEND_QM_ASSIGN;
				OP1_TYPE(op) = IS_CONST;
				OP2_TYPE(op) = IS_UNUSED;
				__OP1_VAL(op) = c;
				zval_copy_ctor(&__OP1_VAL(op));
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
		} 
		/* optimize out known class constants */
		else if (prev != NULL &&
			prev->opcode == ZEND_FETCH_CLASS && OP2_TYPE(prev) == IS_CONST && prev->extended_value == ZEND_FETCH_CLASS_GLOBAL &&
			RES_TYPE(op) == IS_TMP_VAR && op->opcode == ZEND_FETCH_CONSTANT && OP2_TYPE(op) == IS_CONST && OP1_VR(op) == RES_VR(prev) &&
			(
				(OP1_TYPE(NEXT_OP(op)) == IS_TMP_VAR && OP1_VR(NEXT_OP(op)) == RES_VR(op)) ||
				(OP2_TYPE(NEXT_OP(op)) == IS_TMP_VAR && OP2_VR(NEXT_OP(op)) == RES_VR(op))
			)
		) {
			zval c;
			if (optimize_get_constant(OP2_VAL_S(op), OP2_VAL_SL(op), OP2_VAL_S(prev), OP2_VAL_SL(prev), &c TSRMLS_CC)) {
				if (OP1_TYPE(NEXT_OP(op)) == IS_TMP_VAR && OP1_VR(NEXT_OP(op)) == RES_VR(op)) {
					zend_op *x = NEXT_OP(op);

					UNDEFINE_OP(x->op1);
					OP1_TYPE(x) = IS_CONST;
					__OP1_VAL(x) = c;
					zval_copy_ctor(&__OP1_VAL(x));
					INIT_PZVAL(&__OP1_VAL(x));
				}
				if (OP2_TYPE(NEXT_OP(op)) == IS_TMP_VAR && OP2_VR(NEXT_OP(op)) == RES_VR(op)) {
					zend_op *x = NEXT_OP(op);
		
					UNDEFINE_OP(x->op2);
					OP2_TYPE(x) = IS_CONST;
					__OP2_VAL(x) = c;
					zval_copy_ctor(&__OP2_VAL(x));
					INIT_PZVAL(&__OP2_VAL(x));
				}
				UNDEFINE_OP(prev->op1);
				zval_dtor(&__OP2_VAL(prev));
				SET_TO_NOP(prev);
				UNDEFINE_OP(op->op1);
				zval_dtor(&__OP2_VAL(op));
				SET_TO_NOP(op);
			}
		} 
		/* try to resolve partial file paths for include/require statements */
		else if (op->opcode == ZEND_INCLUDE_OR_EVAL && OP1_TYPE(op) == IS_CONST &&
			(OP2_VAL_L(op) & (0x2|0x4|0x8|0x10)) && /* include/require/include_once/require_once */
			OP1_VALT(op) == IS_STRING &&
			!IS_ABSOLUTE_PATH(OP1_VAL_S(op), OP1_VAL_SL(op)) &&
			!php_memnstr(OP1_VAL_S(op), "://", sizeof("://") - 1, OP1_VAL_S(op) + OP1_VAL_SL(op))
		) {
			optimize_fullpath(op, op_array->filename, OP2_VAL_L(op) TSRMLS_CC);
		}
		/* try to optimize out allows $_SERVER['key'] calls */
		else if (server_op && op->opcode == ZEND_FETCH_R && FETCH_TYPE(op) == ZEND_FETCH_GLOBAL &&
			OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_STRING &&
			!strcmp(OP1_VAL_S(op), "_SERVER") && 
			NEXT_OP(op)->opcode == ZEND_FETCH_DIM_R && OP1_VR(NEXT_OP(op)) == RES_VR(op) &&
			OP2_TYPE(NEXT_OP(op)) == IS_CONST && Z_TYPE(__OP2_VAL(NEXT_OP(op))) == IS_STRING &&
			OP2_VAL_SL(NEXT_OP(op)) <= OPTIMIZER_G(optimize_inline_server_l) &&
			optimizeable_server_var(server_op, OP2_VAL_S(NEXT_OP(op)), OP2_VAL_SL(NEXT_OP(op))) &&
			PG(http_globals)[TRACK_VARS_SERVER] &&
			zend_hash_exists(Z_ARRVAL_P(PG(http_globals)[TRACK_VARS_SERVER]), OP2_VAL_S(NEXT_OP(op)), OP2_VAL_SL(NEXT_OP(op))+1)
		) {
			zval **tmp;

			if (zend_hash_find(Z_ARRVAL_P(PG(http_globals)[TRACK_VARS_SERVER]), OP2_VAL_S(NEXT_OP(op)), OP2_VAL_SL(NEXT_OP(op))+1, (void **) &tmp) == SUCCESS) {
				zend_op *x = NEXT_OP(op);

				switch (Z_TYPE_PP(tmp)) {
					case IS_STRING:
						optimize_to_string_ex(Z_STRVAL_PP(tmp), Z_STRLEN_PP(tmp), &op, &x, cb, Ts, OPTIMIZE_TO_FULL TSRMLS_CC);
						OPTIMIZER_STATS_ADD_PEEPHOLE();
						break;
					case IS_LONG:
						optimize_to_long_ex(Z_LVAL_PP(tmp), &op, &x, cb, Ts, OPTIMIZE_TO_FULL TSRMLS_CC);
						OPTIMIZER_STATS_ADD_PEEPHOLE();
						break;
					case IS_DOUBLE:
						optimize_to_double_ex(Z_DVAL_PP(tmp), &op, &x, cb, Ts, OPTIMIZE_TO_FULL TSRMLS_CC);
						OPTIMIZER_STATS_ADD_PEEPHOLE();
						break;
					case IS_BOOL:
						if (Z_LVAL_PP(tmp)) {
							optimize_to_true_ex(&op, &x, cb, Ts, OPTIMIZE_TO_FULL TSRMLS_CC);
						} else {
							optimize_to_false_ex(&op, &x, cb, Ts, OPTIMIZE_TO_FULL TSRMLS_CC);
						}
						OPTIMIZER_STATS_ADD_PEEPHOLE();
						break;
					default: /* shouldn't happen, but can if ARRAY/NULL/RESOURCE/OBJECT types are being used */
						break;
				}
			}
		}
		/* if allowed optimize our assert() calls */
		else if (prev != NULL && op->opcode == ZEND_DO_FCALL && USED_RESULT(op) &&
				op->extended_value == 1 && !strcmp(OP1_VAL_S(op),"assert") &&
				((prev->opcode == ZEND_SEND_VAL && OP1_TYPE(prev) == IS_CONST) || prev->opcode == ZEND_SEND_VAR)
				&& OPTIMIZER_G(optimize_assert) 
		) {
			zval_dtor(&__OP1_VAL(op));
			SET_TO_NOP(op);
			if (prev->opcode != ZEND_SEND_VAR) {
				zval_dtor(&__OP1_VAL(prev));
				SET_TO_NOP(prev);
			}
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* try to optimize out define() calls */
		else if (
			prev != NULL && op->opcode == ZEND_DO_FCALL && OP1_TYPE(op) == IS_CONST &&
			OP1_VALT(op) == IS_STRING && (op->extended_value == 2 || op->extended_value == 3) &&
			!strcmp(OP1_VAL_S(op),"define") && prev->opcode == ZEND_SEND_VAL && OP1_TYPE(prev) == IS_CONST &&
			(op-2)->opcode == ZEND_SEND_VAL && OP1_TYPE((op-2)) == IS_CONST &&
			(op->extended_value < 3 || ((op-3)->opcode == ZEND_SEND_VAL && OP1_TYPE((op-3)) == IS_CONST)) &&
			(!EG(in_execution) || OPTIMIZER_G(optimize_define)) && /* not inside an include */
			!jmp_lines[(op - op_array->opcodes)] /* non-conditional */
		) {
			optimize_define(&prev, &op, cb, Ts TSRMLS_CC);
		}
		/* ternary operator to if() {} else {} optimization */
		else if (
			op->opcode == ZEND_JMPZ && NEXT_OP(op)->opcode == ZEND_QM_ASSIGN && 
			NEXT_OP(NEXT_OP(op))->opcode == ZEND_JMP && NEXT_OP(NEXT_OP(NEXT_OP(op)))->opcode == ZEND_QM_ASSIGN
		) {
			zend_op *x = NEXT_OP(NEXT_OP(NEXT_OP(NEXT_OP(op))));

			if (x->opcode == ZEND_ASSIGN && RES_TYPE(x) == IS_UNUSED) { /* $var = $foo ? a : b */
				NEXT_OP(op)->opcode = NEXT_OP(NEXT_OP(NEXT_OP(op)))->opcode = x->opcode;

				memcpy(&NEXT_OP(op)->op2, &NEXT_OP(op)->op1, sizeof(znode));
				memcpy(&NEXT_OP(NEXT_OP(NEXT_OP(op)))->op2, &NEXT_OP(NEXT_OP(NEXT_OP(op)))->op1, sizeof(znode));

				RESULT_TYPE(NEXT_OP(op)) |= EXT_TYPE_UNUSED;
				RESULT_TYPE(NEXT_OP(NEXT_OP(NEXT_OP(op)))) |= EXT_TYPE_UNUSED;

				UNDEFINE_OP(NEXT_OP(op)->result);
				UNDEFINE_OP(NEXT_OP(NEXT_OP(NEXT_OP(op)))->result);
				UNDEFINE_OP(x->op2);

				RES_TYPE(NEXT_OP(op)) = RES_TYPE(NEXT_OP(NEXT_OP(NEXT_OP(op)))) = IS_UNUSED;

				memcpy(&NEXT_OP(op)->op1, &x->op1, sizeof(znode));
				memcpy(&NEXT_OP(NEXT_OP(NEXT_OP(op)))->op1, &x->op1, sizeof(znode));
				
				SET_TO_NOP(x);
			} else if (x->opcode == ZEND_ECHO) { /* echo $foo ? a : b */
				NEXT_OP(op)->opcode = NEXT_OP(NEXT_OP(NEXT_OP(op)))->opcode = x->opcode;

				UNDEFINE_OP(NEXT_OP(op)->result);
				UNDEFINE_OP(NEXT_OP(NEXT_OP(NEXT_OP(op)))->result);
				UNDEFINE_OP(x->op2);

				RESULT_TYPE(NEXT_OP(op)) |= EXT_TYPE_UNUSED;
				RESULT_TYPE(NEXT_OP(NEXT_OP(NEXT_OP(op)))) |= EXT_TYPE_UNUSED;

				RES_TYPE(NEXT_OP(op)) = RES_TYPE(NEXT_OP(NEXT_OP(NEXT_OP(op)))) = IS_UNUSED;
				SET_TO_NOP(x);
			}
			
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out mathematical operations on constant values */
		else if (
			(
				op->opcode == ZEND_ADD || op->opcode == ZEND_SUB || op->opcode == ZEND_MUL || op->opcode == ZEND_DIV || 
				op->opcode == ZEND_MOD || op->opcode == ZEND_SL || op->opcode == ZEND_SR || op->opcode == ZEND_CONCAT || 
				op->opcode == ZEND_BW_OR || op->opcode == ZEND_BW_AND || op->opcode == ZEND_BW_XOR || 
				op->opcode == ZEND_BOOL_XOR || op->opcode == ZEND_IS_IDENTICAL || op->opcode == ZEND_IS_NOT_IDENTICAL || 
				op->opcode == ZEND_IS_EQUAL || op->opcode == ZEND_IS_NOT_EQUAL || op->opcode == ZEND_IS_SMALLER || 
				op->opcode == ZEND_IS_SMALLER_OR_EQUAL
			) && OP1_TYPE(op) == IS_CONST && OP2_TYPE(op) == IS_CONST && RES_TYPE(op) == IS_TMP_VAR
		) {
			typedef int (*binary_op_type)(zval *, zval *, zval* TSRMLS_DC);

			binary_op_type binary_op = (binary_op_type)get_binary_op(op->opcode);

			if (binary_op != NULL) {
				zval res;

				if (binary_op(&res, &__OP1_VAL(op), &__OP2_VAL(op) TSRMLS_CC) != FAILURE) {
					zval_dtor(&__OP1_VAL(op));
					zval_dtor(&__OP2_VAL(op));
					op->opcode = ZEND_QM_ASSIGN;
					op->extended_value = 0;
					OP1_TYPE(op) = IS_CONST;
					memcpy(&__OP1_VAL(op), &res, sizeof(zval));
					OP2_TYPE(op) = IS_UNUSED;
					
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
			}
		}
		/* optimize unary ops with static constants on both sides */
		else if ((op->opcode == ZEND_BW_NOT || op->opcode == ZEND_BOOL_NOT) && OP1_TYPE(op) == IS_CONST && RES_TYPE(op) == IS_TMP_VAR) {
			int (*unary_op)(zval *result, zval *op1) = unary_op = get_unary_op(op->opcode);

			if (unary_op != NULL) {
				zval res;

				if (unary_op(&res, &__OP1_VAL(op)) != FAILURE) {
					zval_dtor(&__OP1_VAL(op));
					op->opcode = ZEND_QM_ASSIGN;
					op->extended_value = 0;
					OP1_TYPE(op) = IS_CONST;
					memcpy(&__OP1_VAL(op), &res, sizeof(zval));
					OP2_TYPE(op) = IS_UNUSED;
					
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
			}
		}
		/* optimize if (const_value) situations */
		else if ((op->opcode == ZEND_BOOL) && OP1_TYPE(op) == IS_CONST && RES_TYPE(op) == IS_TMP_VAR) {
			zval res;

			ZVAL_BOOL(&res, zend_is_true(&__OP1_VAL(op)));
			zval_dtor(&__OP1_VAL(op));
			op->opcode = ZEND_QM_ASSIGN;
			op->extended_value = 0;
			OP1_TYPE(op) = IS_CONST;
			memcpy(&__OP1_VAL(op), &res, sizeof(zval));
			OP2_TYPE(op) = IS_UNUSED;
			
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize (cast) operations for const values */
		else if (
			op->opcode == ZEND_CAST && OP1_TYPE(op) == IS_CONST && RES_TYPE(op) == IS_TMP_VAR && 
			op->extended_value != IS_ARRAY && op->extended_value != IS_OBJECT && op->extended_value != IS_RESOURCE
		) {
			zval res;
			memcpy(&res, &__OP1_VAL(op), sizeof(zval));
			zval_copy_ctor(&res);

			switch (op->extended_value) {
				case IS_NULL:
					convert_to_null(&res);
					break;
				case IS_BOOL:
					convert_to_boolean(&res);
					break;
				case IS_LONG:
					convert_to_long(&res);
					break;
				case IS_DOUBLE:
					convert_to_double(&res);
					break;
				case IS_STRING:
					convert_to_string(&res);
					break;
				case IS_ARRAY:
					convert_to_array(&res);
					break;
				case IS_OBJECT:
					convert_to_object(&res);
					break;
			}
			zval_dtor(&__OP1_VAL(op));
			op->opcode = ZEND_QM_ASSIGN;
			op->extended_value = 0;
			OP1_TYPE(op) = IS_CONST;
			memcpy(&__OP1_VAL(op), &res, sizeof(zval));
			OP2_TYPE(op) = IS_UNUSED;
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* string creation folding, relating to leftover ops from INIT_STRING */
		else if (op->opcode == ZEND_FREE && OP1_TYPE(op) == IS_CONST) {
			zval_dtor(&__OP1_VAL(op));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* get rid of INIT_STRING */
		else if (op->opcode == ZEND_INIT_STRING) {
			op->opcode = ZEND_QM_ASSIGN;
			OP1_TYPE(op) = IS_CONST;
			OP2_TYPE(op) = IS_UNUSED;
			OP1_VALT(op) = IS_STRING;
			OP1_VAL_SL(op) = 0;
			OP1_VAL_S(op) = STR_EMPTY_ALLOC();
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize ADD_CHAR TO QM_ASSIGN to be later inlined if possible */
		else if (op->opcode == ZEND_ADD_CHAR && OP1_TYPE(op) == IS_CONST) {
			size_t len;

			op->opcode = ZEND_QM_ASSIGN;
			OP1_TYPE(op) = IS_CONST;
			OP2_TYPE(op) = IS_UNUSED;
			convert_to_string(&__OP1_VAL(op));
			len = OP1_VAL_SL(op) + 1;
			STR_REALLOC(OP1_VAL_S(op),len+1);
			OP1_VAL_S(op)[len-1] = (char) OP2_VAL_L(op);
			OP1_VAL_S(op)[len] = 0;
			OP1_VAL_SL(op) = len;
			
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize ADD_STRING TO QM_ASSIGN to be later inlined if possible */
		else if (op->opcode == ZEND_ADD_STRING && OP1_TYPE(op) == IS_CONST) {
			size_t len;

			op->opcode = ZEND_QM_ASSIGN;
			OP1_TYPE(op) = IS_CONST;
			OP2_TYPE(op) = IS_UNUSED;
			convert_to_string(&__OP1_VAL(op));
			convert_to_string(&__OP2_VAL(op));
			len = OP1_VAL_SL(op) + OP2_VAL_SL(op);
			STR_REALLOC(OP1_VAL_S(op),len+1);
			memcpy(OP1_VAL_S(op) + OP1_VAL_SL(op), OP2_VAL_S(op), OP2_VAL_SL(op));
			OP1_VAL_S(op)[len] = 0;
			OP1_VAL_SL(op) = len;
			STR_FREE(OP2_VAL_S(op));
			
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* convert addition of constant to concat */
		else if (op->opcode == ZEND_ADD_VAR && OP1_TYPE(op) == IS_CONST) {
			op->opcode = ZEND_CONCAT;
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out addition of empty character */
		else if (
			op->opcode == ZEND_ADD_CHAR && OP1_TYPE(op) == IS_TMP_VAR && 
			IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_CONCAT && 
			OP1_TYPE(DEFINED_OP(op->op1)) == IS_CONST && OP1_VALT(DEFINED_OP(op->op1)) == IS_STRING 
			&& !OP1_VAL_SL(DEFINED_OP(op->op1))
		) {
			char ch = (char) OP2_VAL_L(op);
			zend_op *x = DEFINED_OP(op->op1);

			UNDEFINE_OP(op->op1);
			memcpy(&op->op1, &x->op2, sizeof(znode));
			op->opcode = ZEND_CONCAT;
			ZVAL_STRINGL(&__OP2_VAL(op), &ch, 1, 1);
			STR_FREE(OP1_VAL_S(x));
			SET_TO_NOP(x);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out addition, concatination and adding of vars to expressions with empty strings */
		else if (
			(op->opcode == ZEND_ADD_STRING || op->opcode == ZEND_CONCAT || op->opcode == ZEND_ADD_VAR) && 
			OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_CONCAT && 
			OP1_TYPE(DEFINED_OP(op->op1)) == IS_CONST && OP1_VALT(DEFINED_OP(op->op1)) == IS_STRING && 
			!OP1_VAL_SL(DEFINED_OP(op->op1))
		) {
			zend_op *x = DEFINED_OP(op->op1);

			UNDEFINE_OP(op->op1);
			op->opcode = ZEND_CONCAT;
			memcpy(&op->op1, &x->op2, sizeof(znode));
			STR_FREE(OP1_VAL_S(x));
			SET_TO_NOP(x);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		} 
		/* optimize out addition of 2 constant chars into a single add string */
		else if (op->opcode == ZEND_ADD_CHAR && OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_ADD_CHAR) {
			char ch1 = (char) OP2_VAL_L(DEFINED_OP(op->op1));
			char ch2 = (char) OP2_VAL_L(op);

			OP2_VALT(DEFINED_OP(op->op1)) = IS_STRING;
			OP2_VAL_S(DEFINED_OP(op->op1)) = emalloc(3);
			OP2_VAL_S(DEFINED_OP(op->op1))[0] = ch1;
			OP2_VAL_S(DEFINED_OP(op->op1))[1] = ch2;
			OP2_VAL_S(DEFINED_OP(op->op1))[2] = '\0';
			OP2_VAL_SL(DEFINED_OP(op->op1)) = 2;
			memcpy(&DEFINED_OP(op->op1)->result, &op->result, sizeof(znode));
			DEFINED_OP(op->op1)->opcode = ZEND_ADD_STRING;
			DEFINE_OP(DEFINED_OP(op->op1));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out addition of const chars to strings or variables */
		else if (
			op->opcode == ZEND_ADD_CHAR && OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && 
			(DEFINED_OP(op->op1)->opcode == ZEND_CONCAT || DEFINED_OP(op->op1)->opcode == ZEND_ADD_STRING) && 
			OP2_TYPE(DEFINED_OP(op->op1)) == IS_CONST
		) {
			size_t len;

			convert_to_string(&__OP2_VAL(DEFINED_OP(op->op1)));
			len = OP2_VAL_SL(DEFINED_OP(op->op1)) + 1;
			STR_REALLOC(OP2_VAL_S(DEFINED_OP(op->op1)),len+1);
			OP2_VAL_S(DEFINED_OP(op->op1))[len-1] = (char) OP2_VAL_L(op);
			OP2_VAL_S(DEFINED_OP(op->op1))[len] = 0;
			OP2_VAL_SL(DEFINED_OP(op->op1)) = len;
			memcpy(&DEFINED_OP(op->op1)->result, &op->result, sizeof(znode));
			if (OP1_TYPE(DEFINED_OP(op->op1)) == RES_TYPE(DEFINED_OP(op->op1)) && OP1_VR(DEFINED_OP(op->op1)) == RES_VR(DEFINED_OP(op->op1))) {
				DEFINED_OP(op->op1)->opcode = ZEND_ADD_STRING;
			}
			DEFINE_OP(DEFINED_OP(op->op1));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize instances such as $a = 'foo'; $a .= 'bar'; */
		else if (
			(op->opcode == ZEND_ASSIGN || op->opcode == ZEND_ASSIGN_CONCAT) && OP2_TYPE(op) == IS_CONST &&
			NEXT_OP(op)->opcode == ZEND_ASSIGN_CONCAT && OP2_TYPE(NEXT_OP(op)) == IS_CONST && OP1_VR(op) == OP1_VR(NEXT_OP(op)) &&
			NEXT_OP(op) < end
		) {
			zend_op *x = NEXT_OP(op);
			size_t len;

			convert_to_string(&__OP2_VAL(op));	
			convert_to_string(&__OP2_VAL(x));

			len = OP2_VAL_SL(op);

			STR_REALLOC(OP2_VAL_S(op), len + OP2_VAL_SL(x) + 1);
			memcpy(OP2_VAL_S(op) + len, OP2_VAL_S(x), OP2_VAL_SL(x) + 1); 
			OP2_VAL_SL(op) += OP2_VAL_SL(x);
			STR_FREE(OP2_VAL_S(x));
 			SET_TO_NOP(x);
 			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimie out concatination of single chars */
		else if (
			(op->opcode == ZEND_ADD_STRING || op->opcode == ZEND_CONCAT) && OP2_TYPE(op) == IS_CONST && 
			OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_ADD_CHAR
		) {
			char ch = (char) OP2_VAL_L(DEFINED_OP(op->op1));
			size_t len;

			convert_to_string(&__OP2_VAL(op));
			len = OP2_VAL_SL(op) + 1;
			OP2_VALT(DEFINED_OP(op->op1)) = IS_STRING;
			OP2_VAL_S(DEFINED_OP(op->op1)) = emalloc(len+1);
			OP2_VAL_S(DEFINED_OP(op->op1))[0] = ch;
			memcpy(OP2_VAL_S(DEFINED_OP(op->op1))+1, OP2_VAL_S(op), OP2_VAL_SL(op));
			OP2_VAL_S(DEFINED_OP(op->op1))[len] = 0;
			OP2_VAL_SL(DEFINED_OP(op->op1)) = len;
			STR_FREE(OP2_VAL_S(op));
			memcpy(&DEFINED_OP(op->op1)->result, &op->result, sizeof(znode));
			DEFINED_OP(op->op1)->opcode = op->opcode;

			if (OP1_TYPE(DEFINED_OP(op->op1)) == RES_TYPE(DEFINED_OP(op->op1)) && OP1_VR(DEFINED_OP(op->op1)) == RES_VR(DEFINED_OP(op->op1))) {
				DEFINED_OP(op->op1)->opcode = ZEND_ADD_STRING;
			}
			DEFINE_OP(DEFINED_OP(op->op1));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* static string addition & concatination optimization */
		else if (
			(op->opcode == ZEND_ADD_STRING || op->opcode == ZEND_CONCAT) && OP2_TYPE(op) == IS_CONST && 
			OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && 
			(DEFINED_OP(op->op1)->opcode == ZEND_CONCAT || DEFINED_OP(op->op1)->opcode == ZEND_ADD_STRING) && 
			OP2_TYPE(DEFINED_OP(op->op1)) == IS_CONST
		) {
			size_t len;

			convert_to_string(&__OP2_VAL(DEFINED_OP(op->op1)));
			convert_to_string(&__OP2_VAL(op));

			len = OP2_VAL_SL(DEFINED_OP(op->op1)) + OP2_VAL_SL(op);
			STR_REALLOC(OP2_VAL_S(DEFINED_OP(op->op1)),len + 1);
			memcpy(OP2_VAL_S(DEFINED_OP(op->op1)) + OP2_VAL_SL(DEFINED_OP(op->op1)), OP2_VAL_S(op), OP2_VAL_SL(op));
			OP2_VAL_S(DEFINED_OP(op->op1))[len] = 0;
			OP2_VAL_SL(DEFINED_OP(op->op1)) = len;
			STR_FREE(OP2_VAL_S(op));
			memcpy(&DEFINED_OP(op->op1)->result, &op->result, sizeof(znode));

			if (op->opcode == ZEND_CONCAT) {
				DEFINED_OP(op->op1)->opcode = ZEND_CONCAT;
			}

			if (OP1_TYPE(DEFINED_OP(op->op1)) == RES_TYPE(DEFINED_OP(op->op1)) && OP1_VR(DEFINED_OP(op->op1)) == RES_VR(DEFINED_OP(op->op1))) {
				DEFINED_OP(op->op1)->opcode = ZEND_ADD_STRING;
			}
			DEFINE_OP(DEFINED_OP(op->op1));
			SET_TO_NOP(op); 
			
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimie out GLOBALS["foo"] into global $foo */
		else if (
			(
				(op->opcode == ZEND_FETCH_DIM_R && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_R) ||
				(op->opcode == ZEND_FETCH_DIM_W && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_W) ||
				(op->opcode == ZEND_FETCH_DIM_RW && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_RW) ||
				(op->opcode == ZEND_FETCH_DIM_IS && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_IS) ||
				(op->opcode == ZEND_FETCH_DIM_FUNC_ARG && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_FUNC_ARG) ||
				(op->opcode == ZEND_FETCH_DIM_UNSET && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_UNSET)
			) && FETCH_TYPE(DEFINED_OP(op->op1)) == ZEND_FETCH_GLOBAL &&
			OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_STRING && 
			!is_numeric_string(OP2_VAL_S(op), OP2_VAL_SL(op), NULL, NULL, 1) &&
			OP1_TYPE(DEFINED_OP(op->op1)) == IS_CONST && OP1_VALT(DEFINED_OP(op->op1)) == IS_STRING &&
			OP1_VAL_SL(DEFINED_OP(op->op1)) == (sizeof("GLOBALS")-1) &&
			!memcmp(OP1_VAL_S(DEFINED_OP(op->op1)), "GLOBALS", sizeof("GLOBALS")-1)
		) {
			zend_op *x = DEFINED_OP(op->op1);
			UNDEFINE_OP(op->op1);
			STR_FREE(OP1_VAL_S(x));
			FETCH_TYPE(x) = ZEND_FETCH_GLOBAL;
			memcpy(&x->op1,&op->op2,sizeof(znode));
			memcpy(&x->result,&op->result,sizeof(znode));
			DEFINE_OP(x);
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out isset()/empty() ops on GLOBALS['foo'] into $foo */
		else if (
			op->opcode == ZEND_ISSET_ISEMPTY_DIM_OBJ && OP1_TYPE(op) == IS_VAR && 
			IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_FETCH_IS &&
			FETCH_TYPE(DEFINED_OP(op->op1)) == ZEND_FETCH_GLOBAL &&
			OP1_TYPE(DEFINED_OP(op->op1)) == IS_CONST && OP1_VALT(DEFINED_OP(op->op1)) == IS_STRING &&
			OP1_VAL_SL(DEFINED_OP(op->op1)) == (sizeof("GLOBALS")-1) &&
			!memcmp(OP1_VAL_S(DEFINED_OP(op->op1)), "GLOBALS", sizeof("GLOBALS")-1)
		) {
			zend_op* x = DEFINED_OP(op->op1);
			STR_FREE(OP1_VAL_S(x));
			x->opcode = ZEND_ISSET_ISEMPTY_VAR;
			x->extended_value = op->extended_value;
			FETCH_TYPE(x) = ZEND_FETCH_GLOBAL;
			memcpy(&x->op1, &op->op2, sizeof(znode));
			memcpy(&x->result, &op->result, sizeof(znode));
			DEFINE_OP(x);
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		else if (op->opcode == ZEND_FREE && OP1_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1)) {
			/* convert post increment to pre increment */
			if (DEFINED_OP(op->op1)->opcode == ZEND_POST_INC && PREV_OP(DEFINED_OP(op->op1))->opcode != ZEND_JMPZNZ) {
				DEFINED_OP(op->op1)->opcode = ZEND_PRE_INC;
				RES_TYPE(DEFINED_OP(op->op1)) = IS_VAR;
				RESULT_TYPE(DEFINED_OP(op->op1)) |= EXT_TYPE_UNUSED;
				UNDEFINE_OP(op->op1);
				SET_TO_NOP(op);
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
			/* convert post decrement to pre decrement */
			else if (DEFINED_OP(op->op1)->opcode == ZEND_POST_DEC && PREV_OP(DEFINED_OP(op->op1))->opcode != ZEND_JMPZNZ) {
				DEFINED_OP(op->op1)->opcode = ZEND_PRE_DEC;
				RES_TYPE(DEFINED_OP(op->op1)) = IS_VAR;
				RESULT_TYPE(DEFINED_OP(op->op1)) |= EXT_TYPE_UNUSED;
				UNDEFINE_OP(op->op1);
				SET_TO_NOP(op);
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
			/* convert print where return value is unused to echo */
			else if (DEFINED_OP(op->op1)->opcode == ZEND_PRINT) {
				DEFINED_OP(op->op1)->opcode = ZEND_ECHO;
				RES_TYPE(DEFINED_OP(op->op1)) = IS_UNUSED;
				UNDEFINE_OP(op->op1);
				SET_TO_NOP(op);
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
			/* convert left over bool into NOP */
			else if (DEFINED_OP(op->op1)->opcode == ZEND_BOOL) {
				SET_TO_NOP(DEFINED_OP(op->op1));
				UNDEFINE_OP(op->op1);
				SET_TO_NOP(op);
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
		} 
		/* convert conditional bool to just bool */
		else if (
			op->opcode == ZEND_BOOL && OP1_TYPE(op) == IS_TMP_VAR && 
			(!global[VAR_NUM(OP1_VR(op))] || (RES_TYPE(op) == IS_TMP_VAR && OP1_VR(op) == RES_VR(op))) && 
			IS_DEFINED(op->op1) && (
				DEFINED_OP(op->op1)->opcode == ZEND_IS_IDENTICAL ||
				DEFINED_OP(op->op1)->opcode == ZEND_IS_NOT_IDENTICAL ||
				DEFINED_OP(op->op1)->opcode == ZEND_IS_EQUAL ||
				DEFINED_OP(op->op1)->opcode == ZEND_IS_NOT_EQUAL ||
				DEFINED_OP(op->op1)->opcode == ZEND_IS_SMALLER ||
				DEFINED_OP(op->op1)->opcode == ZEND_IS_SMALLER_OR_EQUAL
			)
		) {
			memcpy(&DEFINED_OP(op->op1)->result, &op->result, sizeof(znode));
			DEFINE_OP(DEFINED_OP(op->op1));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize boolean ops and booleans & basic jmps */
		else if (
			(
				op->opcode == ZEND_BOOL || op->opcode == ZEND_BOOL_NOT || op->opcode == ZEND_JMPZ ||
				op->opcode == ZEND_JMPNZ || op->opcode == ZEND_JMPZNZ || op->opcode == ZEND_JMPZ_EX ||
				op->opcode == ZEND_JMPNZ_EX
			) &&
			OP1_TYPE(op) == IS_TMP_VAR && 
			(!global[VAR_NUM(OP1_VR(op))] || (RES_TYPE(op) == IS_TMP_VAR && OP1_VR(op) == RES_VR(op))) &&
			IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_BOOL
		) {
			zend_op *x = DEFINED_OP(op->op1);
			UNDEFINE_OP(op->op1);
			memcpy(&op->op1, &x->op1, sizeof(znode));
			SET_TO_NOP(x);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize boolean NOT ops and booleans & basic jmps */
		else if (
			(op->opcode == ZEND_BOOL || op->opcode == ZEND_BOOL_NOT || op->opcode == ZEND_JMPZ || op->opcode == ZEND_JMPNZ) && 
			OP1_TYPE(op) == IS_TMP_VAR && 
			(!global[VAR_NUM(OP1_VR(op))] || (RES_TYPE(op) == IS_TMP_VAR && OP1_VR(op) == RES_VR(op))) &&
			IS_DEFINED(op->op1) && DEFINED_OP(op->op1)->opcode == ZEND_BOOL_NOT
		) {
			zend_op *x = DEFINED_OP(op->op1);

			switch (op->opcode) {
				case ZEND_BOOL:
					op->opcode = ZEND_BOOL_NOT;
					break;
				case ZEND_BOOL_NOT:
					op->opcode = ZEND_BOOL;
					break;
				case ZEND_JMPZ:
					op->opcode = ZEND_JMPNZ;
					break;
				case ZEND_JMPNZ:
					op->opcode = ZEND_JMPZ;
					break;
			}
			UNDEFINE_OP(op->op1);
			memcpy(&op->op1, &x->op1, sizeof(znode));
			SET_TO_NOP(x);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out pointless assignment */
		else if (op->opcode == ZEND_QM_ASSIGN && OP1_TYPE(op) == IS_TMP_VAR && RES_TYPE(op) == IS_TMP_VAR && OP1_VR(op) == RES_VR(op)) {
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		} 
		/* optimize array breakdown of known values */
		else if (
			op->opcode == ZEND_QM_ASSIGN && OP1_TYPE(op) == IS_TMP_VAR && !global[VAR_NUM(OP1_VR(op))] && 
			OP1_VR(op) != RES_VR(op) && IS_DEFINED(op->op1)
		) {
			zend_op *x = DEFINED_OP(op->op1);

			switch (x->opcode) {
				case ZEND_ADD_ARRAY_ELEMENT:
				case ZEND_ADD_STRING:
				case ZEND_ADD_CHAR:
				case ZEND_ADD_VAR:
					break;
				default:
					UNDEFINE_OP(op->op1);
					memcpy(&x->result, &op->result, sizeof(znode));
					DEFINE_OP(x);
					SET_TO_NOP(op);
			}
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* combine multiple echo calls with const params into a single echo */
		else if (prev != NULL && op->opcode == ZEND_ECHO && OP1_TYPE(op) == IS_CONST && prev->opcode == ZEND_ECHO && OP1_TYPE(prev) == IS_CONST) {
			int len;

			convert_to_string(&__OP1_VAL(op));
			//convert_to_string(&__OP1_VAL(op));
			convert_to_string(&__OP1_VAL(prev));

			len = OP1_VAL_SL(prev) + OP1_VAL_SL(op);
			STR_REALLOC(OP1_VAL_S(prev), len + 1);
			memcpy(OP1_VAL_S(prev) + OP1_VAL_SL(prev), OP1_VAL_S(op), OP1_VAL_SL(op));
			OP1_VAL_S(prev)[len] = 0;
			OP1_VAL_SL(prev) = len;
			STR_FREE(OP1_VAL_S(op));
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* convert empty silence blocks to NOP */
		else if (prev != NULL && prev->opcode == ZEND_END_SILENCE && op->opcode == ZEND_BEGIN_SILENCE) {
			zend_op *x = op+1;

			while (x < end) {
				if (x->opcode == ZEND_END_SILENCE && OP1_VR(x) == RES_VR(op)) {
					OP1_VR(x) = OP1_VR(prev);
					SET_TO_NOP(prev);
					SET_TO_NOP(op);
					break;
				}
				x++;
			}
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* convert empty silence blocks to NOP */
		else if (prev != NULL && prev->opcode == ZEND_BEGIN_SILENCE && op->opcode == ZEND_END_SILENCE && RES_VR(prev) == OP1_VR(op)) {
			SET_TO_NOP(prev);
			SET_TO_NOP(op);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize out pointless send by-ref calls to ops that take args by value */
		else if (op->opcode == ZEND_SEND_VAR_NO_REF && (op->extended_value & ZEND_ARG_COMPILE_TIME_BOUND) && !(op->extended_value & ZEND_ARG_SEND_BY_REF)) {
			op->opcode = ZEND_SEND_VAR;
			op->extended_value = ZEND_DO_FCALL;
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* convert INIT_FCALL_BY_NAME and DO_FCALL_BY_NAME to DO_FCALL */
		else if (
			prev != NULL && op->opcode == ZEND_DO_FCALL_BY_NAME && op->extended_value == 0 &&
			OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_STRING && prev->opcode == ZEND_INIT_FCALL_BY_NAME &&
			OP1_TYPE(prev) == IS_UNUSED && OP2_TYPE(prev) == IS_CONST && OP2_VALT(prev) == IS_STRING && 
			OP1_VAL_SL(op) == OP2_VAL_SL(prev) && !memcmp(OP1_VAL_S(op),OP2_VAL_S(prev),OP1_VAL_SL(op))
		) {
			op->opcode = ZEND_DO_FCALL;
			STR_FREE(OP2_VAL_S(prev));
			SET_TO_NOP(prev);
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}
		/* optimize array init such as $a = array(); $a[] = 'foo'; */
		else if (
			op->opcode == ZEND_INIT_ARRAY && NEXT_OP(op)->opcode == ZEND_ASSIGN && 
			OP1_TYPE(op) == IS_UNUSED && OP2_TYPE(op) == IS_UNUSED && NEXT_OP(NEXT_OP(op))->opcode == ZEND_ASSIGN_DIM &&
			OP1_VR(NEXT_OP(NEXT_OP(op))) == OP1_VR(NEXT_OP(op)) && NEXT_OP(NEXT_OP(NEXT_OP(op)))->opcode == ZEND_OP_DATA &&
			(cb->start + cb->len) > NEXT_OP(NEXT_OP(op))
		) {
			zend_op *x = NEXT_OP(op);

			zend_op *key = NEXT_OP(NEXT_OP(op));
			zend_op *val = NEXT_OP(NEXT_OP(NEXT_OP(op)));
			zend_bool move = 0;

			memcpy(&op->op2, &key->op2, sizeof(znode));
			memcpy(&op->op1, &val->op1, sizeof(znode));

			SET_TO_NOP(key);
			SET_TO_NOP(val);

			/* let's see if we have more array elements */
			while (NEXT_OP(val)->opcode == ZEND_ASSIGN_DIM && OP1_VR(NEXT_OP(val)) == OP1_VR(x) && NEXT_OP(NEXT_OP(val))->opcode == ZEND_OP_DATA) {
				SET_TO_NOP(val);

				key = NEXT_OP(val);
				val = NEXT_OP(NEXT_OP(val));

				key->opcode = ZEND_ADD_ARRAY_ELEMENT;
				memcpy(&key->op1, &val->op1, sizeof(znode));
				memcpy(&key->result, &op->result, sizeof(znode));
				move = 1;
			}

			/* move ASSIGN below ADD_ARRAY_ELEMENT */
			if (move) {
				zend_op *x = NEXT_OP(op);
				memcpy(val, x, sizeof(zend_op));
				SET_TO_NOP(x);
			}
			
			OPTIMIZER_STATS_ADD_PEEPHOLE();
		}

		/* optimize operations where a math operation is done on the var itself, $a = $a [op] val; > $a [op]= val; */
		if (op->opcode == ZEND_ASSIGN && OP1_TYPE(op) == IS_VAR && OP2_TYPE(op) == IS_TMP_VAR && IS_DEFINED(op->op1) && IS_DEFINED(op->op2)) {
			zend_op* l = DEFINED_OP(op->op1);
			zend_op* r = DEFINED_OP(op->op2);

			if (l->opcode == ZEND_FETCH_W && OP1_TYPE(l) == IS_CONST && OP1_VALT(l) == IS_STRING &&
					(
						r->opcode == ZEND_ADD || r->opcode == ZEND_SUB || r->opcode == ZEND_MUL ||
						r->opcode == ZEND_DIV || r->opcode == ZEND_MOD || r->opcode == ZEND_SL ||
						r->opcode == ZEND_SR || r->opcode == ZEND_CONCAT || r->opcode == ZEND_BW_OR ||
						r->opcode == ZEND_BW_AND || r->opcode == ZEND_BW_XOR
					) && OP1_TYPE(r) == IS_VAR && IS_DEFINED(r->op1))
			{
				zend_op* rl = DEFINED_OP(r->op1);

				if (rl->opcode == ZEND_FETCH_R && OP1_TYPE(rl) == IS_CONST && OP1_VALT(rl) == IS_STRING &&
					FETCH_TYPE(rl) == FETCH_TYPE(l) && OP1_VAL_SL(l) == OP1_VAL_SL(rl) &&
					!memcmp(OP1_VAL_S(l), OP1_VAL_S(rl), OP1_VAL_SL(l))
				) {
					switch (r->opcode) {
						case ZEND_ADD:
							op->opcode = ZEND_ASSIGN_ADD;
							break;
						case ZEND_SUB:
							op->opcode = ZEND_ASSIGN_SUB;
							break;
						case ZEND_MUL:
							op->opcode = ZEND_ASSIGN_MUL;
							break;
						case ZEND_DIV:
							op->opcode = ZEND_ASSIGN_DIV;
							break;
						case ZEND_MOD:
							op->opcode = ZEND_ASSIGN_MOD;
							break;
						case ZEND_SL:
							op->opcode = ZEND_ASSIGN_SL;
							break;
						case ZEND_SR:
							op->opcode = ZEND_ASSIGN_SR;
							break;
						case ZEND_CONCAT:
							op->opcode = ZEND_ASSIGN_CONCAT;
							break;
						case ZEND_BW_OR:
							op->opcode = ZEND_ASSIGN_BW_OR;
							break;
						case ZEND_BW_AND:
							op->opcode = ZEND_ASSIGN_BW_AND;
							break;
						case ZEND_BW_XOR:
							op->opcode = ZEND_ASSIGN_BW_XOR;
							break;
					}
					memcpy(&op->op2, &r->op2, sizeof(znode));
					l->opcode = ZEND_FETCH_RW;
					SET_TO_NOP(r);
					STR_FREE(OP1_VAL_S(rl));
					SET_TO_NOP(rl);
				}
				
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			}
		}

		/* one time optimizations that the 1st pass should always resolved, so we only try to run them once */
		if (pass == 1) {
			/* FETCH_W var,$x + ASSIGN $x,?,_  + FETCH_R var,$y => FETCH_W var,$x + ASSIGN $x,?,$y */
			if (
				op->opcode == ZEND_UNSET_VAR || op->opcode == ZEND_DO_FCALL || op->opcode == ZEND_DO_FCALL_BY_NAME ||
				op->opcode == ZEND_POST_INC || op->opcode == ZEND_POST_DEC || op->opcode == ZEND_UNSET_DIM ||
				op->opcode == ZEND_UNSET_OBJ || op->opcode == ZEND_INCLUDE_OR_EVAL || op->opcode == ZEND_ASSIGN_DIM ||
				op->opcode == ZEND_ASSIGN_OBJ
			) {
				zend_hash_clean(&assigns);
				zend_hash_clean(&fetch_dim);
				OPTIMIZER_STATS_ADD_PEEPHOLE();
			} else if (
				op->opcode == ZEND_ASSIGN_REF || op->opcode == ZEND_ASSIGN || op->opcode == ZEND_PRE_INC ||
				op->opcode == ZEND_PRE_DEC || op->opcode == ZEND_ASSIGN_ADD || op->opcode == ZEND_ASSIGN_SUB ||
				op->opcode == ZEND_ASSIGN_MUL || op->opcode == ZEND_ASSIGN_DIV || op->opcode == ZEND_ASSIGN_MOD ||
				op->opcode == ZEND_ASSIGN_SL || op->opcode == ZEND_ASSIGN_SR || op->opcode == ZEND_ASSIGN_CONCAT ||
				op->opcode == ZEND_ASSIGN_BW_OR || op->opcode == ZEND_ASSIGN_BW_AND || op->opcode == ZEND_ASSIGN_BW_XOR
			) {
				zend_hash_clean(&assigns);
				zend_hash_clean(&fetch_dim);

				if (USED_RESULT(op) && OP1_TYPE(op) == IS_VAR && op->extended_value != ZEND_ASSIGN_DIM && op->extended_value != ZEND_ASSIGN_OBJ && IS_DEFINED(op->op1)) {
					zend_op *x = DEFINED_OP(op->op1);

					if ((x->opcode == ZEND_FETCH_W || x->opcode == ZEND_FETCH_RW) && OP1_TYPE(x) == IS_CONST && OP1_VALT(x) == IS_STRING) {
						union {
							zend_op *v;
							void *ptr;
						} op_copy;

						char *s = emalloc(OP1_VAL_SL(x) + 2);

						op_copy.v = op;
						memcpy(s, OP1_VAL_S(x), OP1_VAL_SL(x));
						s[OP1_VAL_SL(x)] = (char)FETCH_TYPE(x);
						s[OP1_VAL_SL(x) + 1] = '\0';
						zend_hash_update(&assigns, s, OP1_VAL_SL(x) + 2, &op_copy.ptr, sizeof(void*), NULL);
						efree(s);
					}
					
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
			} else if ((op->opcode == ZEND_FETCH_R || op->opcode == ZEND_FETCH_IS) && !global[VAR_NUM(RES_VR(op))] && OP1_TYPE(op) == IS_CONST && OP1_VALT(op) == IS_STRING) {
				union {
					zend_op *v;
					void *ptr;
				} x;

				char *s = emalloc(OP1_VAL_SL(op) + 2);
				memcpy(s, OP1_VAL_S(op), OP1_VAL_SL(op));
				s[OP1_VAL_SL(op)] = (char)FETCH_TYPE(op);
				s[OP1_VAL_SL(op) + 1] = '\0';

				if (zend_hash_find(&assigns, s, OP1_VAL_SL(op) + 2, &x.ptr) == SUCCESS) {
					x.v = *(zend_op**)x.v;

					memcpy(&x.v->result, &op->result, sizeof(znode));
					RESULT_TYPE(x.v) = 0;
					DEFINE_OP(x.v);
					zend_hash_del(&assigns, s, OP1_VAL_SL(op) + 2);
					STR_FREE(OP1_VAL_S(op));
					SET_TO_NOP(op);
					OPTIMIZER_STATS_ADD_PEEPHOLE();
				}
				efree(s);
			} else if (op->opcode == ZEND_FETCH_DIM_R && op->extended_value != ZEND_FETCH_ADD_LOCK && OP1_TYPE(op) == IS_VAR && IS_DEFINED(op->op1)) {
				zend_op *x = DEFINED_OP(op->op1);

				while ((x->opcode == ZEND_ASSIGN_REF ||
						x->opcode == ZEND_ASSIGN ||
						x->opcode == ZEND_PRE_INC ||
						x->opcode == ZEND_PRE_DEC ||
						x->opcode == ZEND_ASSIGN_ADD ||
						x->opcode == ZEND_ASSIGN_SUB ||
						x->opcode == ZEND_ASSIGN_MUL ||
						x->opcode == ZEND_ASSIGN_DIV ||
						x->opcode == ZEND_ASSIGN_MOD ||
						x->opcode == ZEND_ASSIGN_SL ||
						x->opcode == ZEND_ASSIGN_SR ||
						x->opcode == ZEND_ASSIGN_CONCAT ||
						x->opcode == ZEND_ASSIGN_BW_OR ||
						x->opcode == ZEND_ASSIGN_BW_AND ||
						x->opcode == ZEND_ASSIGN_BW_XOR) &&
						OP1_TYPE(x) == IS_VAR &&
						IS_DEFINED(x->op1)) 
				{
					x = DEFINED_OP(x->op1);
				}
				if ((x->opcode == ZEND_FETCH_R || x->opcode == ZEND_FETCH_W || x->opcode == ZEND_FETCH_RW) && OP1_TYPE(x) == IS_CONST && OP1_VALT(x) == IS_STRING) {
					union {
						zend_op *v;
						void *ptr;
					} y;
					union {
						zend_op *v;
						void *ptr;
					} op_copy;

					char *s = emalloc(OP1_VAL_SL(x) + 2);

					op_copy.v = op;
					memcpy(s, OP1_VAL_S(x), OP1_VAL_SL(x));
					s[OP1_VAL_SL(x)] = (char)FETCH_TYPE(x);
					s[OP1_VAL_SL(x) + 1] = '\0';

					if (zend_hash_find(&fetch_dim, s, OP1_VAL_SL(x) + 2, &y.ptr) == SUCCESS) {
						y.v = *(zend_op**)y.v;
						y.v->extended_value = ZEND_FETCH_ADD_LOCK;
						zend_hash_update(&fetch_dim, s, OP1_VAL_SL(x) + 2, &op_copy.ptr, sizeof(void*), NULL);
						UNDEFINE_OP(x->result);
						STR_FREE(OP1_VAL_S(x));
						SET_TO_NOP(x);
						memcpy(&op->op1, &y.v->op1, sizeof(znode));
					} else {
						zend_hash_update(&fetch_dim, s, OP1_VAL_SL(x) + 2, &op_copy.ptr, sizeof(void*), NULL);
					}
					efree(s);
				}
			}
		}

		if (op->opcode != ZEND_NOP) {
			prev = op;
		}

		if ((RES_TYPE(op) == IS_VAR && (op->opcode == ZEND_RECV || op->opcode == ZEND_RECV_INIT || !USED_RESULT(op))) || (RES_TYPE(op) == IS_TMP_VAR)) {
			if (op->opcode == ZEND_RECV || op->opcode == ZEND_RECV_INIT) {
				UNDEFINE_OP(op->result);
			} else {
				DEFINE_OP(op);
			}
		}
		++op;
	}

	/* Remove left over NOP opcodes */
	op = cb->start;
	end = op + cb->len;
	while (op < end) {
		if (op->opcode == ZEND_NOP) {
			zend_op *next = op + 1;
			while (next < end && next->opcode == ZEND_NOP) {
				next++;
			}
			if (next < end) {
				memcpy(op, next,(end - next) * sizeof(zend_op));
				while (next > op) {
					--end;
					SET_TO_NOP(end);
					--next;
				}
			} else {
				end -= (next-op);
			}
		} else {
			++op;
		}
	}
	cb->len = end - cb->start;

	zend_hash_destroy(&fetch_dim);
	zend_hash_destroy(&assigns);
	free_alloca(Ts);
}



/*
 * This is a direct copy from old optimizer, APC makes a mess of oplines, let is restore them back to sane values
 */
static void apc_restore_opline_num(zend_op_array *op_array)
{
	zend_op *opline, *end;
	opline = op_array->opcodes;
	end = opline + op_array->last;

	while (opline < end) {
		switch (opline->opcode) {
			case ZEND_JMP:
				opline->op1.u.opline_num = opline->op1.u.jmp_addr - op_array->opcodes;
				break;
			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
			case ZEND_JMPNZ_EX:
				opline->op2.u.opline_num = opline->op2.u.jmp_addr - op_array->opcodes;
				break;
		}
		opline++;
	}
}



/*
 * Find All Blocks in op_array and build Control Flow Graph (CFG)
 */
static zend_bool build_cfg(zend_op_array *op_array, CB* cb)
{
	zend_op* op = op_array->opcodes;
	int len = op_array->last;
	int line_num;
	CB* p;
	zend_bool remove_brk_cont_array = 1;
	zend_uint deep_catch;

	/* APC specific cleanup */
	apc_restore_opline_num(op_array);

	if (op_array->last_try_catch > 0) {
		int i;
		zend_try_catch_element* tc_element = op_array->try_catch_array;

		for (i = 0; i < op_array->last_try_catch; i++, tc_element++) {
			cb[tc_element->try_op].start = &op_array->opcodes[tc_element->try_op];
			cb[tc_element->try_op].exn = 1;

			cb[tc_element->catch_op].start = &op_array->opcodes[tc_element->catch_op];
			cb[tc_element->catch_op].exn = 1;
			
			cb[tc_element->try_op].jmp_ext = &cb[tc_element->catch_op];
		}
	}

	/* Find the start of all the blocks */
	cb[0].start = op;
	for (line_num = 0; line_num < len; op++,line_num++) {
		const unsigned int mask = op->opcode < MAX_ZOPS ? zops[op->opcode] : -1;
		if (mask != -1) {
			if ((mask & OP1_MASK) == OP1_UCLASS) {
				if (OP1_TYPE(op) != IS_UNUSED) {
					OP1_TYPE(op) = IS_VAR;
				}
			} else if ((mask & OP1_MASK) == OP1_CLASS) {
				OP1_TYPE(op) = IS_VAR;
			} else if ((mask & OP1_MASK) == OP1_UNUSED) {
				OP1_TYPE(op) = IS_UNUSED;
			} 

			if ((mask & OP2_MASK) == OP2_CLASS) {
				OP2_TYPE(op) = IS_VAR;
			} else if ((mask & OP2_MASK) == OP2_UNUSED) {
				OP2_TYPE(op) = IS_UNUSED;
			} else if ((mask & OP2_MASK) == OP2_FETCH && FETCH_TYPE(op) == ZEND_FETCH_STATIC_MEMBER) {
				OP2_TYPE(op) = IS_VAR;
			}

			if ((mask & RES_MASK) == RES_CLASS) {
				RES_TYPE(op) = IS_VAR;
				RESULT_TYPE(op) &= ~EXT_TYPE_UNUSED;
			} else if ((mask & RES_MASK) == RES_UNUSED) {
				RES_TYPE(op) = IS_UNUSED;
			}
		}

		switch (op->opcode) {
			case ZEND_RETURN:
			case ZEND_EXIT:
				cb[line_num+1].start = op+1;
				break;
			case ZEND_JMP:
				cb[op->op1.u.opline_num].start = &op_array->opcodes[op->op1.u.opline_num];
				cb[line_num+1].start = op+1;
				break;
			case ZEND_JMPZNZ:
				cb[op->extended_value].start = &op_array->opcodes[op->extended_value];
				cb[op->op2.u.opline_num].start = &op_array->opcodes[op->op2.u.opline_num];
				cb[line_num+1].start = op+1;
				break;
			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
			case ZEND_JMPNZ_EX:
			case ZEND_NEW:
			case ZEND_FE_RESET:
			case ZEND_FE_FETCH:
				cb[line_num+1].start = op+1;
				cb[op->op2.u.opline_num].start = &op_array->opcodes[op->op2.u.opline_num];
				break;
			case ZEND_BRK:
				/* Replace BRK by JMP */
				if (op->op1.u.opline_num == -1) {
					/* nothing to do here */
				} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_LONG) {
					int level = OP2_VAL_L(op);
					int offset = op->op1.u.opline_num;
					zend_brk_cont_element *jmp_to;

					do {
						if (offset == -1) {
							/* invalid brk */
							goto brk_failed;
						}
						jmp_to = &op_array->brk_cont_array[offset];
						offset = jmp_to->parent;
					} while (--level > 0);
					
					op->opcode = ZEND_JMP;
					op->op1.u.opline_num = jmp_to->brk;
					OP2_TYPE(op) = IS_UNUSED;
					op->extended_value = ZEND_BRK; /* Mark the opcode as former ZEND_BRK */
					cb[op->op1.u.opline_num].start = &op_array->opcodes[jmp_to->brk];
				} else {
brk_failed:
					remove_brk_cont_array = 0;
				}
				cb[line_num+1].start = op+1;
				break;
			case ZEND_CONT:
				/* Replace CONT by JMP */
				if (op->op1.u.opline_num == -1) {
					/* nothing to do here */
				} else if (OP2_TYPE(op) == IS_CONST && OP2_VALT(op) == IS_LONG) {
					int level = OP2_VAL_L(op);
					int offset = op->op1.u.opline_num;
					zend_brk_cont_element *jmp_to;

					do {
						if (offset == -1) {
							/* invalid cont */
							goto cont_failed;
						}
						jmp_to = &op_array->brk_cont_array[offset];
						offset = jmp_to->parent;
					} while (--level > 0);

					op->opcode = ZEND_JMP;
					op->op1.u.opline_num = jmp_to->cont;
					OP2_TYPE(op) = IS_UNUSED;
					op->extended_value = ZEND_CONT; /* Mark the opcode as former ZEND_CONT */
					cb[op->op1.u.opline_num].start = &op_array->opcodes[jmp_to->cont];
				} else {
cont_failed:
					remove_brk_cont_array = 0;
				}
				cb[line_num+1].start = op+1;
				break;
			case ZEND_CATCH:
				cb[op->extended_value].start = &op_array->opcodes[op->extended_value];
				cb[line_num+1].start = op+1;
				break;
			case ZEND_THROW:
				if (op->op2.u.opline_num != -1) {
					cb[op->op2.u.opline_num].start = &op_array->opcodes[op->op2.u.opline_num];
				}
				cb[line_num+1].start = op+1;
#if 0
				{
					zend_op *x = op + 1, *e = op_array->opcodes + op_array->last;
					/* throw OP adds JMP to the end of script return for reason yet to be explained to me
					 * we need to kill this pointless JMP to avoid JMP optimiziation making all code following the
					 * JMP from being marked NOP 
					 */
					while (x < e) {
						if (x->opcode == ZEND_JMP && x->op1.u.opline_num == op_array->last - 2) {
							SET_TO_NOP(x);
							break;
						}
						x++;
					}
				}
#endif
				break;
			case ZEND_DO_FCALL:
			case ZEND_DO_FCALL_BY_NAME:
				if (op->op2.u.opline_num != -1) {
					cb[op->op2.u.opline_num].start = &op_array->opcodes[op->op2.u.opline_num];
					cb[line_num+1].start = op+1;
				}
				break;
			case ZEND_UNSET_VAR:
			case ZEND_UNSET_DIM:
			case ZEND_UNSET_OBJ:
				RES_TYPE(op) = IS_UNUSED;
				break;
			default:
				break;
		}
	}

	/* Find lengths of the found blocks and build CFG */
	p = cb;
	for (line_num = 1; line_num < len; line_num++) {
		/* Calculate innermost CATCH op */
		deep_catch = 0;
		if (op_array->last_try_catch > 0) {
			int i;
			zend_try_catch_element* tc_element = op_array->try_catch_array;
			for (i = 0; i<op_array->last_try_catch; i++, tc_element++) {
				/* silence compile warnings, line_num can't be negative here so casting is safe. */
				if (tc_element->try_op <= (zend_uint)line_num-1 && (zend_uint)line_num-1 < tc_element->catch_op && (deep_catch == 0 || deep_catch > tc_element->catch_op)) {
					deep_catch = tc_element->catch_op;
				}
			}
		}
		if (cb[line_num].start != NULL) {
			p->len = cb[line_num].start - p->start;
			p->next = &cb[line_num];
			op = &p->start[p->len-1];
			switch (op->opcode) {
				case ZEND_JMP:
					p->jmp_1 = &cb[op->op1.u.opline_num];
					if (op->extended_value == ZEND_BRK || op->extended_value == ZEND_CONT) {
						/* Converted ZEND_BRK or ZEND_CONT opcode */
						p->follow = (deep_catch > 0) ? &cb[deep_catch] : &cb[len-1];
						op->extended_value = 0;
					}
					break;
				case ZEND_JMPZNZ:
					p->jmp_2 = &cb[op->op2.u.opline_num];
					p->jmp_ext = &cb[op->extended_value];
					break;
				case ZEND_JMPZ:
				case ZEND_JMPNZ:
				case ZEND_JMPZ_EX:
				case ZEND_JMPNZ_EX:
				case ZEND_NEW:
				case ZEND_FE_RESET:
				case ZEND_FE_FETCH:
					p->jmp_2 = &cb[op->op2.u.opline_num];
					p->follow = &cb[line_num];
					break;
				case ZEND_BRK:
				case ZEND_RETURN:
				case ZEND_EXIT:
				case ZEND_CONT:
					p->follow = (deep_catch > 0) ? &cb[deep_catch] : &cb[len-1];
					break;
				case ZEND_DO_FCALL:
				case ZEND_DO_FCALL_BY_NAME:
					if (op->op2.u.opline_num != -1) {
						p->jmp_2 = &cb[op->op2.u.opline_num];
					}
					p->follow = &cb[line_num];
					break;
				case ZEND_CATCH:
					p->jmp_ext = &cb[op->extended_value];
					p->follow = &cb[line_num];
					break;
				case ZEND_THROW:
					if (op->op2.u.opline_num != -1) {
						p->jmp_2 = &cb[op->op2.u.opline_num];
					}
					p->follow = &cb[line_num];
					break;
				default:
					p->follow = &cb[line_num];
			}
			p = &cb[line_num];
		}
	}
	p->len = (op_array->opcodes + op_array->last) - p->start;

	/* Remove Unused brk_cont_array (BRK and CONT instructions replaced by JMP)*/
	if (remove_brk_cont_array) {
		if (op_array->brk_cont_array != NULL) {
			efree(op_array->brk_cont_array);
			op_array->brk_cont_array = NULL;
		}
		op_array->last_brk_cont = 0;
	}
	
	return remove_brk_cont_array;
}

/*
 * Emits Optimized Code
 */
static void generate_cfg(zend_op_array *op_array, CB* cb)
{
	/* Compacting Optimized Code */
	CB* p = cb;
	zend_op* start = op_array->opcodes;
	zend_op* op = start;
	zend_op* end = op + op_array->last;

	while (p != NULL) {
		if (p->used) {
			if (p->len > 0 && op != p->start) {
				memcpy(op, p->start, p->len * sizeof(zend_op));
			}
			p->start = op;
			op += p->len;
		}
		p = p->next;
	}

	op_array->last = op - start;
	op_array->start_op = NULL;

	while (op < end) {
		SET_TO_NOP(op);
		op++;
	}
	

//vld_dump_oparray(op_array);
	/* Set Branch Targets */
	p = cb;
	while (p != NULL) {
		if (p->used && p->len > 0) {
			if (p->jmp_1 != NULL) {
				p->start[p->len - 1].op1.u.opline_num = p->jmp_1->start - start;
			}
			if (p->jmp_2 != NULL) {
				p->start[p->len - 1].op2.u.opline_num = p->jmp_2->start - start;
			}
			if (p->jmp_ext != NULL) {
				/*
				 * NEED to limit this to ONLY ops that use extended_value for 
				 *   oplines. Otherwise the ext value for the first op in a try 
				 *   block will be corrupted
				 *
				 */
				if (p->start[p->len - 1].opcode == ZEND_CATCH) {
					p->start[p->len - 1].extended_value = p->jmp_ext->start - start;
				}
			}
		}
		p = p->next;
//vld_dump_oparray(op_array);
	}


	if (op_array->last_try_catch > 0) {
		int i;
		int last_try_catch = op_array->last_try_catch;
		zend_try_catch_element *old_tc_element = op_array->try_catch_array;

		for (i = 0; i < op_array->last_try_catch; i++, old_tc_element++) {
			while (!cb[old_tc_element->try_op].used && cb[old_tc_element->try_op].follow) {
				old_tc_element->try_op += cb[old_tc_element->try_op].start - cb[old_tc_element->try_op].follow->start;
			}
			while (!cb[old_tc_element->catch_op].used && cb[old_tc_element->catch_op].follow) {
				old_tc_element->catch_op += cb[old_tc_element->catch_op].start - cb[old_tc_element->catch_op].follow->start;
			}

			if (cb[old_tc_element->try_op].used && cb[old_tc_element->catch_op].used) {
				old_tc_element->try_op = cb[old_tc_element->try_op].start - start;
				old_tc_element->catch_op = cb[old_tc_element->catch_op].start - start;
			} else { 
				old_tc_element->try_op = 0;
				old_tc_element->catch_op = 0;
				last_try_catch--;
			}
		}

		if (op_array->last_try_catch > last_try_catch) {
			zend_try_catch_element* new_tc_array = NULL;
			if (last_try_catch > 0) {
				/* Lost some try & catch blocks */
				zend_try_catch_element* new_tc_element = safe_emalloc(sizeof(zend_try_catch_element), last_try_catch, 0);
				new_tc_array = new_tc_element;
				old_tc_element = op_array->try_catch_array;
				for (i = 0; i < op_array->last_try_catch; i++, old_tc_element++) {
					if (old_tc_element->try_op != old_tc_element->catch_op) {
						new_tc_element->try_op = old_tc_element->try_op;
						new_tc_element->catch_op = old_tc_element->catch_op;
						new_tc_element++;
					}
				}
			}

			/* Otherwise lost all try & catch blocks */
			efree(op_array->try_catch_array);
			op_array->try_catch_array = new_tc_array;
			op_array->last_try_catch = last_try_catch;
		}
	}
}

#define GET_REG(R) {	                                                   \
	if (assigned[(R)] < 0) {	                                           \
		zend_uint j = 0;	                                               \
		while (j < op_array->T) {	                                       \
			if (reg_pool[j] == 0 && (global[(R)] == 0 || used[j] == 0)) {  \
				reg_pool[j] = 1;	                                       \
				assigned[(R)] = j;	                                       \
				if (j + 1 > n) {	                                       \
					n = j + 1;	                                           \
				}	                                                       \
				break;	                                                   \
			}	                                                           \
			j++;	                                                       \
		}	                                                               \
	}	                                                                   \
	used[assigned[(R)]] = 1;	                                           \
}                                                                          \



static void rebuild_var_list(zend_op_array *op_array, CB* p, char *global)
{
	zend_uint i, n = 0;

	int *assigned  = do_alloca(op_array->T * sizeof(int));
	char *reg_pool = do_alloca(op_array->T * sizeof(char));
	char *used     = do_alloca(op_array->T * sizeof(char));

	memset(assigned, -1, op_array->T * sizeof(int));
	memset(reg_pool, 0, op_array->T * sizeof(char));
	memset(used, 0, op_array->T * sizeof(char));

	while (p != NULL) {
		if (p->used && p->len > 0) {
			zend_op* start = p->start;
			zend_op* op = start + p->len;
			zend_op* op_data;

			for (i = 0; i < op_array->T; i++) {
				if (!global[i]) {
					if (assigned[i] >= 0) {
						reg_pool[assigned[i]] = 0;
					}
					assigned[i] = -1;
				}
			}

			while (start < op) {
				--op;
				op_data = NULL;

				if (op->opcode == ZEND_DO_FCALL_BY_NAME && OP1_TYPE(op) == IS_CONST) {
					zval_dtor(&__OP1_VAL(op));
					OP1_TYPE(op) = IS_UNUSED;
				}
				if (OP1_TYPE(op) == IS_VAR || OP1_TYPE(op) == IS_TMP_VAR) {
					int r = VAR_NUM(OP1_VR(op));
					GET_REG(r);

					if (op->opcode == ZEND_DO_FCALL_BY_NAME) {
						OP1_TYPE(op) = IS_UNUSED;
					} else if (op->opcode == ZEND_FETCH_CONSTANT && OP1_TYPE(op) == IS_VAR) {
						OP1_VR(op) = VAR_VAL(assigned[r]);
						/* restore op1 type from VAR to CONST (the opcode handler expects this or bombs out with invalid opcode) */
						OP1_TYPE(op) = IS_CONST;
					} else {
						OP1_VR(op) = VAR_VAL(assigned[r]);
					}
				}
				if (OP2_TYPE(op) == IS_VAR || OP2_TYPE(op) == IS_TMP_VAR) {
					int r = VAR_NUM(OP2_VR(op));
					GET_REG(r);
					OP2_VR(op) = VAR_VAL(assigned[r]);
				}
				if (op->opcode == ZEND_DECLARE_INHERITED_CLASS) {
					int r = VAR_NUM(op->extended_value);
					GET_REG(r);
					op->extended_value = VAR_VAL(assigned[r]);
				}
				if (RES_TYPE(op) == IS_VAR || RES_TYPE(op) == IS_TMP_VAR) {
					int r = VAR_NUM(RES_VR(op));

					GET_REG(r);
					RES_VR(op) = VAR_VAL(assigned[r]);

					if (
						op->opcode != ZEND_RECV && op->opcode != ZEND_RECV_INIT &&
						(
							(RES_TYPE(op) == IS_VAR && USED_RESULT(op)) ||
							(
								!(OP1_TYPE(op) == RES_TYPE(op) && OP1_VR(op) == RES_VR(op)) && 
								!(OP2_TYPE(op) == RES_TYPE(op) && OP2_VR(op) == RES_VR(op)) && 
								!global[r] && op->opcode != ZEND_ADD_ARRAY_ELEMENT
							)
						)
					) {
						reg_pool[VAR_NUM(RES_VR(op))] = 0;
					}
				}
			}
		}
		p = p->next;
	}
	op_array->T = n;

	free_alloca(used);
	free_alloca(reg_pool);
	free_alloca(assigned);
}


#define OPTIMIZER_STATS_CALC(stats_struct, op_array)   \
	stats_struct.op_row_count = op_array->last;        \
	stats_struct.var_count    = op_array->last_var;    \
	stats_struct.temp_count   = op_array->T;           \



/*
 * Main Optimization Routine
 */
zend_op_array *optimize_op_array(zend_op_array *op_array TSRMLS_DC)
{
	CB *p, *cb;
	int i;
	int *jmp_lines;
	HashTable fh;
	
#if OPTIMIZER_TEST_COMPILE
	if (OPTIMIZER_G(optimizer_test)) {
		php_printf("PRE OPTIMIZATION\n");
		optimizer_test_dump_op_array(op_array TSRMLS_CC);
	}
#endif
	
	optimizer_parse_level(TSRMLS_CC);
	
	if (OPTIMIZER_G(optimize_out_funcs) != NULL) {
		OPTIMIZER_G(optimize_out_funcs_l) = strlen(OPTIMIZER_G(optimize_out_funcs));
	} else {
		OPTIMIZER_G(optimize_out_funcs_l) = 0;
	}
	
	if (OPTIMIZER_G(optimize_inline_server) != NULL) {
		OPTIMIZER_G(optimize_inline_server_l) = strlen(OPTIMIZER_G(optimize_inline_server));
	} else {
		OPTIMIZER_G(optimize_inline_server_l) = 0;
	}

#if OPTIMIZER_STATS_COMPILE
	optimizer_stats _op_array_stats = {0};
	optimizer_stats* op_array_stats = &_op_array_stats;
	op_array_stats->optimization_count = 0;
	OPTIMIZER_STATS_CALC(op_array_stats->pre, op_array);
#endif
	

	if (op_array->type != ZEND_USER_FUNCTION) { /* eval() is scary */
		return op_array;
	}

	if ((jmp_lines = do_alloca(sizeof(int) * op_array->last)) == NULL) {
		return NULL;
	}


	fcr_table_ctor(&fh);
	

	/* Allocate memory for CFG */
	if ((cb = do_alloca(sizeof(CB)*(op_array->last+1))) == NULL) {
		free_alloca(jmp_lines);
		fcr_table_dtor(&fh);
		return NULL;
	}
	memset(cb, 0, sizeof(CB) * (op_array->last + 1));

#if OPTIMIZER_DEBUG_COMPILE
	//vld_dump_oparray(op_array);
	optimizer_dump_op_array(op_array);
#endif

	/* Find all blocks and build CFG */
	if (build_cfg(op_array, cb)) {
		char *global = do_alloca(op_array->T * sizeof(char));

		//optimizer_dump_cfg_graphviz(cb, op_array);
		
		if (global == NULL) {
			free_alloca(cb);
			free_alloca(jmp_lines);
			fcr_table_dtor(&fh);
			return NULL;
		}

		int peephole_passes = OPTIMIZER_G(peephole_passes);
		//peephole_passes = 1;
		for (i = 0; i < peephole_passes; i++) {
			/* Determine used blocks and its parent chain */
			mark_used_cb(cb);

			/* JMP Optimization */
			optimize_jmp(cb, op_array, jmp_lines);

			get_var(cb, op_array, global);

			/* Optimize Each block section */
			p = cb;
			while (p != NULL) {
				optimize_code_block(p, op_array, global, i, fh, jmp_lines OPTIMIZER_STATS_PARAM_PASS TSRMLS_CC);
				p = p->next;
			}

			/* Mark All Blocks as unused and destroy and parent chain */
			p = cb;
			while (p != NULL) {
				rm_cb(p);
				p = p->next;
			}
		}
		
		/* Last jump optimization */
		mark_used_cb(cb);
		optimize_jmp(cb, op_array, jmp_lines);
		p = cb;
		while (p != NULL) {
			rm_cb(p);
			p = p->next;
		}
		
		/* Mark used blocks */
		mark_used_cb2(cb);	
		
		/* Remove unused blocks */
		p = cb;
		while (p->next != NULL) {
			if (p->next->used) {
				p = p->next;
			} else {
				del_cb(p->next);
				p->next = p->next->next;
			}
		}
		

#if OPTIMIZER_DEBUG_COMPILE
		//optimizer_dump_cfg_graphviz(cb, op_array);
#endif


		/* Store Optimized Code */
		generate_cfg(op_array, cb);
		rebuild_var_list(op_array, cb, global);
		free_alloca(global);
	}
	free_alloca(cb);
	free_alloca(jmp_lines);
	
	
	fcr_table_dtor(&fh);

    /*
     * part of zend_opcode.c...
     */
	pass_two(op_array TSRMLS_CC);
	
#if OPTIMIZER_STATS_COMPILE
    OPTIMIZER_STATS_CALC(op_array_stats->post, op_array);
    // Replace with a call to a debug function that will dump the info for us
	//php_printf("optimization_count=%d\n\n", op_array_stats->optimization_count);
	//php_printf("op-row count (pre): %d\n\n", op_array_stats->op_row_count_pre);
	//php_printf("STATS DUMP\n");
	optimizer_dump_op_array_stats(op_array_stats, op_array TSRMLS_CC);
	//php_printf("END STATS DUMP\n");
#endif

#if OPTIMIZER_DEBUG_COMPILE
	//vld_dump_oparray(op_array);
	optimizer_dump_op_array(op_array);
#endif

#if OPTIMIZER_TEST_COMPILE
	if (OPTIMIZER_G(optimizer_test)) {
		php_printf("POST OPTIMIZATION\n");
		optimizer_test_dump_op_array(op_array TSRMLS_CC);
	}
#endif

	/* If necessary identify all function calls with static parameter */
	if(OPTIMIZER_G(optimize_detect_static_function_calls) || OPTIMIZER_G(optimize_report_constants)) {
		zend_op* op = op_array->opcodes;	
		int i = 0, j;
		zend_bool ok;
		int len = op_array->last;
		
 		int only_native = OPTIMIZER_G(optimize_report_only_native_func);
 		int const_check = OPTIMIZER_G(optimize_report_constants);
		int func_check  = OPTIMIZER_G(optimize_detect_static_function_calls);


		while (i++ < len) {
			if (func_check && op->opcode == ZEND_DO_FCALL && op->extended_value > 0 && (op-1)->opcode == ZEND_SEND_VAL && OP1_TYPE((op-1)) == IS_CONST) {
				if (only_native) {
					zend_function *func;
					if (zend_hash_find(EG(function_table), OP1_VAL_S(op), OP1_VAL_SL(op) + 1, (void **)&func) == SUCCESS) {
						if (func->type != ZEND_INTERNAL_FUNCTION) {
							op++;
							continue;
						}
					}
				}

				ok = j = 1;
				while (j <= op->extended_value) {
					if ((op-j)->opcode != ZEND_SEND_VAL || OP1_TYPE((op-j)) != IS_CONST) {
						ok = 0;
						break;
					}
					j++;
				}
				if (ok) {
					fprintf(stderr, "Function %s(argc=%ld) on %s:%d\n", OP1_VAL_S(op), op->extended_value, op_array->filename, op->lineno);
				}
			} else if (const_check && op->opcode == ZEND_FETCH_CONSTANT && OP2_TYPE(op) == IS_CONST) {
				if (RES_TYPE(op) == IS_TMP_VAR && OP1_TYPE(op) == IS_UNUSED && !op->extended_value) {
					fprintf(stderr, "Constant %s on %s:%d\n", OP2_VAL_S(op), op_array->filename, op->lineno);
				} else if (i && (op-1)->opcode == ZEND_FETCH_CLASS && OP2_TYPE((op-1)) == IS_CONST && (op-1)->extended_value == ZEND_FETCH_CLASS_GLOBAL) {
					fprintf(stderr, "Class Constant %s::%s on %s:%d\n", OP2_VAL_S(op), OP2_VAL_S((op-1)), op_array->filename, op->lineno);
				}
			}
			op++;
		}
	}
	return op_array;
}


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
